@@ -1,4 +1,4 @@
-{-# LANGUAGE StandaloneDeriving, GeneralizedNewtypeDeriving #-}
+{-# LANGUAGE GeneralizedNewtypeDeriving #-}
 module Darcs.Test.Patch.Properties.Check ( Check(..), checkAPatch ) where
 
 import Prelude ()
@@ -15,7 +15,6 @@ import Darcs.Patch.RegChars ( regChars )
 import Darcs.Util.ByteString ( linesPS )
 import qualified Data.ByteString as B ( ByteString, null, concat )
 import qualified Data.ByteString.Char8 as BC ( break, pack )
-import Darcs.Util.Path ( fn2fp )
 import qualified Data.IntMap as M ( mapMaybe )
 
 import Darcs.Patch ( invert, effect, PrimPatch )
@@ -23,10 +22,11 @@ import Darcs.Patch.Invert ( Invert )
 import Darcs.Patch.V1 ( )
 import Darcs.Patch.V1.Core ( RepoPatchV1(..) )
 import Darcs.Patch.V2.RepoPatch ( RepoPatchV2, isConsistent )
+import Darcs.Patch.V3.Core ( RepoPatchV3 )
 import qualified Darcs.Patch.V1.Prim as V1 ( Prim(..) )
 import qualified Darcs.Patch.V2.Prim as V2 ( Prim(..) )
-import Darcs.Patch.V1.Core ( isMerger )
 import qualified Darcs.Patch.Prim.FileUUID as FileUUID ( Prim )
+import Darcs.Patch.Prim.WithName ( PrimWithName(..) )
 import Darcs.Patch.Prim.V1.Core ( Prim(..), DirPatchType(..), FilePatchType(..) )
 import Darcs.Patch.Witnesses.Ordered
 
@@ -50,42 +50,48 @@ checkAPatch p = doCheck $ do checkPatch p
 instance PrimPatch prim => Check (RepoPatchV2 prim) where
   checkPatch p = maybe isValid (const inconsistent) $ isConsistent p
 
-instance Check (RepoPatchV1 Prim1) where
-   checkPatch p | isMerger p = checkPatch $ effect p
-   checkPatch (Merger _ _ _ _) = impossible
-   checkPatch (Regrem _ _ _ _) = impossible
-   checkPatch (PP p) = checkPatch p
+instance (PrimPatch prim, Check prim) => Check (RepoPatchV1 prim) where
+  checkPatch = checkPatch . effect
+
+instance Check prim => Check (RepoPatchV3 name prim) where
+  checkPatch = checkPatch . effect
 
 deriving instance Check Prim1
 deriving instance Check Prim2
 
+instance Check prim => Check (PrimWithName name prim) where
+  checkPatch = checkPatch . wnPatch
+
 instance Check FileUUID.Prim where
   checkPatch _ = isValid -- XXX
 
 instance Check Prim where
 
-   checkPatch (FP f RmFile) = removeFile $ fn2fp f
-   checkPatch (FP f AddFile) =  createFile $ fn2fp f
+   checkPatch (FP f RmFile) = removeFile f
+   checkPatch (FP f AddFile) =  createFile f
+   -- This is stupid but was designed that way ages ago:
+   -- empty hunks commute with everything, so the file need
+   -- not even exist, nor the line in the file.
+   -- Perhaps we should avoid generating empty hunks.
+   checkPatch (FP _ (Hunk _ [] [])) = isValid
    checkPatch (FP f (Hunk line old new)) = do
-       fileExists $ fn2fp f
-       mapM_ (deleteLine (fn2fp f) line) old
-       mapM_ (insertLine (fn2fp f) line) (reverse new)
-       isValid
+       fileExists f
+       mapM_ (deleteLine f line) old
+       mapM_ (insertLine f line) (reverse new)
    checkPatch (FP f (TokReplace t old new)) =
-       modifyFile (fn2fp f) (tryTokPossibly t old new)
+       modifyFile f (tryTokPossibly t old new)
    -- note that the above isn't really a sure check, as it leaves PSomethings
    -- and PNothings which may have contained new...
    checkPatch (FP f (Binary o n)) = do
-       fileExists $ fn2fp f
-       mapM_ (deleteLine (fn2fp f) 1) (linesPS o)
-       fileEmpty $ fn2fp f
-       mapM_ (insertLine (fn2fp f) 1) (reverse $ linesPS n)
-       isValid
+       fileExists f
+       mapM_ (deleteLine f 1) (linesPS o)
+       fileEmpty f
+       mapM_ (insertLine f 1) (reverse $ linesPS n)
 
-   checkPatch (DP d AddDir) = createDir $ fn2fp d
-   checkPatch (DP d RmDir) = removeDir $ fn2fp d
+   checkPatch (DP d AddDir) = createDir d
+   checkPatch (DP d RmDir) = removeDir d
 
-   checkPatch (Move f f') = checkMove (fn2fp f) (fn2fp f')
+   checkPatch (Move f f') = checkMove f f'
    checkPatch (ChangePref _ _ _) = isValid
 
 tryTokPossibly :: String -> String -> String
