@@ -16,7 +16,7 @@
 -- Boston, MA 02110-1301, USA.
 
 {-# OPTIONS_GHC -fno-warn-missing-methods #-}
-{-# LANGUAGE MultiParamTypeClasses, GeneralizedNewtypeDeriving #-}
+{-# LANGUAGE MultiParamTypeClasses #-}
 
 module Darcs.Repository.ApplyPatches
     ( applyPatches
@@ -25,20 +25,20 @@ module Darcs.Repository.ApplyPatches
     , DefaultIO, runDefault
     ) where
 
-import Prelude hiding ( Applicative )
 import Control.Exception ( catch, SomeException, IOException )
 import Data.Char ( toLower )
 import Data.List ( isSuffixOf )
 import System.IO ( stderr )
 import System.IO.Error ( isDoesNotExistError, isPermissionError )
 import Control.Monad ( unless, mplus )
-import Control.Applicative (Applicative)
 import System.Directory ( createDirectory,
                           removeDirectory, removeFile,
                           renameFile, renameDirectory,
                           doesDirectoryExist, doesFileExist
                         )
 
+import Darcs.Prelude
+
 import Darcs.Patch.ApplyMonad( ApplyMonad(..), ApplyMonadTree(..) )
 import Darcs.Patch.ApplyPatches ( applyPatches )
 import Darcs.Patch.MonadProgress ( MonadProgress(..), ProgressAction(..) )
@@ -47,14 +47,17 @@ import Darcs.Util.Lock ( writeAtomicFilePS )
 
 import Darcs.Util.Exception ( prettyException )
 import Darcs.Util.Progress ( beginTedious, endTedious, tediousSize, finishedOneIO )
-import Darcs.Util.Printer ( hPutDocLn )
-import Darcs.Util.Printer.Color ( showDoc )
+import Darcs.Util.Printer ( hPutDocLn, renderString )
 import Darcs.Util.External ( backupByCopying, backupByRenaming )
-import Darcs.Util.Path ( FileName, fn2fp )
+import Darcs.Util.Path ( AnchoredPath, anchorPath )
 import qualified Data.ByteString as B (empty, null, readFile)
 
 import Darcs.Util.Tree( Tree )
 
+
+ap2fp :: AnchoredPath -> FilePath
+ap2fp = anchorPath ""
+
 newtype DefaultIO a = DefaultIO { runDefaultIO :: IO a }
     deriving (Functor, Applicative, Monad)
 
@@ -66,7 +69,7 @@ instance MonadProgress DefaultIO where
        mapM_ go items
        endTedious what
     where go item =
-            do finishedOneIO what (showDoc $ paMessage item)
+            do finishedOneIO what (renderString $ paMessage item)
                runDefaultIO (paAction item) `catch` \e ->
                  do hPutDocLn stderr $ paOnError item
                     ioError e
@@ -75,23 +78,23 @@ instance ApplyMonad Tree DefaultIO where
     type ApplyMonadBase DefaultIO = IO
 
 instance ApplyMonadTree DefaultIO where
-    mDoesDirectoryExist = DefaultIO . doesDirectoryExist . fn2fp
+    mDoesDirectoryExist = DefaultIO . doesDirectoryExist . ap2fp
     mChangePref a b c = DefaultIO $ changePrefval a b c
-    mModifyFilePS f j = DefaultIO $ B.readFile (fn2fp f) >>= runDefaultIO . j >>= writeAtomicFilePS (fn2fp f)
-    mCreateDirectory = DefaultIO . createDirectory . fn2fp
+    mModifyFilePS f j = DefaultIO $ B.readFile (ap2fp f) >>= runDefaultIO . j >>= writeAtomicFilePS (ap2fp f)
+    mCreateDirectory = DefaultIO . createDirectory . ap2fp
     mCreateFile f = DefaultIO $
-                    do exf <- doesFileExist (fn2fp f)
-                       if exf then fail $ "File '"++fn2fp f++"' already exists!"
-                              else do exd <- doesDirectoryExist $ fn2fp f
-                                      if exd then fail $ "File '"++fn2fp f++"' already exists!"
-                                             else writeAtomicFilePS (fn2fp f) B.empty
+                    do exf <- doesFileExist (ap2fp f)
+                       if exf then fail $ "File '"++ap2fp f++"' already exists!"
+                              else do exd <- doesDirectoryExist $ ap2fp f
+                                      if exd then fail $ "File '"++ap2fp f++"' already exists!"
+                                             else writeAtomicFilePS (ap2fp f) B.empty
     mRemoveFile f = DefaultIO $
-                    do let fp = fn2fp f
+                    do let fp = ap2fp f
                        x <- B.readFile fp
                        unless (B.null x) $
                             fail $ "Cannot remove non-empty file "++fp
                        removeFile fp
-    mRemoveDirectory = DefaultIO . removeDirectory . fn2fp
+    mRemoveDirectory = DefaultIO . removeDirectory . ap2fp
     mRename a b = DefaultIO $
                   catch
                   (renameDirectory x y `mplus` renameFile x y)
@@ -99,8 +102,8 @@ instance ApplyMonadTree DefaultIO where
                   -- versions of darcs allowed users to rename nonexistent
                   -- files.  :(
                   (\e -> unless (isDoesNotExistError e) $ ioError e)
-      where x = fn2fp a
-            y = fn2fp b
+      where x = ap2fp a
+            y = ap2fp b
 
 class (Functor m, Monad m) => TolerantMonad m where
     warning :: IO () -> m ()
@@ -155,13 +158,13 @@ instance TolerantMonad m => ApplyMonadTree (TolerantWrapper m) where
                                  (\(e :: IOException) ->
                                    if "(Directory not empty)" `isSuffixOf` show e
                                    then ioError $ userError $
-                                            "Not deleting " ++ fn2fp d ++ " because it is not empty."
+                                            "Not deleting " ++ ap2fp d ++ " because it is not empty."
                                    else ioError $ userError $
-                                            "Not deleting " ++ fn2fp d ++ " because:\n" ++ show e)
+                                            "Not deleting " ++ ap2fp d ++ " because:\n" ++ show e)
     mRename a b = warning $ catch
                           (let do_backup = if map toLower x == map toLower y
-                                           then backupByCopying (fn2fp b) -- avoid making the original vanish
-                                           else backupByRenaming (fn2fp b)
+                                           then backupByCopying (ap2fp b) -- avoid making the original vanish
+                                           else backupByRenaming (ap2fp b)
                            in do_backup >> runDefaultIO (mRename a b))
                           (\e -> case () of
                                  _ | isPermissionError e -> ioError $ userError $
@@ -171,9 +174,9 @@ instance TolerantMonad m => ApplyMonadTree (TolerantWrapper m) where
                                    | otherwise -> ioError e
                           )
        where
-        x = fn2fp a
-        y = fn2fp b
+        x = ap2fp a
+        y = ap2fp b
         couldNotRename = "Could not rename " ++ x ++ " to " ++ y
 
-backup :: FileName -> IO ()
-backup f = backupByRenaming (fn2fp f)
+backup :: AnchoredPath -> IO ()
+backup f = backupByRenaming (ap2fp f)
