@@ -18,80 +18,101 @@ should attempt to automatically create the patch-index.
 
 See <http://darcs.net/Internals/PatchIndex> for more information.
 -}
-module Darcs.Repository.PatchIndex (
-  doesPatchIndexExist,
-  isPatchIndexDisabled,
-  isPatchIndexInSync,
-  canUsePatchIndex,
-  createPIWithInterrupt,
-  createOrUpdatePatchIndexDisk,
-  deletePatchIndex,
-  attemptCreatePatchIndex,
-  PatchFilter,
-  maybeFilterPatches,
-  getRelevantSubsequence,
-  dumpPatchIndex,
-  piTest
-) where
+module Darcs.Repository.PatchIndex
+    ( doesPatchIndexExist
+    , isPatchIndexDisabled
+    , isPatchIndexInSync
+    , canUsePatchIndex
+    , createPIWithInterrupt
+    , createOrUpdatePatchIndexDisk
+    , deletePatchIndex
+    , attemptCreatePatchIndex
+    , PatchFilter
+    , maybeFilterPatches
+    , getRelevantSubsequence
+    , dumpPatchIndex
+    , piTest
+    ) where
 
-import Prelude ()
 import Darcs.Prelude
 
-import Data.Binary ( Binary, encodeFile, decodeFileOrFail )
-import Data.Word ( Word32 )
-import Data.Int ( Int8 )
-import Data.List ( group, mapAccumL, sort, isPrefixOf, nub, (\\) )
-import Data.Maybe ( fromJust, fromMaybe, isJust )
-import Data.Set (Set)
-import Data.Map (Map)
-import qualified Data.Map as M
-import qualified Data.Set as S
 import Control.Exception ( catch )
 import Control.Monad ( forM_, unless, when )
 import Control.Monad.State.Strict ( evalState, execState, State, gets, modify )
-import System.Directory ( createDirectory, renameDirectory, doesFileExist, doesDirectoryExist )
-import Darcs.Repository.Format ( formatHas, RepoProperty( HashedInventory ) )
-import Darcs.Repository.InternalTypes ( Repository, repoLocation, repoFormat )
-import Darcs.Patch.Witnesses.Ordered ( mapFL, RL(..), FL(..), reverseRL )
-import Darcs.Patch.Witnesses.Sealed ( Sealed2(..), Sealed(..), seal, seal2, unsafeUnseal )
-import Darcs.Patch.PatchInfoAnd ( PatchInfoAnd(..), info )
-import Darcs.Util.Lock ( withPermDir, rmRecursive )
+
+import Data.Binary ( Binary, encodeFile, decodeFileOrFail )
+import qualified Data.ByteString as B
+import Data.Int ( Int8 )
+import Data.List ( group, mapAccumL, sort, nub, (\\) )
+import Data.Maybe ( fromJust, fromMaybe, isJust )
+import qualified Data.Map as M
+import qualified Data.Set as S
+import Data.Word ( Word32 )
+
+import System.Directory
+    ( createDirectory
+    , doesDirectoryExist
+    , doesFileExist
+    , removeDirectoryRecursive
+    , removeFile
+    , renameDirectory
+    )
+import System.FilePath( (</>) )
+import System.IO ( openFile, IOMode(WriteMode), hClose )
+
 import Darcs.Patch ( RepoPatch, listTouchedFiles )
-import Darcs.Util.Path ( FileName, fp2fn, fn2fp, toFilePath )
 import Darcs.Patch.Apply ( ApplyState(..) )
-import Darcs.Patch.Set ( PatchSet(..), patchSet2FL, Origin, patchSet2FL )
-import Darcs.Patch.Inspect ( PatchInspect )
-import Darcs.Util.Global ( darcsdir )
-import Darcs.Util.Progress ( debugMessage )
 import Darcs.Patch.Index.Types
 import Darcs.Patch.Index.Monad ( applyToFileMods, makePatchID )
-import System.FilePath( (</>) )
-import System.IO (openFile, IOMode(WriteMode), hClose)
-import qualified Data.ByteString as B
+import Darcs.Patch.Inspect ( PatchInspect )
+import Darcs.Patch.PatchInfoAnd ( PatchInfoAnd, info )
+import Darcs.Patch.Progress (progressFL )
+import Darcs.Patch.Set ( PatchSet, patchSet2FL, Origin, patchSet2FL )
+import Darcs.Patch.Witnesses.Ordered ( mapFL, RL(..), FL(..), reverseRL )
+import Darcs.Patch.Witnesses.Sealed
+    ( Sealed2(..)
+    , Sealed(..)
+    , seal
+    , seal2
+    , unseal
+    , unseal2
+    )
+import Darcs.Patch.Witnesses.Unsafe ( unsafeCoerceP, unsafeCoercePEnd )
+
+import Darcs.Repository.Format ( formatHas, RepoProperty( HashedInventory ) )
+import Darcs.Repository.InternalTypes ( Repository, repoLocation, repoFormat )
+
+import Darcs.Util.Global ( darcsdir )
 import Darcs.Util.Hash ( sha256sum, showAsHex )
-import Darcs.Util.Tree ( Tree(..) )
-import Darcs.Patch.Witnesses.Unsafe ( unsafeCoerceP )
+import Darcs.Util.Lock ( withPermDir )
+import Darcs.Util.Path ( AnchoredPath, displayPath, toFilePath, isPrefix )
+import Darcs.Util.Progress ( debugMessage )
 import Darcs.Util.SignalHandler ( catchInterrupt )
+import Darcs.Util.Tree ( Tree(..) )
+
+type Map = M.Map
+type Set = S.Set
 
 data FileIdSpan = FidSpan
-                    !FileId           -- the fileid has some fixed name in the
-                    !PatchId          -- span starting here
-                    !(Maybe PatchId)  -- and (maybe) ending here
-  deriving (Show,Eq,Ord)
+  !FileId                   -- ^ the fileid has some fixed name in the
+  !PatchId                  -- ^ span starting here
+  !(Maybe PatchId)          -- ^ and (maybe) ending here
+  deriving (Show, Eq, Ord)
 
 data FilePathSpan = FpSpan
-                      !FileName         -- the file path has some fixed fileid in the
-                      !PatchId          -- span starting here
-                      !(Maybe PatchId)  -- and (maybe) ending here
-  deriving (Show,Eq,Ord)
+  !AnchoredPath             -- ^ the file path has some fixed fileid in the
+  !PatchId                  -- ^ span starting here
+  !(Maybe PatchId)          -- ^ and (maybe) ending here
+  deriving (Show, Eq, Ord)
 
 -- | info about a given fileid, e.g.. is a file or a directory
-data FileInfo = FileInfo { isFile::Bool,
-                           touching::Set Word32} -- first word of patch hash
-  deriving (Show,Eq,Ord)
+data FileInfo = FileInfo
+  { isFile :: Bool
+  , touching :: Set Word32  -- ^ first word of patch hash
+  } deriving (Show, Eq, Ord)
 
 -- | timespans where a certain filename corresponds to a file with a given id
-type FileIdSpans = Map FileName [FileIdSpan]
+type FileIdSpans = Map AnchoredPath [FileIdSpan]
 
 -- | timespans where a file with a certain id corresponds to given filenames
 type FilePathSpans = Map FileId [FilePathSpan]
@@ -100,39 +121,40 @@ type FilePathSpans = Map FileId [FilePathSpan]
 type InfoMap = Map FileId FileInfo
 
 -- | the patch-index
-data PatchIndex =
-    PatchIndex {
-        -- |all the PatchIds tracked by this patch index, with the most
-        -- recent patch at the head of the list (note, stored in the
-        -- reverse order to this on disk for backwards compatibility
-        -- with an older format).
-        pids::[PatchId],
-        fidspans::FileIdSpans,
-        fpspans::FilePathSpans,
-        infom::InfoMap
-    }
+data PatchIndex = PatchIndex
+  { pids :: [PatchId]
+    -- ^ all the 'PatchId's tracked by this patch index, with the most
+    -- recent patch at the head of the list (note, stored in the
+    -- reverse order to this on disk for backwards compatibility
+    -- with an older format).
+  , fidspans :: FileIdSpans
+  , fpspans :: FilePathSpans
+  , infom :: InfoMap
+  }
 
 -- | On-disk version of patch index
 --   version 1 is the one introduced in darcs 2.10
 --           2 changes the pids order to newer-to-older
+--           3 changes FileName to AnchoredPath everywhere, which has
+--             different Binary (and Ord) instances
 version :: Int8
-version = 2
+version = 3
 
 type PIM a = State PatchIndex a
 
 -- | 'applyPatchMods pmods pindex' applies a list of PatchMods to the given
 --   patch index pindex
-applyPatchMods :: [(PatchId, [PatchMod FileName])] -> PatchIndex -> PatchIndex
+applyPatchMods :: [(PatchId, [PatchMod AnchoredPath])] -> PatchIndex -> PatchIndex
 applyPatchMods pmods pindex =
   flip execState pindex $ mapM_ goList pmods
- where goList :: (PatchId, [PatchMod FileName]) -> PIM ()
+ where goList :: (PatchId, [PatchMod AnchoredPath]) -> PIM ()
        goList (pid, mods) = do
            modify (\pind -> pind{pids = pid:pids pind})
            mapM_ (curry go pid) (nubSeq mods)
        -- nubSeq handles invalid patch in darcs repo:
        --   move with identical target name "rename darcs_patcher to darcs-patcher."
        nubSeq = map head . group
-       go :: (PatchId, PatchMod FileName) -> PIM ()
+       go :: (PatchId, PatchMod AnchoredPath) -> PIM ()
        go (pid, PCreateFile fn) = do
          fid <- createFidStartSpan fn pid
          startFpSpan fid fn pid
@@ -168,7 +190,7 @@ applyPatchMods pmods pindex =
                               ++" in FileIdSpans in duplicate, empty list"
 
 -- | create new filespan for created file
-createFidStartSpan :: FileName -> PatchId -> PIM FileId
+createFidStartSpan :: AnchoredPath -> PatchId -> PIM FileId
 createFidStartSpan fn pstart = do
   fidspans <- gets fidspans
   case M.lookup fn fidspans of
@@ -182,7 +204,7 @@ createFidStartSpan fn pstart = do
       return fid
 
 -- | start new span for name fn for file fid starting with patch pid
-startFpSpan :: FileId -> FileName -> PatchId -> PIM ()
+startFpSpan :: FileId -> AnchoredPath -> PatchId -> PIM ()
 startFpSpan fid fn pstart = modify (\pind -> pind {fpspans=M.alter alt fid (fpspans pind)})
   where alt Nothing = Just [FpSpan fn pstart Nothing]
         alt (Just spans) = Just (FpSpan fn pstart Nothing:spans)
@@ -197,13 +219,13 @@ stopFpSpan fid pend = modify (\pind -> pind {fpspans=M.alter alt fid (fpspans pi
         alt _ = error $ "impossible: span already ended for " ++ show fid
 
 -- | start new span for name fn for file fid starting with patch pid
-startFidSpan :: FileName -> PatchId -> FileId -> PIM ()
+startFidSpan :: AnchoredPath -> PatchId -> FileId -> PIM ()
 startFidSpan fn pstart fid = modify (\pind -> pind {fidspans=M.alter alt fn (fidspans pind)})
   where alt Nothing = Just [FidSpan fid pstart Nothing]
         alt (Just spans) = Just (FidSpan fid pstart Nothing:spans)
 
 -- | stop current span for file name fn
-stopFidSpan :: FileName -> PatchId -> PIM ()
+stopFidSpan :: AnchoredPath -> PatchId -> PIM ()
 stopFidSpan fn pend = modify (\pind -> pind {fidspans=M.alter alt fn (fidspans pind)})
   where alt Nothing = error $ "impossible: no span for " ++ show fn
         alt (Just []) = error $ "impossible: no span for " ++ show fn++", empty list"
@@ -220,19 +242,19 @@ createInfo fid isF = modify (\pind -> pind {infom=M.alter alt fid (infom pind)})
 -- | insert touching patchid for given file id
 insertTouch :: FileId -> PatchId -> PIM ()
 insertTouch fid pid = modify (\pind -> pind {infom=M.alter alt fid (infom pind)})
-  where alt Nothing =  impossible "Fileid does not exist"
+  where alt Nothing =  error "impossible: Fileid does not exist"
         alt (Just (FileInfo isF pids)) = Just (FileInfo isF (S.insert (short pid) pids))
 
 -- | lookup current fid of filepath
-lookupFid :: FileName -> PIM FileId
+lookupFid :: AnchoredPath -> PIM FileId
 lookupFid fn = do
     maybeFid <- lookupFid' fn
     case maybeFid of
-        Nothing -> bug $ "couldn't find " ++ fn2fp fn ++ " in patch index"
+        Nothing -> error $ "couldn't find " ++ displayPath fn ++ " in patch index"
         Just fid -> return fid
 
 -- | lookup current fid of filepatch, returning a Maybe to allow failure
-lookupFid' :: FileName -> PIM (Maybe FileId)
+lookupFid' :: AnchoredPath -> PIM (Maybe FileId)
 lookupFid' fn = do
    fidm <- gets fidspans
    case M.lookup fn fidm of
@@ -241,7 +263,7 @@ lookupFid' fn = do
 
 
 -- | lookup all the file ids of a given path
-lookupFidf' :: FileName -> PIM [FileId]
+lookupFidf' :: AnchoredPath -> PIM [FileId]
 lookupFidf' fn = do
    fidm <- gets fidspans
    case M.lookup fn fidm of
@@ -251,16 +273,17 @@ lookupFidf' fn = do
 
 -- |  return all fids of matching subpaths
 --    of the given filepath
-lookupFids :: FileName -> PIM [FileId]
+lookupFids :: AnchoredPath -> PIM [FileId]
 lookupFids fn = do
    fid_spans <- gets fidspans
-   file_idss <- mapM (lookupFidf' . fp2fn) $ filter (isPrefixOf (fn2fp fn)) (fpSpans2filePaths' fid_spans)
+   file_idss <- mapM lookupFidf' $
+      filter (isPrefix fn) (fpSpans2filePaths' fid_spans)
    return $ nub $ concat file_idss
 
 -- | returns a single file id if the given path is a file
 --   if it is a directory, if returns all the file ids of all paths inside it,
 --   at any point in repository history
-lookupFids' :: FileName -> PIM [FileId]
+lookupFids' :: AnchoredPath -> PIM [FileId]
 lookupFids' fn = do
   info_map <- gets infom
   fps_spans <- gets fpspans
@@ -282,12 +305,12 @@ createPatchIndexDisk
   -> PatchSet rt p Origin wR
   -> IO ()
 createPatchIndexDisk repository ps = do
-  let patches = mapFL Sealed2 $ patchSet2FL ps
+  let patches = mapFL Sealed2 $ progressFL "Create patch index" $ patchSet2FL ps
   createPatchIndexFrom repository $ patches2patchMods patches S.empty
 
 -- | convert patches to patchmods
 patches2patchMods :: (Apply p, PatchInspect p, ApplyState p ~ Tree)
-                  => [Sealed2 (PatchInfoAnd rt p)] -> Set FileName -> [(PatchId, [PatchMod FileName])]
+                  => [Sealed2 (PatchInfoAnd rt p)] -> Set AnchoredPath -> [(PatchId, [PatchMod AnchoredPath])]
 patches2patchMods patches fns = snd $ mapAccumL go fns patches
   where
     go filenames (Sealed2 p) = (filenames', (pid, pmods_effect ++ pmods_dup))
@@ -298,7 +321,7 @@ patches2patchMods patches fns = snd $ mapAccumL go fns patches
             touched pm = case pm of {PTouch f -> [f]; PRename a b -> [a,b];
                                      PCreateDir f -> [f]; PCreateFile f -> [f];
                                      PRemove f -> [f]; _ -> []}
-            touched_all = map fp2fn $ listTouchedFiles p
+            touched_all = listTouchedFiles p
             touched_effect = concatMap touched pmods_effect
             touched_invalid = [ f | (PInvalid f) <- pmods_effect]
             -- listTouchedFiles returns all files that touched by these
@@ -310,7 +333,7 @@ patches2patchMods patches fns = snd $ mapAccumL go fns patches
                                             S.fromList touched_effect)
 
 -- | return set of current filenames in patch index
-fpSpans2fileNames :: FilePathSpans -> Set FileName
+fpSpans2fileNames :: FilePathSpans -> Set AnchoredPath
 fpSpans2fileNames fpSpans =
   S.fromList [fn | (FpSpan fn _ Nothing:_)<- M.elems fpSpans]
 
@@ -325,7 +348,7 @@ removePidSuffix pid2idx oldpids@(oldpid:_) (PatchIndex pids fidspans fpspans inf
                (M.mapMaybe removefp fpspans)
                infom -- leave hashes in infom, false positives are harmless
   where
-    findIdx pid = fromMaybe (impossible "removePidSuffix") (M.lookup pid pid2idx)
+    findIdx pid = fromMaybe (error "impossible case") (M.lookup pid pid2idx)
     oldidx = findIdx oldpid
     from `after` idx = findIdx from > idx
     mto `afterM` idx | Just to <- mto, findIdx to > idx = True
@@ -355,7 +378,7 @@ updatePatchIndexDisk repo patches = do
     let repodir = repoLocation repo
     (_,_,pid2idx,pindex) <- loadPatchIndex repodir
     -- check that patch index is up to date
-    let flpatches = patchSet2FL patches
+    let flpatches = progressFL "Update patch index" $ patchSet2FL patches
     let pidsrepo = mapFL (makePatchID . info) flpatches
         (oldpids,_,len_common) = uncommon (reverse $ pids pindex) pidsrepo
         pindex' = removePidSuffix pid2idx oldpids pindex
@@ -377,7 +400,7 @@ updatePatchIndexDisk repo patches = do
 -- | 'createPatchIndexFrom repo pmods' creates a patch index from the given
 --   patchmods.
 createPatchIndexFrom :: Repository rt p wR wU wT
-                     -> [(PatchId, [PatchMod FileName])] -> IO ()
+                     -> [(PatchId, [PatchMod AnchoredPath])] -> IO ()
 createPatchIndexFrom repo pmods = do
     inv_hash <- getInventoryHash repodir
     storePatchIndex repodir cdir inv_hash (applyPatchMods pmods emptyPatchIndex)
@@ -446,7 +469,7 @@ createOrUpdatePatchIndexDisk :: (RepoPatch p, ApplyState p ~ Tree)
                              => Repository rt p wR wU wT -> PatchSet rt p Origin wR -> IO ()
 createOrUpdatePatchIndexDisk repo ps = do
    let repodir = repoLocation repo
-   rmRecursive (repodir </> darcsdir </> noPatchIndex) `catch` \(_ :: IOError) -> return ()
+   removeFile (repodir </> darcsdir </> noPatchIndex) `catch` \(_ :: IOError) -> return ()
    dpie <- doesPatchIndexExist repodir
    if dpie
       then updatePatchIndexDisk repo ps
@@ -467,7 +490,7 @@ canUsePatchIndex repo = do
      case (piExists, piDisabled) of
         (True, False) -> return True
         (False, True) -> return False
-        (True, True) -> error "patch index exists, and patch index is disabled. run optimize enable-patch-index or disable-patch-index to rectify."
+        (True, True) -> fail "patch index exists, and patch index is disabled. run optimize enable-patch-index or disable-patch-index to rectify."
         (False, False) -> return False
 
 -- | Creates patch-index (ignoring whether it is explicitely disabled).
@@ -498,7 +521,7 @@ isPatchIndexInSync repo = do
 storePatchIndex :: FilePath -> FilePath -> String -> PatchIndex -> IO ()
 storePatchIndex repodir cdir inv_hash (PatchIndex pids fidspans fpspans infom) = do
   createDirectory cdir `catch` \(_ :: IOError) -> return ()
-  tmpdir <- withPermDir (repodir </> "filecache-tmp") $ \dir -> do
+  tmpdir <- withPermDir repodir $ \dir -> do
               debugMessage "About to create patch index..."
               let tmpdir = toFilePath dir
               storeRepoState (tmpdir </> repoStateFile) inv_hash
@@ -508,7 +531,7 @@ storePatchIndex repodir cdir inv_hash (PatchIndex pids fidspans fpspans infom) =
               storeFpMap (tmpdir </> fpMapFile) fpspans
               debugMessage "Patch index created"
               return tmpdir
-  rmRecursive cdir `catch` \(_ :: IOError) -> return ()
+  removeDirectoryRecursive cdir `catch` \(_ :: IOError) -> return ()
   renameDirectory tmpdir cdir
 
 decodeFile :: Binary a => FilePath -> IO a
@@ -577,22 +600,22 @@ deletePatchIndex :: FilePath -> IO ()
 deletePatchIndex repodir = do
     exists <- doesDirectoryExist indexDir
     when exists $
-         rmRecursive indexDir
-            `catch` \(e :: IOError) -> error $ "Error: Could not delete patch index\n" ++ show e
+         removeDirectoryRecursive indexDir
+            `catch` \(e :: IOError) -> fail $ "Error: Could not delete patch index\n" ++ show e
     (openFile (repodir </> darcsdir </> noPatchIndex) WriteMode >>= hClose)
-            `catch` \(e :: IOError) -> error $ "Error: Could not disable patch index\n" ++ show e
+            `catch` \(e :: IOError) -> fail $ "Error: Could not disable patch index\n" ++ show e
 
 dumpRepoState :: [PatchId] -> String
 dumpRepoState = unlines . map pid2string
 
 dumpFileIdSpans :: FileIdSpans -> String
 dumpFileIdSpans fidspans =
-  unlines [fn2fp fn++" -> "++showFileId fid++" from "++pid2string from++" to "++maybe "-" pid2string mto
+  unlines [displayPath fn++" -> "++showFileId fid++" from "++pid2string from++" to "++maybe "-" pid2string mto
            | (fn, fids) <- M.toList fidspans, FidSpan fid from mto <- fids]
 
 dumpFilePathSpans :: FilePathSpans -> String
 dumpFilePathSpans fpspans =
-  unlines [showFileId fid++" -> "++ fn2fp fn++" from "++pid2string from++" to "++maybe "-" pid2string mto
+  unlines [showFileId fid++" -> "++ displayPath fn++" from "++pid2string from++" to "++maybe "-" pid2string mto
            | (fid, fns) <- M.toList fpspans, FpSpan fn from mto <- fns]
 
 dumpTouchingMap :: InfoMap -> String
@@ -602,12 +625,12 @@ dumpTouchingMap infom = unlines [showFileId fid++(if isF then "" else "/")++" ->
 -- | return set of current filepaths in patch index
 fpSpans2filePaths :: FilePathSpans -> InfoMap -> [FilePath]
 fpSpans2filePaths fpSpans infom =
-  sort [fn2fp fn ++ (if isF then "" else "/") | (fid,FpSpan fn _ Nothing:_) <- M.toList fpSpans,
+  sort [displayPath fn ++ (if isF then "" else "/") | (fid,FpSpan fn _ Nothing:_) <- M.toList fpSpans,
                                                 let Just (FileInfo isF _) = M.lookup fid infom]
 
 -- | return set of current filepaths in patch index, for internal use
-fpSpans2filePaths' :: FileIdSpans -> [FilePath]
-fpSpans2filePaths' fidSpans = [fn2fp fp | (fp, _)  <- M.toList fidSpans]
+fpSpans2filePaths' :: FileIdSpans -> [AnchoredPath]
+fpSpans2filePaths' fidSpans = [fp | (fp, _)  <- M.toList fidSpans]
 
 -- | Checks if patch index can be created and build it with interrupt.
 attemptCreatePatchIndex
@@ -630,31 +653,38 @@ canCreatePI repo =
     repodir = repoLocation repo
     doesntHaveHashedInventory = return . not . formatHas HashedInventory
 
--- | Returns an RL in which the order of patches matters. Useful for the @annotate@ command.
---   If patch-index does not exist and is not explicitely disabled, silently create it.
---   (Also, if it is out-of-sync, which should not happen, silently update it).
-getRelevantSubsequence :: (RepoPatch p, ApplyState p ~ Tree, a ~ PatchInfoAnd rt p)
-                       => Sealed ((RL a) wK)          -- ^ Sequence of patches you want to filter
-                       -> Repository rt p wR wU wR    -- ^ The repository (to attempt loading patch-index from its path)
-                       -> PatchSet rt p Origin wR     -- ^ PatchSet of repository (in case we need to create patch-index)
-                       -> [FileName]                  -- ^ File(s) about which you want patches from given sequence
-                       -> IO (Sealed ((RL a) Origin)) -- ^ Filtered sequence of patches.
+-- | Returns an RL in which the order of patches matters. Useful for the
+-- @annotate@ command. If patch-index does not exist and is not explicitely
+-- disabled, silently create it. (Also, if it is out-of-sync, which should not
+-- happen, silently update it).
+getRelevantSubsequence
+    :: (RepoPatch p, ApplyState p ~ Tree, a ~ PatchInfoAnd rt p)
+    => Sealed ((RL a) wK)
+    -- ^ Sequence of patches you want to filter
+    -> Repository rt p wR wU wR
+    -- ^ The repository (to attempt loading patch-index from its path)
+    -> PatchSet rt p Origin wR
+    -- ^ PatchSet of repository (in case we need to create patch-index)
+    -> [AnchoredPath]
+    -- ^ File(s) about which you want patches from given sequence
+    -> IO (Sealed ((RL a) Origin))
+    -- ^ Filtered sequence of patches
 getRelevantSubsequence pxes repository ps fns = do
-   pi@(PatchIndex _ _ _ infom) <- loadSafePatchIndex repository ps
-   let fids = map (\fn -> evalState (lookupFid fn) pi) fns
-       pidss = map ((\(FileInfo _ a) -> a).fromJust.(`M.lookup` infom)) fids
-       pids = S.unions pidss
-   let flpxes = reverseRL $ unsafeUnseal pxes
-   return.seal $ keepElems flpxes NilRL pids
+    pi@(PatchIndex _ _ _ infom) <- loadSafePatchIndex repository ps
+    let fids = map (\fn -> evalState (lookupFid fn) pi) fns
+        pidss = map ((\(FileInfo _ a) -> a) . fromJust . (`M.lookup` infom)) fids
+        pids = S.unions pidss
+    let flpxes = reverseRL $ unseal unsafeCoercePEnd pxes
+    return . seal $ keepElems flpxes NilRL pids
+  where
+    keepElems :: (RepoPatch p, ApplyState p ~ Tree, a ~ PatchInfoAnd rt p)
+              => FL a wX wY -> RL a wB wX -> S.Set Word32 -> RL a wP wQ
+    keepElems NilFL acc _ = unsafeCoerceP acc
+    keepElems (x :>: xs) acc pids
+      | short (makePatchID $ info x) `S.member` pids = keepElems xs (acc :<: x) pids
+      | otherwise = keepElems (unsafeCoerceP xs) acc pids
 
-  where keepElems :: (RepoPatch p, ApplyState p ~ Tree, a ~ PatchInfoAnd rt p)
-                  => FL a wX wY -> RL a wB wX -> S.Set Word32 -> RL a wP wQ
-        keepElems NilFL acc _ = unsafeCoerceP acc
-        keepElems (x:>:xs) acc pids
-          | short (makePatchID $ info x) `S.member` pids = keepElems xs (acc:<:x) pids
-          | otherwise                                    = keepElems (unsafeCoerceP xs) acc pids
-
-type PatchFilter rt p = [FilePath] -> [Sealed2 (PatchInfoAnd rt p)] -> IO [Sealed2 (PatchInfoAnd rt p)]
+type PatchFilter rt p = [AnchoredPath] -> [Sealed2 (PatchInfoAnd rt p)] -> IO [Sealed2 (PatchInfoAnd rt p)]
 
 -- | If a patch index is available, returns a filter that takes a list of files and returns
 --   a @PatchFilter@ that only keeps patches that modify the given list of files.
@@ -671,9 +701,10 @@ maybeFilterPatches repo ps fps ops = do
     if usePI
       then do
         pi@(PatchIndex _ _ _ infom) <- loadSafePatchIndex repo ps
-        let fids = concatMap ((\fn -> evalState (lookupFids' fn) pi). fp2fn) fps
+        let fids = concatMap ((\fn -> evalState (lookupFids' fn) pi)) fps
             npids = S.unions $ map (touching.fromJust.(`M.lookup` infom)) fids
-        return $ filter (flip S.member npids . (\(Sealed2 (PIAP pin _)) -> short $ makePatchID pin)) ops
+        return $ filter
+          (flip S.member npids . (unseal2 (short . makePatchID . info))) ops
       else return ops
 
 -- | Dump information in patch index. Patch-index should be checked to exist beforehand. Read-only.
@@ -713,8 +744,8 @@ piTest repodir = do
           g (FidSpan _ x (Just y)) = [y,x]
           g (FidSpan _ x _) = [x]
           ascTs = reverse . nub . concat $ map g spans
-      unless (isInOrder ascTs pids) (error $ "In order test failed! filename: " ++ show fn)
-      forM_ spans $ \(FidSpan fid _ _) -> unless (M.member fid fpspans) (error $ "Valid file id test failed! fid: " ++ show fid)
+      unless (isInOrder ascTs pids) (fail $ "In order test failed! filename: " ++ show fn)
+      forM_ spans $ \(FidSpan fid _ _) -> unless (M.member fid fpspans) (fail $ "Valid file id test failed! fid: " ++ show fid)
    putStrLn "fidspans tests passed"
 
    -- test fpspans
@@ -725,12 +756,12 @@ piTest repodir = do
           g (FpSpan _ x (Just y)) = [y,x]
           g (FpSpan _ x _) = [x]
           ascTs = reverse . nub . concat $ map g spans
-      unless (isInOrder ascTs pids) (error $ "In order test failed! fileid: " ++ show fid)
-      forM_ spans $ \(FpSpan fn _ _) -> unless (M.member fn fidspans) (error $ "Valid file name test failed! file name: " ++ show fn)
+      unless (isInOrder ascTs pids) (fail $ "In order test failed! fileid: " ++ show fid)
+      forM_ spans $ \(FpSpan fn _ _) -> unless (M.member fn fidspans) (fail $ "Valid file name test failed! file name: " ++ show fn)
       let f :: FilePathSpan -> FilePathSpan -> Bool
           f (FpSpan _ x _) (FpSpan _ _ (Just y)) = x == y
           f _ _ = error "adj test of fpspans fail"
-      unless (and $ zipWith f spans (tail spans)) (error $ "Adjcency test failed! fid: " ++ show fid)
+      unless (and $ zipWith f spans (tail spans)) (fail $ "Adjcency test failed! fid: " ++ show fid)
    putStrLn "fpspans tests passed"
 
    -- test infomap
