@@ -23,7 +23,7 @@ module Darcs.Repository.Prefs
     , getGlobal
     , environmentHelpHome
     , defaultrepo
-    , getDefaultRepoPath
+    , getDefaultRepo
     , addRepoSource
     , getPrefval
     , setPrefval
@@ -31,8 +31,7 @@ module Darcs.Repository.Prefs
     , defPrefval
     , writeDefaultPrefs
     , boringRegexps
-    , boringFileFilter
-    , darcsdirFilter
+    , isBoring
     , FileType(..)
     , filetypeFunction
     , getCaches
@@ -43,17 +42,18 @@ module Darcs.Repository.Prefs
     , showMotd
     , prefsUrl
     , prefsDirPath
+    , prefsFilePath
+    , getPrefLines -- exported for darcsden, don't remove
     -- * documentation of prefs files
     , prefsFilesHelp
     ) where
 
-import Prelude ()
 import Darcs.Prelude
 
 import Control.Exception ( catch )
 import Control.Monad ( unless, when, liftM )
 import Data.Char ( toUpper )
-import Data.List ( nub, isPrefixOf, union, sortBy, lookup )
+import Data.List ( nub, isPrefixOf, union, lookup )
 import Data.Maybe ( isJust, fromMaybe, mapMaybe, catMaybes, maybeToList )
 import qualified Control.Exception as C
 import qualified Data.ByteString       as B  ( empty, null, hPut, ByteString )
@@ -62,19 +62,25 @@ import System.Directory ( getAppUserDataDirectory, doesDirectoryExist,
                           createDirectory, doesFileExist )
 import System.Environment ( getEnvironment )
 import System.FilePath.Posix ( normalise, dropTrailingPathSeparator, (</>) )
-import System.IO.Error ( isDoesNotExistError )
+import System.IO.Error ( isDoesNotExistError, catchIOError )
 import System.IO ( stdout, stderr )
 import System.Info ( os )
+import System.Posix.Files ( getFileStatus, fileOwner )
 import Text.Regex ( Regex, mkRegex, matchRegex )
 
-import Darcs.Repository.Cache ( Cache(..), CacheType(..), CacheLoc(..),
-                                WritableOrNot(..), compareByLocality )
+import Darcs.Repository.Cache ( Cache, mkCache, CacheType(..), CacheLoc(..),
+                                WritableOrNot(..) )
 import Darcs.Util.External ( gzFetchFilePS , fetchFilePS, Cachable(..))
-import Darcs.Repository.Flags( UseCache (..), DryRun (..), SetDefault (..),
-                               RemoteRepos (..) )
+import Darcs.Repository.Flags
+    ( UseCache (..)
+    , DryRun (..)
+    , SetDefault (..)
+    , InheritDefault (..)
+    , RemoteRepos (..)
+    )
 import Darcs.Util.Lock( readTextFile, writeTextFile )
 import Darcs.Util.Exception ( catchall )
-import Darcs.Util.Global ( darcsdir )
+import Darcs.Util.Global ( darcsdir, debugMessage )
 import Darcs.Util.Path ( AbsolutePath, ioAbsolute, toFilePath,
                          getCurrentDirectory )
 import Darcs.Util.Printer( hPutDocLn, text )
@@ -186,6 +192,8 @@ defaultBoring = map ("# " ++) boringFileInternalHelp ++
     , "(^|/)\\.DS_Store$"
     , "# emacs saved sessions (desktops)"
     , "(^|.*/)\\.emacs\\.desktop(\\.lock)?$"
+    , " # stack"
+    , "(^|/)\\.stack-work($|/)"
     ]
 
 boringFileInternalHelp :: [String]
@@ -201,17 +209,6 @@ boringFileInternalHelp =
     , "See regex(7) for a description of extended regular expressions."
     ]
 
-darcsdirFilter :: [FilePath] -> [FilePath]
-darcsdirFilter = filter (not . isDarcsdir)
-
-isDarcsdir :: FilePath -> Bool
-isDarcsdir ('.' : '/' : f) = isDarcsdir f
-isDarcsdir "." = True
-isDarcsdir "" = True
-isDarcsdir ".." = True
-isDarcsdir "../" = True
-isDarcsdir fp = (darcsdir ++ "/") `isPrefixOf` fp || fp == darcsdir
-
 -- | The path of the global preference directory; @~/.darcs@ on Unix,
 -- and @%APPDATA%/darcs@ on Windows.
 globalPrefsDir :: IO (Maybe FilePath)
@@ -272,12 +269,10 @@ boringRegexps = do
     globalBores <- getGlobal "boring"
     liftM catMaybes $ mapM tryMakeBoringRegexp $ localBores ++ globalBores
 
-boringFileFilter :: IO ([FilePath] -> [FilePath])
-boringFileFilter = filterBoringAndDarcsdir `fmap` boringRegexps
-  where
-    filterBoringAndDarcsdir regexps = filter (notBoring regexps . doNormalise)
-    notBoring regexps file = not $
-        isDarcsdir file || any (\r -> isJust $ matchRegex r file) regexps
+isBoring :: IO (FilePath -> Bool)
+isBoring = do
+  regexps <- boringRegexps
+  return $ \file -> any (\r -> isJust $ matchRegex r file) regexps
 
 noncomments :: [String] -> [String]
 noncomments = filter nonComment
@@ -439,12 +434,12 @@ fixRepoPath p
 defaultrepo :: RemoteRepos -> AbsolutePath -> [String] -> IO [String]
 defaultrepo (RemoteRepos rrepos) _ [] =
   do case rrepos of
-       [] -> maybeToList `fmap` getDefaultRepoPath
+       [] -> maybeToList `fmap` getDefaultRepo
        rs -> mapM fixRepoPath rs
 defaultrepo _ _ r = return r
 
-getDefaultRepoPath :: IO (Maybe String)
-getDefaultRepoPath = do
+getDefaultRepo :: IO (Maybe String)
+getDefaultRepo = do
     defaults <- getPreflist defaultRepoPref
     case defaults of
          [] -> return Nothing
@@ -456,19 +451,26 @@ defaultRepoPref = "defaultrepo"
 -- | addRepoSource adds a new entry to _darcs/prefs/repos and sets it as default
 --   in _darcs/prefs/defaultrepo, unless --no-set-default or --dry-run is passed,
 --   or it is the same repository as the current one.
-addRepoSource :: String -> DryRun -> RemoteRepos -> SetDefault -> IO ()
-addRepoSource r isDryRun (RemoteRepos rrepos) setDefault = (do
+addRepoSource :: String
+              -> DryRun
+              -> RemoteRepos
+              -> SetDefault
+              -> InheritDefault
+              -> Bool
+              -> IO ()
+addRepoSource r isDryRun (RemoteRepos rrepos) setDefault inheritDefault isInteractive = (do
     olddef <- getPreflist defaultRepoPref
+    newdef <- newDefaultRepo
     let shouldDoIt = null noSetDefault && greenLight
-        greenLight = shouldAct && not rIsTmp && (olddef /= [r] || olddef == [])
+        greenLight = shouldAct && not rIsTmp && (olddef /= [newdef] || olddef == [])
     -- the nuance here is that we should only notify when the reason we're not
     -- setting default is the --no-set-default flag, not the various automatic
     -- show stoppers
     if shouldDoIt
-       then setPreflist defaultRepoPref [r]
-       else when (True `notElem` noSetDefault && greenLight) $
+       then setPreflist defaultRepoPref [newdef]
+       else when (True `notElem` noSetDefault && greenLight && inheritDefault == NoInheritDefault) $
                 putStr . unlines $ setDefaultMsg
-    addToPreflist "repos" r) `catchall` return ()
+    addToPreflist "repos" newdef) `catchall` return ()
   where
     shouldAct = isDryRun == NoDryRun
     rIsTmp = r `elem` rrepos
@@ -476,11 +478,39 @@ addRepoSource r isDryRun (RemoteRepos rrepos) setDefault = (do
                        NoSetDefault x -> [x]
                        _ -> []
     setDefaultMsg =
-        [ "HINT: if you want to change the default remote repository to"
+        [ "By the way, to change the default remote repository to"
         , "      " ++ r ++ ","
-        , "      quit now and issue the same command with the --set-default "
-          ++ "flag."
+        , "you can " ++
+          (if isInteractive then "quit now and " else "") ++
+          "issue the same command with the --set-default flag."
         ]
+    newDefaultRepo :: IO String
+    newDefaultRepo = case inheritDefault of
+      YesInheritDefault -> getRemoteDefaultRepo
+      NoInheritDefault -> return r
+    -- TODO It would be nice if --inherit-default could be made to work with
+    -- arbitrary remote repos; for security reasons we currently allow only
+    -- repos on the same host which must also be owned by ourselves. This is
+    -- because the defaultrepo file is read and written as a text file, and
+    -- therefore encoded in the user's locale encoding. See
+    -- http://bugs.darcs.net/issue2627 for a more detailed discussion.
+    getRemoteDefaultRepo
+      | isValidLocalPath r = do
+          sameOwner r "." >>= \case
+            True -> do
+              defs <-
+                getPreffile (r </> darcsdir </> "prefs/defaultrepo")
+                `catchIOError`
+                const (return [r])
+              case defs of
+                defrepo:_ -> do
+                  debugMessage "using defaultrepo of remote"
+                  return defrepo
+                [] -> return r
+            False -> return r
+      | otherwise = return r
+    sameOwner p q =
+      (==) <$> (fileOwner <$> getFileStatus p) <*> (fileOwner <$> getFileStatus q)
 
 -- | delete references to other repositories.
 --   Used when cloning to a ssh destination.
@@ -509,7 +539,7 @@ getCaches useCache repodir = do
         thatrepo = [Cache Repo NotWritable repodir]
         tempCache = nub $ thisrepo ++ globalcache ++ globalsources ++ here
                           ++ thatrepo ++ filterExternalSources there
-    return $ Ca $ sortBy compareByLocality tempCache
+    return $ mkCache tempCache
   where
     sourcesFile = darcsdir ++ "/prefs/sources"
 
@@ -553,6 +583,9 @@ prefsDir = "prefs"
 prefsDirPath :: FilePath
 prefsDirPath = darcsdir </> prefsDir
 
+prefsFilePath :: FilePath
+prefsFilePath = prefsDirPath </> "prefs"
+
 prefsFilesHelp :: [(String,String)]
 prefsFilesHelp  =
     [ ("motd", unlines
@@ -573,7 +606,7 @@ prefsFilesHelp  =
       , "e.g. `David Roundy <droundy@abridgegame.org>`. This file overrides the"
       , "contents of the environment variables `$DARCS_EMAIL` and `$EMAIL`."])
     , ("defaults", unlines
-      [ "Default values for darcs commands. Each line of this file has the"
+      [ "Default options for darcs commands. Each line of this file has the"
       , "following form:"
       , ""
       , "    COMMAND FLAG VALUE"
@@ -581,12 +614,15 @@ prefsFilesHelp  =
       , "where `COMMAND` is either the name of the command to which the default"
       , "applies, or `ALL` to indicate that the default applies to all commands"
       , "accepting that flag. The `FLAG` term is the name of the long argument"
-      , "option without the `--`, i.e. `verbose` rather than `--verbose`."
+      , "option with or without the `--`, i.e. `verbose` or `--verbose`."
       , "Finally, the `VALUE` option can be omitted if the flag does not involve"
       , "a value. If the value has spaces in it, use single quotes, not double"
       , "quotes, to surround it. Each line only takes one flag. To set multiple"
       , "defaults for the same command (or for `ALL` commands), use multiple lines."
       , ""
+      , "Options listed in the defaults file are just that: defaults. You can"
+      , "override any default on the command line."
+      , ""
       , "Note that the use of `ALL` easily can have unpredicted consequences,"
       , "especially if commands in newer versions of darcs accepts flags that"
       , "they did not in previous versions. Only use safe flags with `ALL`."
@@ -604,38 +640,10 @@ prefsFilesHelp  =
       , ""
       , "    amend disable"
       , ""
-      , "Also, a global preferences file can be created with the name"
-      , "`.darcs/defaults` in your home directory. Options present there will be"
-      , "added to the repository-specific preferences if they do not conflict."])
-    , ("sources", unlines
-      [ "The `_darcs/prefs/sources` file is used to indicate alternative locations"
-      , "from which to download patches. This file contains lines such as:"
-      , ""
-      , "    cache:/home/droundy/.cache/darcs"
-      , "    readonly:/home/otheruser/.cache/darcs"
-      , "    repo:http://darcs.net"
-      , ""
-      , "This would indicate that darcs should first look in"
-      , "`/home/droundy/.cache/darcs` for patches that might be missing, and if"
-      , "the patch is not there, it should save a copy there for future use."
-      , "In that case, darcs will look in `/home/otheruser/.cache/darcs` to see if"
-      , "that user might have downloaded a copy, but will not try to save a copy"
-      , "there, of course. Finally, it will look in `http://darcs.net`. Note that"
-      , "the `sources` file can also exist in `~/.darcs/`. Also note that the"
-      , "sources mentioned in your `sources` file will be tried *before* the"
-      , "repository you are pulling from. This can be useful in avoiding"
-      , "downloading patches multiple times when you pull from a remote"
-      , "repository to more than one local repository."
-      , ""
-      , "A global cache is enabled by default in your home directory. The cache"
-      , "allows darcs to avoid re-downloading patches (for example, when doing a"
-      , "second darcs clone of the same repository), and also allows darcs to use"
-      , "hard links to reduce disk usage."
-      , ""
-      , "Note that the cache directory should reside on the same filesystem as"
-      , "your repositories, so you may need to vary this. You can also use"
-      , "multiple cache directories on different filesystems, if you have several"
-      , "filesystems on which you use darcs."])
+      , "A global defaults file can be created with the name"
+      , "`.darcs/defaults` in your home directory. In case of conflicts,"
+      , "the defaults for a specific repository take precedence."
+      ])
     , ("boring", unlines
       [ "The `_darcs/prefs/boring` file may contain a list of regular expressions"
       , "describing files, such as object files, that you do not expect to add to"
@@ -675,6 +683,46 @@ prefsFilesHelp  =
       [ "Contains the URL of the default remote repository used by commands `pull`,"
       , "`push`, `send` and `optimize relink`. Darcs edits this file automatically"
       , "or when the flag `--set-default` is used."])
+    , ("sources", unlines
+      [ "Besides the defaultrepo, darcs also keeps track of any other locations"
+      , "used in commands for exchanging patches (e.g. push, pull, send)."
+      , "These are subsequently used as alternatives from which to download"
+      , "patches. The file contains lines such as:"
+      , ""
+      , "    cache:/home/droundy/.cache/darcs"
+      , "    readonly:/home/otheruser/.cache/darcs"
+      , "    repo:http://darcs.net"
+      , ""
+      , "The prefix `cache:` indicates that darcs can use this as a read-write"
+      , "cache for patches, `read-only:` indicates a cache that is only"
+      , "readable, and `repo:` denotes a (possibly remote) repository. The order"
+      , "of the entries is immaterial: darcs will always try local paths before"
+      , "remote ones, and only local ones will be used as potentially writable."
+      , ""
+      , "A global cache is enabled by default in your home directory under"
+      , "`.cache/darcs` (older versions of darcs used `.darcs/cache` for this),"
+      , "or `$XDG_CACHE_HOME/darcs` if the environment variable is set, see"
+      , "https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html."
+      , "The cache allows darcs to avoid re-downloading patches (for example, when"
+      , "doing a second darcs clone of the same repository), and also allows darcs"
+      , "to use hard links to reduce disk usage."
+      , ""
+      , "Note that the cache directory should reside on the same filesystem as"
+      , "your repositories, so you may need to vary this. You can also use"
+      , "multiple cache directories on different filesystems, if you have several"
+      , "filesystems on which you use darcs."
+      , ""
+      , "While darcs automatically adds entries to `_darcs/prefs/sources`, it does"
+      , "not currently remove them. If one or more of the entries aren't accessible"
+      , "(e.g. because they resided on a removable media), then darcs will bugger"
+      , "you with a hint, suggesting you remove those entries. This is done because"
+      , "certain systems have extremely long timeouts associated with some remotely"
+      , "accessible media (e.g. NFS over automounter on Linux), which can slow down"
+      , "darcs operations considerably. On the other hand, when you clone a repo"
+      , "with --lazy from a no longer accessible location, then the hint may give"
+      , "you an idea where the patches could be found, so you can try to restore"
+      , "access to them."
+      ])
     , ("tmpdir", unlines
       [ "By default temporary directories are created in `/tmp`, or if that doesn't"
       , "exist, in `_darcs` (within the current repo).  This can be overridden by"
