@@ -3,7 +3,6 @@ module Darcs.Repository.Repair ( replayRepository, checkIndex,
                                  RepositoryConsistency(..) )
        where
 
-import Prelude ()
 import Darcs.Prelude
 
 import Control.Monad ( when, unless )
@@ -11,11 +10,23 @@ import Control.Monad.Trans ( liftIO )
 import Control.Exception ( catch, finally, IOException )
 import Data.Maybe ( catMaybes )
 import Data.List ( sort, (\\) )
-import System.Directory ( createDirectoryIfMissing, getCurrentDirectory,
-                          setCurrentDirectory )
+import System.Directory
+    ( createDirectoryIfMissing
+    , getCurrentDirectory
+    , removeDirectoryRecursive
+    , setCurrentDirectory
+    )
 import System.FilePath ( (</>) )
 import Darcs.Util.Path( anchorPath, AbsolutePath, ioAbsolute, toFilePath )
-import Darcs.Patch.PatchInfoAnd ( hopefully, PatchInfoAnd, info, winfo, WPatchInfo, unWPatchInfo, compareWPatchInfo )
+import Darcs.Patch.PatchInfoAnd
+    ( PatchInfoAnd
+    , WPatchInfo
+    , compareWPatchInfo
+    , hopefully
+    , info
+    , unWPatchInfo
+    , winfo
+    )
 
 import Darcs.Patch.Witnesses.Eq ( EqCheck(..) )
 import Darcs.Patch.Witnesses.Ordered
@@ -28,30 +39,37 @@ import Darcs.Patch.Info ( displayPatchInfo )
 import Darcs.Patch.Set ( Origin, PatchSet(..), patchSet2FL, patchSet2RL )
 import Darcs.Patch ( RepoPatch, IsRepoType, PrimOf, isInconsistent )
 
-import Darcs.Repository.Flags
-    ( Verbosity(..), Compression, DiffAlgorithm )
-import Darcs.Repository.Format ( identifyRepoFormat,
-                                 RepoProperty ( HashedInventory ), formatHas )
 import Darcs.Repository.Cache ( HashedDir( HashedPristineDir ) )
+import Darcs.Repository.Diff( treeDiff )
+import Darcs.Repository.Flags ( Verbosity(..), Compression, DiffAlgorithm )
+import Darcs.Repository.Format
+    ( identifyRepoFormat
+    , RepoProperty ( HashedInventory )
+    , formatHas
+    )
 import Darcs.Repository.HashedIO ( cleanHashdir )
-import Darcs.Repository.Hashed ( readHashedPristineRoot, writeAndReadPatch )
+import Darcs.Repository.Hashed ( readRepo, writeAndReadPatch )
 import Darcs.Repository.InternalTypes ( Repository, repoCache, repoLocation )
 import Darcs.Repository.Prefs ( filetypeFunction )
-import Darcs.Repository.Hashed ( readRepo )
+import Darcs.Repository.Pristine ( readHashedPristineRoot )
 import Darcs.Repository.State
     ( readRecorded
     , readIndex
     , readRecordedAndPending
     )
-import Darcs.Repository.Diff( treeDiff )
 
-import Darcs.Util.Progress ( debugMessage, beginTedious, endTedious, tediousSize, finishedOneIO )
+import Darcs.Util.Progress
+    ( beginTedious
+    , debugMessage
+    , endTedious
+    , finishedOneIO
+    , tediousSize
+    )
 import Darcs.Util.File ( withCurrentDirectory )
 import Darcs.Util.Exception ( catchall )
 import Darcs.Util.Global ( darcsdir )
-import Darcs.Util.Lock( rmRecursive, withTempDir )
-import Darcs.Util.Printer ( Doc, putDocLn, text )
-import Darcs.Util.Printer.Color ( showDoc )
+import Darcs.Util.Lock( withDelayedDir )
+import Darcs.Util.Printer ( Doc, putDocLn, text, renderString )
 
 import Darcs.Util.Hash( Hash(NoHash), encodeBase16 )
 import Darcs.Util.Tree( Tree, emptyTree, list, restrict, expand, itemHash, zipTrees )
@@ -66,14 +84,17 @@ replaceInFL :: FL (PatchInfoAnd rt a) wX wY
             -> [Sealed2 (WPatchInfo :||: PatchInfoAnd rt a)]
             -> FL (PatchInfoAnd rt a) wX wY
 replaceInFL orig [] = orig
-replaceInFL NilFL _ = impossible
+replaceInFL NilFL _ = error "impossible case"
 replaceInFL (o:>:orig) ch@(Sealed2 (o':||:c):ch_rest)
     | IsEq <- winfo o `compareWPatchInfo` o' = c:>:replaceInFL orig ch_rest
     | otherwise = o:>:replaceInFL orig ch
 
-applyAndFix :: forall rt p wR wU wT. (IsRepoType rt, RepoPatch p, ApplyState p ~ Tree)
-            => Repository rt p wR wU wT -> Compression -> FL (PatchInfoAnd rt p) Origin wR
-            -> TreeIO (FL (PatchInfoAnd rt p) Origin wR, Bool)
+applyAndFix
+  :: forall rt p wR wU wT. (IsRepoType rt, RepoPatch p, ApplyState p ~ Tree)
+  => Repository rt p wR wU wT
+  -> Compression
+  -> FL (PatchInfoAnd rt p) Origin wR
+  -> TreeIO (FL (PatchInfoAnd rt p) Origin wR, Bool)
 applyAndFix _ _ NilFL = return (NilFL, True)
 applyAndFix r compr psin =
     do liftIO $ beginTedious k
@@ -83,7 +104,8 @@ applyAndFix r compr psin =
        orig <- liftIO $ patchSet2FL `fmap` readRepo r
        return (replaceInFL orig repaired, ok)
     where k = "Replaying patch"
-          aaf :: FL (PatchInfoAnd rt p) wW wZ -> TreeIO ([Sealed2 (WPatchInfo :||: PatchInfoAnd rt p)], Bool)
+          aaf :: FL (PatchInfoAnd rt p) wW wZ
+              -> TreeIO ([Sealed2 (WPatchInfo :||: PatchInfoAnd rt p)], Bool)
           aaf NilFL = return ([], True)
           aaf (p:>:ps) = do
             mp' <- applyAndTryToFix p
@@ -91,7 +113,8 @@ applyAndFix r compr psin =
               Just err -> liftIO $ putDocLn err
               Nothing -> return ()
             let !winfp = winfo p -- assure that 'p' can be garbage collected.
-            liftIO $ finishedOneIO k $ showDoc $ displayPatchInfo $ unWPatchInfo winfp
+            liftIO $ finishedOneIO k $ renderString $
+              displayPatchInfo $ unWPatchInfo winfp
             (ps', restok) <- aaf ps
             case mp' of
               Nothing -> return (ps', restok)
@@ -123,9 +146,15 @@ hasDuplicate li = hd $ sort li
           hd [] = Nothing
           hd (x1:x2:xs) | x1 == x2 = Just x1
                         | otherwise = hd (x2:xs)
-replayRepository' ::
-    forall rt p wR wU wT . (IsRepoType rt, RepoPatch p, ApplyState p ~ Tree)
-               => DiffAlgorithm -> AbsolutePath -> Repository rt p wR wU wT -> Compression -> Verbosity -> IO (RepositoryConsistency rt p wR)
+
+replayRepository'
+  :: forall rt p wR wU wT. (IsRepoType rt, RepoPatch p, ApplyState p ~ Tree)
+  => DiffAlgorithm
+  -> AbsolutePath
+  -> Repository rt p wR wU wT
+  -> Compression
+  -> Verbosity
+  -> IO (RepositoryConsistency rt p wR)
 replayRepository' dflag whereToReplay' repo compr verbosity = do
   let whereToReplay = toFilePath whereToReplay'
       putVerbose s = when (verbosity == Verbose) $ putDocLn s
@@ -133,7 +162,10 @@ replayRepository' dflag whereToReplay' repo compr verbosity = do
   checkUniqueness putVerbose putInfo repo
   createDirectoryIfMissing False whereToReplay
   putVerbose $ text "Reading recorded state..."
-  pris <- readRecorded repo `catch` \(_ :: IOException) -> return emptyTree
+  pris <-
+    (readRecorded repo >>= expand >>= darcsUpdateHashes)
+    `catch`
+    \(_ :: IOException) -> return emptyTree
   putVerbose $ text "Applying patches..."
   patches <- readRepo repo
   debugMessage "Fixing any broken patches..."
@@ -146,7 +178,8 @@ replayRepository' dflag whereToReplay' repo compr verbosity = do
 
   debugMessage "Checking pristine against slurpy"
   ftf <- filetypeFunction
-  is_same <- do Sealed diff <- unFreeLeft `fmap` treeDiff dflag ftf pris newpris :: IO (Sealed (FL (PrimOf p) wR))
+  is_same <- do Sealed diff <- unFreeLeft `fmap` treeDiff dflag ftf pris newpris
+                  :: IO (Sealed (FL (PrimOf p) wR))
                 return $ nullFL diff
               `catchall` return False
   -- TODO is the latter condition needed? Does a broken patch imply pristine
@@ -162,23 +195,40 @@ cleanupRepositoryReplay r = do
   let c = repoCache r
   rf <- identifyRepoFormat "."
   unless (formatHas HashedInventory rf) $
-         rmRecursive $ darcsdir ++ "/pristine.hashed"
+         removeDirectoryRecursive $ darcsdir ++ "/pristine.hashed"
   when (formatHas HashedInventory rf) $ do
        current <- readHashedPristineRoot r
        cleanHashdir c HashedPristineDir $ catMaybes [current]
 
-replayRepositoryInTemp :: (IsRepoType rt, RepoPatch p, ApplyState p ~ Tree)
-                       => DiffAlgorithm -> Repository rt p wR wU wT -> Compression -> Verbosity
-                          -> IO (RepositoryConsistency rt p wR)
+replayRepositoryInTemp
+  :: (IsRepoType rt, RepoPatch p, ApplyState p ~ Tree)
+  => DiffAlgorithm
+  -> Repository rt p wR wU wT
+  -> Compression
+  -> Verbosity
+  -> IO (RepositoryConsistency rt p wR)
 replayRepositoryInTemp dflag r compr verb = do
   repodir <- getCurrentDirectory
-  withTempDir "darcs-check" $ \tmpDir -> do
+  {- The reason we use withDelayedDir here, instead of withTempDir, is that
+  replayRepository' may return a new pristine that is read from the 
+  temporary location and reading a Tree is done using lazy ByteStrings (for
+  file contents). Then we check if there is a difference to our stored
+  pristine, but when there are differences the check may terminate early
+  and not all of the new pristine was read/evaluated. This may then cause
+  does-not-exist-failures later on when the tree is evaluated further.
+  -}
+  withDelayedDir "darcs-check" $ \tmpDir -> do
     setCurrentDirectory repodir
     replayRepository' dflag tmpDir r compr verb
 
-replayRepository :: (IsRepoType rt, RepoPatch p, ApplyState p ~ Tree)
-                 => DiffAlgorithm -> Repository rt p wR wU wT -> Compression -> Verbosity
-                 -> (RepositoryConsistency rt p wR -> IO a) -> IO a
+replayRepository
+  :: (IsRepoType rt, RepoPatch p, ApplyState p ~ Tree)
+  => DiffAlgorithm
+  -> Repository rt p wR wU wT
+  -> Compression
+  -> Verbosity
+  -> (RepositoryConsistency rt p wR -> IO a)
+  -> IO a
 replayRepository dflag r compr verb f =
   run `finally` cleanupRepositoryReplay r
     where run = do
@@ -187,7 +237,11 @@ replayRepository dflag r compr verb f =
             st <- replayRepository' dflag hashedPristine r compr verb
             f st
 
-checkIndex :: (RepoPatch p, ApplyState p ~ Tree) => Repository rt p wR wU wT -> Bool -> IO Bool
+checkIndex
+  :: (RepoPatch p, ApplyState p ~ Tree)
+  => Repository rt p wR wU wR
+  -> Bool
+  -> IO Bool
 checkIndex repo quiet = do
   index <- updateIndex =<< readIndex repo
   pristine <- expand =<< readRecordedAndPending repo
@@ -201,7 +255,8 @@ checkIndex repo quiet = do
       gethashes p (Just i1) Nothing   = (p, itemHash i1, NoHash)
       gethashes p   Nothing (Just i2) = (p,      NoHash, itemHash i2)
       gethashes p   Nothing Nothing   = error $ "Bad case at " ++ show p
-      mismatches = [ miss | miss@(_, h1, h2) <- zipTrees gethashes index working_hashed, h1 /= h2 ]
+      mismatches =
+        [miss | miss@(_, h1, h2) <- zipTrees gethashes index working_hashed, h1 /= h2]
 
       format paths = unlines $ map (("  " ++) . anchorPath "") paths
       mismatches_disp = unlines [ anchorPath "" p ++
