@@ -24,9 +24,8 @@
 
 module Darcs.Repository.State
     ( restrictSubpaths, restrictBoring, TreeFilter(..), restrictDarcsdir
-    , maybeRestrictSubpaths
     -- * Diffs
-    , unrecordedChanges, readPending
+    , unrecordedChanges
     -- * Trees
     , readRecorded, readUnrecorded, readRecordedAndPending, readWorking
     , readPendingAndWorking, readUnrecordedFiltered
@@ -38,13 +37,12 @@ module Darcs.Repository.State
     , addPendingDiffToPending, addToPending
     ) where
 
-import Prelude ()
 import Darcs.Prelude
 
 import Control.Monad ( when, foldM, forM )
 import Control.Monad.State ( StateT, runStateT, get, put, liftIO )
 import Control.Exception ( catch, IOException )
-import Data.Maybe ( fromJust, isJust )
+import Data.Maybe ( isJust )
 import Data.Ord ( comparing )
 import Data.List ( sortBy, union, delete )
 import Text.Regex( matchRegex )
@@ -56,99 +54,115 @@ import System.FilePath
     , (<.>)
 #endif
     )
+import System.IO ( hPutStrLn, stderr )
+import System.IO.Error ( catchIOError )
+
 import qualified Data.ByteString as B
-    ( ByteString, readFile, drop, writeFile, empty, concat )
+    ( ByteString, readFile, writeFile, empty, concat )
 import qualified Data.ByteString.Char8 as BC
-    ( pack, unpack, split )
+    ( pack, unpack )
 import qualified Data.ByteString.Lazy as BL ( toChunks )
 
-import Darcs.Patch ( RepoPatch, PrimOf, sortCoalesceFL, fromPrims
+import Darcs.Patch ( RepoPatch, PrimOf, sortCoalesceFL
                    , PrimPatch, maybeApplyToTree
                    , tokreplace, forceTokReplace, move )
-import Darcs.Patch.Named.Wrapped ( anonymous )
-import Darcs.Patch.Apply ( ApplyState, applyToTree, effectOnFilePaths )
-import Darcs.Patch.Witnesses.Ordered ( RL(..), FL(..), (+>+)
+import Darcs.Patch.Named ( anonymous )
+import Darcs.Patch.Apply ( ApplyState, applyToTree, effectOnPaths )
+import Darcs.Patch.Witnesses.Ordered ( FL(..), (+>+)
                                      , (:>)(..), reverseRL, reverseFL
                                      , mapFL, concatFL, toFL, nullFL )
 import Darcs.Patch.Witnesses.Eq ( EqCheck(IsEq, NotEq) )
 import Darcs.Patch.Witnesses.Unsafe ( unsafeCoerceP, unsafeCoercePEnd )
 import Darcs.Patch.Witnesses.Sealed ( Sealed(Sealed), seal, unFreeLeft, mapSeal
                                     , freeGap, emptyGap, joinGap, FreeLeft, Gap(..) )
-import Darcs.Patch.Commute ( selfCommuter, commuteFL )
-import Darcs.Patch.CommuteFn ( commuterIdRL )
+import Darcs.Patch.Commute ( commuteFL )
 import Darcs.Patch.Permutations ( partitionConflictingFL, genCommuteWhatWeCanRL )
 import Darcs.Patch.PatchInfoAnd ( PatchInfoAnd, n2pia )
 import Darcs.Patch.FileHunk ( FileHunk(..), IsHunk(..) )
 import Darcs.Patch.TokenReplace ( breakToTokens, defaultToks )
 
 import Darcs.Repository.Flags ( UseIndex(..), ScanKnown(..), DiffAlgorithm(..)
-                              , UpdateWorking(..), LookForMoves(..), LookForReplaces(..) )
-import Darcs.Util.Global ( darcsdir )
+                              , UpdatePending(..), LookForMoves(..), LookForReplaces(..) )
 
-import Darcs.Repository.InternalTypes ( Repository, repoFormat )
+import Darcs.Repository.InternalTypes ( Repository, repoFormat, repoLocation )
 import Darcs.Repository.Format(formatHas, RepoProperty(NoWorkingDir))
 import qualified Darcs.Repository.Pending as Pending
 import Darcs.Repository.Prefs ( filetypeFunction, boringRegexps )
 import Darcs.Repository.Diff ( treeDiff )
+import Darcs.Repository.Inventory ( peekPristineHash, getValidHash )
+import Darcs.Repository.Paths
+    ( pristineDirPath
+    , hashedInventoryPath
+    , oldPristineDirPath
+    , oldCurrentDirPath
+    , patchesDirPath
+    , indexPath
+    , indexInvalidPath
+    )
 
 import Darcs.Util.Path
-    ( AnchoredPath(..), anchorPath, floatPath, fn2fp
-    , SubPath, sp2fn, filterPaths, FileName
-    , parents, replacePrefixPath, anchoredRoot
-    , toFilePath, simpleSubPath, normPath, floatSubPath, makeName
+    ( AnchoredPath
+    , anchorPath
+    , filterPaths
+    , inDarcsdir
+    , parents
+    , movedirfilename
     )
 import Darcs.Util.Hash( Hash( NoHash ) )
 import Darcs.Util.Tree( Tree, restrict, FilterTree, expand, emptyTree, overlay, find
                       , ItemType(..), itemType, readBlob, modifyTree, findFile, TreeItem(..)
                       , makeBlobBS, expandPath )
-import Darcs.Util.Tree.Plain( readPlainTree )
-import Darcs.Util.Tree.Hashed( darcsTreeHash, readDarcsHashed, decodeDarcsHash, decodeDarcsSize )
+import qualified Darcs.Util.Tree.Plain as PlainTree ( readPlainTree )
+import Darcs.Util.Tree.Hashed
+    ( darcsTreeHash, readDarcsHashed, decodeDarcsHash, decodeDarcsSize )
 import qualified Darcs.Util.Index as I
 import qualified Darcs.Util.Tree as Tree
 import Darcs.Util.Index ( listFileIDs, getFileID )
 
+#define TEST_INDEX 0
+
+#if TEST_INDEX
+import Control.Monad ( unless )
+import Darcs.Util.Path ( displayPath )
+import Darcs.Util.Tree ( list )
+#endif
+
 newtype TreeFilter m = TreeFilter { applyTreeFilter :: forall tr . FilterTree tr m => tr m -> tr m }
 
--- | From a repository and a list of SubPath's, construct a filter that can be
+-- | From a repository and a list of AnchoredPath's, construct a filter that can be
 -- used on a Tree (recorded or unrecorded state) of this repository. This
 -- constructed filter will take pending into account, so the subpaths will be
 -- translated correctly relative to pending move patches.
 restrictSubpaths :: (RepoPatch p, ApplyState p ~ Tree)
-                 => Repository rt p wR wU wT -> [SubPath]
+                 => Repository rt p wR wU wT -> [AnchoredPath]
                  -> IO (TreeFilter m)
-restrictSubpaths repo subpaths = do
+restrictSubpaths repo paths = do
   Sealed pending <- Pending.readPending repo
-  restrictSubpathsAfter pending repo subpaths
+  restrictSubpathsAfter pending repo paths
 
 -- | Like 'restrictSubpaths' but with the pending patch passed as a parameter.
 -- The 'Repository' parameter is not used, we need it only to avoid
 -- abiguous typing of @p@.
 restrictSubpathsAfter :: (RepoPatch p, ApplyState p ~ Tree)
-                      => FL (PrimOf p) wT wP
+                      => FL (PrimOf p) wR wP
                       -> Repository rt p wR wU wT
-                      -> [SubPath]
+                      -> [AnchoredPath]
                       -> IO (TreeFilter m)
-restrictSubpathsAfter pending _repo subpaths = do
-  let paths = map (fn2fp . sp2fn) subpaths
-      paths' = paths `union` effectOnFilePaths pending paths
-      anchored = map floatPath paths'
+restrictSubpathsAfter pending _repo paths = do
+  let paths' = paths `union` effectOnPaths pending paths
       restrictPaths :: FilterTree tree m => tree m -> tree m
-      restrictPaths = Tree.filter (filterPaths anchored)
+      restrictPaths = Tree.filter (filterPaths paths')
   return (TreeFilter restrictPaths)
 
+-- note we assume pending starts at the recorded state
 maybeRestrictSubpaths :: (RepoPatch p, ApplyState p ~ Tree)
-                      => FL (PrimOf p) wT wP
+                      => FL (PrimOf p) wR wP
                       -> Repository rt p wR wU wT
-                      -> Maybe [SubPath]
+                      -> Maybe [AnchoredPath]
                       -> IO (TreeFilter m)
 maybeRestrictSubpaths pending repo =
   maybe (return $ TreeFilter id) (restrictSubpathsAfter pending repo)
 
--- |Is the given path in (or equal to) the _darcs metadata directory?
-inDarcsDir :: AnchoredPath -> Bool
-inDarcsDir (AnchoredPath (x:_)) | x == makeName darcsdir = True
-inDarcsDir _ = False
-
 -- | Construct a 'TreeFilter' that removes any boring files that are not also
 -- contained in the argument 'Tree'.
 --
@@ -158,7 +172,7 @@ inDarcsDir _ = False
 restrictBoring :: Tree m -> IO (TreeFilter m)
 restrictBoring guide = do
   boring <- boringRegexps
-  let boring' p | inDarcsDir p = False
+  let boring' p | inDarcsdir p = False
       boring' p = not $ any (\rx -> isJust $ matchRegex rx p') boring
           where p' = anchorPath "" p
       restrictTree :: FilterTree t m => t m -> t m
@@ -170,13 +184,13 @@ restrictBoring guide = do
 -- | Construct a Tree filter that removes any darcs metadata files the
 -- Tree might have contained.
 restrictDarcsdir :: TreeFilter m
-restrictDarcsdir = TreeFilter $ Tree.filter $ \p _ -> not (inDarcsDir p)
+restrictDarcsdir = TreeFilter $ Tree.filter $ \p _ -> not (inDarcsdir p)
 
 {- |
 For a repository and an optional list of paths (when 'Nothing', take
 everything) compute a (forward) list of prims (i.e. a patch) going from the
 recorded state of the repository (pristine) to the unrecorded state of the
-repository (the working copy + pending). When a list of paths is given, at
+repository (the working tree + pending). When a list of paths is given, at
 least the files that live under any of these paths in either recorded or
 unrecorded will be included in the resulting patch. NB. More patches may be
 included in this list, eg. the full contents of the pending patch. This is
@@ -211,8 +225,8 @@ unrecordedChanges :: (RepoPatch p, ApplyState p ~ Tree)
                   => (UseIndex, ScanKnown, DiffAlgorithm)
                   -> LookForMoves
                   -> LookForReplaces
-                  -> Repository rt p wR wU wT
-                  -> Maybe [SubPath] -> IO (FL (PrimOf p) wT wU)
+                  -> Repository rt p wR wU wR
+                  -> Maybe [AnchoredPath] -> IO (FL (PrimOf p) wR wU)
 unrecordedChanges dopts lfm lfr r paths = do
   (pending :> working) <- readPendingAndWorking dopts lfm lfr r paths
   return $ sortCoalesceFL (pending +>+ working)
@@ -220,18 +234,18 @@ unrecordedChanges dopts lfm lfr r paths = do
 -- Implementation note: it is important to do things in the right order: we
 -- first have to read the pending patch, then detect moves, then detect adds,
 -- then detect replaces.
-readPendingAndWorking :: forall rt p wR wU wT. (RepoPatch p, ApplyState p ~ Tree)
+readPendingAndWorking :: (RepoPatch p, ApplyState p ~ Tree)
                       => (UseIndex, ScanKnown, DiffAlgorithm)
                       -> LookForMoves
                       -> LookForReplaces
-                      -> Repository rt p wR wU wT
-                      -> Maybe [SubPath]
-                      -> IO ((FL (PrimOf p) :> FL (PrimOf p)) wT wU)
+                      -> Repository rt p wR wU wR
+                      -> Maybe [AnchoredPath]
+                      -> IO ((FL (PrimOf p) :> FL (PrimOf p)) wR wU)
 readPendingAndWorking _ _ _ r _ | formatHas NoWorkingDir (repoFormat r) = do
   IsEq <- return $ workDirLessRepoWitness r
   return (NilFL :> NilFL)
 readPendingAndWorking (useidx, scan, diffalg) lfm lfr repo mbpaths = do
-  (pending_tree, working_tree, pending) <-
+  (pending_tree, working_tree, (pending :> moves)) <-
     readPendingAndMovesAndUnrecorded repo useidx scan lfm mbpaths
   (pending_tree_with_replaces, Sealed replaces) <-
     getReplaces lfr diffalg repo pending_tree working_tree
@@ -239,18 +253,18 @@ readPendingAndWorking (useidx, scan, diffalg) lfm lfr repo mbpaths = do
   wrapped_diff <- treeDiff diffalg ft pending_tree_with_replaces working_tree
   case unFreeLeft wrapped_diff of
     Sealed diff -> do
-      return (pending +>+ unsafeCoercePEnd replaces :> unsafeCoercePEnd diff)
+      return $ unsafeCoercePEnd $ pending :> (moves +>+ replaces +>+ diff)
 
 readPendingAndMovesAndUnrecorded
   :: (RepoPatch p, ApplyState p ~ Tree)
-  => Repository rt p wR wU wT
+  => Repository rt p wR wU wR
   -> UseIndex
   -> ScanKnown
   -> LookForMoves
-  -> Maybe [SubPath]
+  -> Maybe [AnchoredPath]
   -> IO ( Tree IO             -- pristine with (pending + moves)
         , Tree IO             -- working
-        , FL (PrimOf p) wT wU -- pending + moves
+        , (FL (PrimOf p) :> FL (PrimOf p)) wR wU -- pending :> moves
         )
 readPendingAndMovesAndUnrecorded repo useidx scan lfm mbpaths = do
   (pending_tree, Sealed pending) <- readPending repo
@@ -261,44 +275,47 @@ readPendingAndMovesAndUnrecorded repo useidx scan lfm mbpaths = do
     applyTreeFilter relevant <$> applyToTree moves pending_tree
   let useidx' = if nullFL moves then useidx else IgnoreIndex
   index <-
-    applyToTree moves =<< I.updateIndex =<<
-    applyTreeFilter relevant <$> readIndex repo
-  working_tree <- filteredWorking useidx' scan relevant index pending_tree'
-  return (pending_tree', working_tree, unsafeCoercePEnd pending')
+    applyToTree moves =<< readIndexOrPlainTree repo useidx relevant pending_tree
+  working_tree <- filteredWorking repo useidx' scan relevant index pending_tree'
+  return (pending_tree', working_tree, unsafeCoercePEnd (pending :> moves))
 
 -- | @filteredWorking useidx scan relevant index pending_tree@ reads the
 -- working tree and filters it according to options and @relevant@ file paths.
 -- The @pending_tree@ is understood to have @relevant@ already applied and is
--- used (only) if @useidx == 'IgnoreIndex'@ and @scan == 'ScanKnown'@ to act as
+-- used (only) if @useidx == 'IgnoreIndex'@ and @scan /= 'ScanBoring'@ to act as
 -- a guide for filtering the working tree.
--- Note that even if @useidx '==' 'IgnoreIndex'@, the index is still used
--- to avoid filtering boring files that darcs knows about (see 'restrictBoring').
-filteredWorking :: UseIndex
+filteredWorking :: Repository rt p wR wU wR
+                -> UseIndex
                 -> ScanKnown
                 -> TreeFilter IO
                 -> Tree IO
                 -> Tree IO
                 -> IO (Tree IO)
-filteredWorking useidx scan relevant index pending_tree = do
-  applyTreeFilter restrictDarcsdir <$> case scan of
-    ScanKnown -> case useidx of
-      UseIndex -> return index
-      IgnoreIndex -> do
-        guide <- expand pending_tree
-        applyTreeFilter relevant . restrict guide <$> readPlainTree "."
-    ScanAll -> do
-      nonboring <- restrictBoring index
-      plain <- applyTreeFilter relevant . applyTreeFilter nonboring <$> readPlainTree "."
-      return $ case useidx of
-        UseIndex -> plain `overlay` index
-        IgnoreIndex -> plain
-    ScanBoring -> do
-      plain <- applyTreeFilter relevant <$> readPlainTree "."
-      return $ case useidx of
-        UseIndex -> plain `overlay` index
-        IgnoreIndex -> plain
+filteredWorking repo useidx scan relevant index pending_tree =
+  applyTreeFilter restrictDarcsdir <$> applyTreeFilter relevant <$> do
+    case useidx of
+      UseIndex ->
+        case scan of
+          ScanKnown -> return index
+          ScanAll -> do
+            nonboring <- restrictBoring index
+            plain <- applyTreeFilter nonboring <$> readPlainTree repo
+            return $ plain `overlay` index
+          ScanBoring -> do
+            plain <- readPlainTree repo
+            return $ plain `overlay` index
+      IgnoreIndex ->
+        case scan of
+          ScanKnown -> do
+            guide <- expand pending_tree
+            restrict guide <$> readPlainTree repo
+          ScanAll -> do
+            guide <- expand pending_tree
+            nonboring <- restrictBoring guide
+            applyTreeFilter nonboring <$> readPlainTree repo
+          ScanBoring -> readPlainTree repo
 
--- | Witnesses the fact that in the absence of a working directory, we
+-- | Witnesses the fact that in the absence of a working tree, we
 -- pretend that the working dir updates magically to the tentative state.
 workDirLessRepoWitness :: Repository rt p wR wU wT -> EqCheck wU wT
 workDirLessRepoWitness r
@@ -310,24 +327,20 @@ workDirLessRepoWitness r
 -- applying all the repository's patches to an empty directory.
 readRecorded :: Repository rt p wR wU wT -> IO (Tree IO)
 readRecorded _repo = do
-  let h_inventory = darcsdir </> "hashed_inventory"
-  hashed <- doesFileExist h_inventory
+  hashed <- doesFileExist hashedInventoryPath
   if hashed
-     then do inv <- B.readFile h_inventory
-             let linesInv = BC.split '\n' inv
-             case linesInv of
-               [] -> return emptyTree
-               (pris_line:_) -> do
-                          let hash = decodeDarcsHash $ B.drop 9 pris_line
-                              size = decodeDarcsSize $ B.drop 9 pris_line
-                          when (hash == NoHash) $
-                              fail $ "Bad pristine root: " ++ show pris_line
-                          readDarcsHashed (darcsdir </> "pristine.hashed") (size, hash)
-     else do have_pristine <- doesDirectoryExist $ darcsdir </> "pristine"
-             have_current <- doesDirectoryExist $ darcsdir </> "current"
+     then do inv <- B.readFile hashedInventoryPath
+             let pris = peekPristineHash inv
+                 hash = decodeDarcsHash $ BC.pack $ getValidHash pris
+                 size = decodeDarcsSize $ BC.pack $ getValidHash pris
+             when (hash == NoHash) $
+                 fail $ "Bad pristine root: " ++ getValidHash pris
+             readDarcsHashed pristineDirPath (size, hash)
+     else do have_pristine <- doesDirectoryExist $ oldPristineDirPath
+             have_current <- doesDirectoryExist $ oldCurrentDirPath
              case (have_pristine, have_current) of
-               (True, _) -> readPlainTree $ darcsdir </> "pristine"
-               (False, True) -> readPlainTree $ darcsdir </> "current"
+               (True, _) -> PlainTree.readPlainTree $ oldPristineDirPath
+               (False, True) -> PlainTree.readPlainTree $ oldCurrentDirPath
                (_, _) -> fail "No pristine tree is available!"
 
 -- | Obtains a Tree corresponding to the "unrecorded" state of the repository:
@@ -338,48 +351,98 @@ readRecorded _repo = do
 -- parts of the index do not need to go through an up-to-date check (which
 -- involves a relatively expensive lstat(2) per file.
 readUnrecorded :: (RepoPatch p, ApplyState p ~ Tree)
-               => Repository rt p wR wU wT -> Maybe [SubPath] -> IO (Tree IO)
-readUnrecorded repo mbpaths = do
-  Sealed pending <- Pending.readPending repo
+               => Repository rt p wR wU wR
+               -> UseIndex
+               -> Maybe [AnchoredPath]
+               -> IO (Tree IO)
+readUnrecorded repo useidx mbpaths = do
+#if TEST_INDEX
+  t1 <- expand =<< readUnrecordedFiltered repo useidx ScanKnown NoLookForMoves mbpaths
+  (pending_tree, Sealed pending) <- readPending repo
   relevant <- maybeRestrictSubpaths pending repo mbpaths
-  readIndex repo >>= I.updateIndex . applyTreeFilter relevant
+  t2 <- readIndexOrPlainTree repo useidx relevant pending_tree
+  assertEqualTrees "indirect" t1 "direct" t2
+  return t1
+#else
+  expand =<< readUnrecordedFiltered repo useidx ScanKnown NoLookForMoves mbpaths
+#endif
+
+#if TEST_INDEX
+assertEqualTrees :: String -> Tree m -> String -> Tree m -> IO ()
+assertEqualTrees n1 t1 n2 t2 =
+  unless (t1 `eqTree` t2) $
+    fail $ "Trees are not equal!\n" ++ showTree n1 t1 ++ showTree n2 t2
+
+eqTree :: Tree m -> Tree m -> Bool
+eqTree t1 t2 = map fst (list t1) == map fst (list t2)
+
+showTree :: String -> Tree m -> String
+showTree name tree = unlines (name : map (("  "++) . displayPath . fst) (list tree))
+#endif
+
+readIndexOrPlainTree :: (ApplyState p ~ Tree, RepoPatch p)
+                     => Repository rt p wR wU wR
+                     -> UseIndex
+                     -> TreeFilter IO
+                     -> Tree IO
+                     -> IO (Tree IO)
+#if TEST_INDEX
+readIndexOrPlainTree repo useidx treeFilter pending_tree = do
+  indexTree <-
+    I.updateIndex =<< applyTreeFilter treeFilter <$> readIndex repo
+  plainTree <- do
+    guide <- expand pending_tree
+    expand =<< applyTreeFilter treeFilter . restrict guide <$> readPlainTree repo
+  assertEqualTrees "index tree" indexTree "plain tree" plainTree
+  return $
+    case useidx of
+      UseIndex -> indexTree
+      IgnoreIndex -> plainTree
+#else
+readIndexOrPlainTree repo UseIndex treeFilter pending_tree =
+  (I.updateIndex =<< applyTreeFilter treeFilter <$> readIndex repo)
+    `catchIOError` \e -> do
+      hPutStrLn stderr ("Warning, cannot access the index:\n" ++ show e)
+      readIndexOrPlainTree repo IgnoreIndex treeFilter pending_tree
+readIndexOrPlainTree repo IgnoreIndex treeFilter pending_tree = do
+  guide <- expand pending_tree
+  expand =<< applyTreeFilter treeFilter . restrict guide <$> readPlainTree repo
+#endif
 
 -- | A variant of 'readUnrecorded' that takes the UseIndex and ScanKnown
 -- options into account, similar to 'readPendingAndWorking'. We are only
 -- interested in the resulting tree, not the patch, so the 'DiffAlgorithm' option
 -- is irrelevant.
 readUnrecordedFiltered :: (RepoPatch p, ApplyState p ~ Tree)
-                       => Repository rt p wR wU wT
+                       => Repository rt p wR wU wR
                        -> UseIndex
                        -> ScanKnown
                        -> LookForMoves
-                       -> Maybe [SubPath] -> IO (Tree IO)
+                       -> Maybe [AnchoredPath] -> IO (Tree IO)
 readUnrecordedFiltered repo useidx scan lfm mbpaths = do
   (_, working_tree, _) <-
     readPendingAndMovesAndUnrecorded repo useidx scan lfm mbpaths
   return working_tree
 
--- | Obtains a Tree corresponding to the complete working copy of the
--- repository (modified and non-modified files).
-readWorking :: IO (Tree IO)
-readWorking = expand =<< (applyTreeFilter restrictDarcsdir <$> readPlainTree ".")
+-- | Obtains the relevant (according to the given filter) part of the working tree.
+readWorking :: TreeFilter IO -> IO (Tree IO)
+readWorking relevant =
+  expand =<<
+  (applyTreeFilter relevant . applyTreeFilter restrictDarcsdir <$>
+   PlainTree.readPlainTree ".")
 
 -- | Obtains the recorded 'Tree' with the pending patch applied.
 readRecordedAndPending :: (RepoPatch p, ApplyState p ~ Tree)
-                       => Repository rt p wR wU wT -> IO (Tree IO)
+                       => Repository rt p wR wU wR -> IO (Tree IO)
 readRecordedAndPending repo = fst `fmap` readPending repo
 
 -- | Obtains the recorded 'Tree' with the pending patch applied, plus
 --   the pending patch itself. The pending patch should start at the
 --   recorded state (we even verify that it applies, and degrade to
---   renaming pending and starting afresh if it doesn't), but we've set to
---   say it starts at the tentative state.
---
---   Question (Eric Kow) Is this a bug? Darcs.Repository.Pending.readPending
---   says it is
+--   renaming pending and starting afresh if it doesn't).
 readPending :: (RepoPatch p, ApplyState p ~ Tree)
-            => Repository rt p wR wU wT
-            -> IO (Tree IO, Sealed (FL (PrimOf p) wT))
+            => Repository rt p wR wU wR
+            -> IO (Tree IO, Sealed (FL (PrimOf p) wR))
 readPending repo = do
   pristine <- readRecorded repo
   Sealed pending <- Pending.readPending repo
@@ -387,55 +450,51 @@ readPending repo = do
     \(err :: IOException) -> do
        putStrLn $ "Yikes, pending has conflicts! " ++ show err
        putStrLn "Stashing the buggy pending as _darcs/patches/pending_buggy"
-       renameFile (darcsdir </> "patches" </> "pending")
-                  (darcsdir </> "patches" </> "pending_buggy")
+       renameFile (patchesDirPath </> "pending")
+                  (patchesDirPath </> "pending_buggy")
        return (pristine, seal NilFL)
 
-index_file, index_invalid :: FilePath
-index_file = darcsdir </> "index"
-index_invalid = darcsdir </> "index_invalid"
-
 -- | Mark the existing index as invalid. This has to be called whenever the
 -- listing of pristine changes and will cause darcs to update the index next
 -- time it tries to read it. (NB. This is about files added and removed from
 -- pristine: changes to file content in either pristine or working are handled
 -- transparently by the index reading code.)
 invalidateIndex :: t -> IO ()
-invalidateIndex _ = B.writeFile index_invalid B.empty
+invalidateIndex _ = B.writeFile indexInvalidPath B.empty
 
 readIndex :: (RepoPatch p, ApplyState p ~ Tree)
-          => Repository rt p wR wU wT -> IO I.Index
+          => Repository rt p wR wU wR -> IO I.Index
 readIndex repo = do
   (invalid, exists, formatValid) <- checkIndex
   if not exists || invalid || not formatValid
      then do pris <- readRecordedAndPending repo
-             idx <- I.updateIndexFrom index_file darcsTreeHash pris
-             when invalid $ removeFile index_invalid
+             idx <- I.updateIndexFrom indexPath darcsTreeHash pris
+             when invalid $ removeFile indexInvalidPath
              return idx
-     else I.readIndex index_file darcsTreeHash
+     else I.readIndex indexPath darcsTreeHash
 
 updateIndex :: (RepoPatch p, ApplyState p ~ Tree)
-            => Repository rt p wR wU wT -> IO ()
+            => Repository rt p wR wU wR -> IO ()
 updateIndex repo = do
   (invalid, _, _) <- checkIndex
   pris <- readRecordedAndPending repo
-  _ <- I.updateIndexFrom index_file darcsTreeHash pris
-  when invalid $ removeFile index_invalid
+  _ <- I.updateIndexFrom indexPath darcsTreeHash pris
+  when invalid $ removeFile indexInvalidPath
 
 checkIndex :: IO (Bool, Bool, Bool)
 checkIndex = do
-  invalid <- doesFileExist $ index_invalid
-  exists <- doesFileExist index_file
+  invalid <- doesFileExist $ indexInvalidPath
+  exists <- doesFileExist indexPath
   formatValid <- if exists
-                     then I.indexFormatValid index_file
+                     then I.indexFormatValid indexPath
                      else return True
   when (exists && not formatValid) $ do
 -- TODO this conditional logic (rename or delete) is mirrored in
 -- Darcs.Util.Index.updateIndexFrom and should be refactored
 #if mingw32_HOST_OS
-    renameFile index_file (index_file <.> "old")
+    renameFile indexPath (indexPath <.> "old")
 #else
-    removeFile index_file
+    removeFile indexPath
 #endif
   return (invalid, exists, formatValid)
 
@@ -443,20 +502,21 @@ checkIndex = do
 -- conflict with the recorded or unrecorded changes in a repo
 filterOutConflicts
   :: (RepoPatch p, ApplyState p ~ Tree)
-  => RL (PatchInfoAnd rt p) wX wT -- ^Recorded patches from repository, starting from
-                                  --  same context as the patches to filter
-  -> Repository rt p wR wU wT     -- ^Repository itself, used for grabbing
+  => Repository rt p wR wU wR     -- ^Repository itself, used for grabbing
                                   --  unrecorded changes
+  -> FL (PatchInfoAnd rt p) wX wR -- ^Recorded patches from repository, starting from
+                                  --  same context as the patches to filter
   -> FL (PatchInfoAnd rt p) wX wZ -- ^Patches to filter
   -> IO (Bool, Sealed (FL (PatchInfoAnd rt p) wX))
                                   -- ^True iff any patches were removed,
                                   --  possibly filtered patches
-filterOutConflicts us repository them
-     = do let commuter = commuterIdRL selfCommuter
-          unrec <- fmap n2pia . anonymous . fromPrims
+filterOutConflicts repository us them
+     = do -- Note: use of anonymous is benign here since we only try to merge cleanly
+          unrec <- fmap n2pia . anonymous
                      =<< unrecordedChanges (UseIndex, ScanKnown, MyersDiff)
                           NoLookForMoves NoLookForReplaces repository Nothing
-          them' :> rest <- return $ partitionConflictingFL commuter them (us :<: unrec)
+          them' :> rest <-
+            return $ partitionConflictingFL them (us +>+ unrec :>: NilFL)
           return (check rest, Sealed them')
   where check :: FL p wA wB -> Bool
         check NilFL = False
@@ -464,21 +524,21 @@ filterOutConflicts us repository them
 
 -- | Automatically detect file moves using the index.
 -- TODO: This function lies about the witnesses.
-getMoves :: forall rt p wR wU wT wB prim.
+getMoves :: forall rt p wR wU wB prim.
             (RepoPatch p, ApplyState p ~ Tree, prim ~ PrimOf p)
          => LookForMoves
-         -> Repository rt p wR wU wT
-         -> Maybe [SubPath]
+         -> Repository rt p wR wU wR
+         -> Maybe [AnchoredPath]
          -> IO (FL prim wB wB)
 getMoves NoLookForMoves _ _ = return NilFL
 getMoves YesLookForMoves repository files =
     mkMovesFL <$> getMovedFiles repository files
   where
     mkMovesFL [] = NilFL
-    mkMovesFL ((a,b,_):xs) = move (anchorPath "" a) (anchorPath "" b) :>: mkMovesFL xs
+    mkMovesFL ((a,b,_):xs) = move a b :>: mkMovesFL xs
 
-    getMovedFiles :: Repository rt p wR wU wT
-                  -> Maybe [SubPath]
+    getMovedFiles :: Repository rt p wR wU wR
+                  -> Maybe [AnchoredPath]
                   -> IO [(AnchoredPath, AnchoredPath, ItemType)]
     getMovedFiles repo fs = do
         old <- sortBy (comparing snd) <$> (listFileIDs =<< readIndex repo)
@@ -489,7 +549,7 @@ getMoves YesLookForMoves repository files =
                                                Just fid -> ((p, it), fid):xs) []
         new <- sortBy (comparing snd) <$>
                  (addIDs . map (\(a,b) -> (a, itemType b)) . Tree.list  =<<
-                   expand =<< applyTreeFilter nonboring <$> readPlainTree ".")
+                   expand =<< applyTreeFilter nonboring <$> readPlainTree repository)
         let match (x:xs) (y:ys)
               | snd x > snd y = match (x:xs) ys
               | snd x < snd y = match xs (y:ys)
@@ -500,9 +560,9 @@ getMoves YesLookForMoves repository files =
             fmovedfiles =
               case fs of
                 Nothing -> movedfiles
-                Just subpath ->
+                Just paths ->
                   filter (\(f1, f2, _) -> any (`elem` selfiles) [f1, f2]) movedfiles
-                  where selfiles = map (floatPath . toFilePath) subpath
+                  where selfiles = paths
         return (resolve fmovedfiles)
 
     resolve :: [(AnchoredPath, AnchoredPath, ItemType)]
@@ -538,14 +598,15 @@ getMoves YesLookForMoves repository files =
                follow _ [] currentSmallest = currentSmallest
 
         -- rewrite [d/ -> e/, .., d/f -> e/h] to [d/ -> e/, .., e/f -> e/h]
+        -- and throw out moves that don't move anything (can they be in there?)
         fixPaths [] = []
         fixPaths (y@(f1,f2,t):ys)
-                        | f1 == f2         = fixPaths ys
+                        | f1 == f2         = fixPaths ys -- no effect, throw out
                         | TreeType <- t    = y:fixPaths (map replacepp ys)
                         | otherwise        = y:fixPaths ys
-         where replacepp i@(if1,if2,it) | nfst == anchoredRoot = i
-                                        | otherwise = (nfst, if2, it)
-                where nfst = replacePrefixPath f1 f2 if1
+         -- TODO why adapt only if1 here and not if2?
+         --      is this a bug?
+         where replacepp (if1,if2,it) = (movedirfilename f1 f2 if1, if2, it)
 
 -- | Search for possible replaces between the recordedAndPending state
 -- and the unrecorded (or working) state. Return a Sealed FL list of
@@ -567,13 +628,11 @@ getReplaces YesLookForReplaces diffalg _repo pending working = do
         replaces = rmInvalidReplaces allModifiedTokens
     (patches, new_pending) <-
       flip runStateT pending $
-        forM replaces $ \(f,a,b) ->
-          doReplace defaultToks
-            (fromJust $ simpleSubPath $ fn2fp $ normPath f)
-            (BC.unpack a) (BC.unpack b)
+        forM replaces $ \(path, a, b) ->
+          doReplace defaultToks path (BC.unpack a) (BC.unpack b)
     return (new_pending, mapSeal concatFL $ toFL patches)
   where
-    modifiedTokens :: PrimOf p wX wY -> [(FileName, B.ByteString, B.ByteString)]
+    modifiedTokens :: PrimOf p wX wY -> [(AnchoredPath, B.ByteString, B.ByteString)]
     modifiedTokens p = case isHunk p of
       Just (FileHunk f _ old new) ->
         map (\(a,b) -> (f, a, b)) (concatMap checkModified $
@@ -591,22 +650,21 @@ getReplaces YesLookForReplaces diffalg _repo pending working = do
           rmInvalidReplaces $ filter (\(f'',a',_) -> f'' /= f || a' /= old) rs
     rmInvalidReplaces (r:rs) = r:rmInvalidReplaces (filter (/=r) rs)
 
-    doReplace toks f old new = do
+    doReplace toks path old new = do
         pend <- get
         mpend' <- liftIO $ maybeApplyToTree replacePatch pend
         case mpend' of
-          Nothing -> getForceReplace f toks old new
+          Nothing -> getForceReplace path toks old new
           Just pend' -> do
             put pend'
             return $ joinGap (:>:) (freeGap replacePatch) (emptyGap NilFL)
       where
-        replacePatch = tokreplace (toFilePath f) toks old new
+        replacePatch = tokreplace path toks old new
 
     getForceReplace :: (PrimPatch prim, ApplyState prim ~ Tree)
-                    => SubPath -> String -> String -> String
+                    => AnchoredPath -> String -> String -> String
                     -> StateT (Tree IO) IO (FreeLeft (FL prim))
-    getForceReplace f toks old new = do
-        let path = floatSubPath f
+    getForceReplace path toks old new = do
         -- the tree here is the "current" pending state
         tree <- get
         -- It would be nice if we could fuse the two traversals here, that is,
@@ -615,7 +673,7 @@ getReplaces YesLookForReplaces diffalg _repo pending working = do
         expandedTree <- liftIO $ expandPath tree path
         content <- case findFile expandedTree path of
           Just blob -> liftIO $ readBlob blob
-          Nothing -> bug $ "getForceReplace: not in tree: " ++ show path
+          Nothing -> error $ "getForceReplace: not in tree: " ++ show path
         let newcontent = forceTokReplace toks (BC.pack new) (BC.pack old)
                             (B.concat $ BL.toChunks content)
             tree' = modifyTree expandedTree path . Just . File $ makeBlobBS newcontent
@@ -623,11 +681,11 @@ getReplaces YesLookForReplaces diffalg _repo pending working = do
         normaliseNewTokPatch <- liftIO $ treeDiff diffalg ftf expandedTree tree'
         -- make sure we can apply them to the pending state
         patches <- return $ joinGap (+>+) normaliseNewTokPatch $ freeGap $
-            tokreplace (toFilePath f) toks old new :>: NilFL
+            tokreplace path toks old new :>: NilFL
         mtree'' <- case unFreeLeft patches of
             Sealed ps -> liftIO $ maybeApplyToTree ps tree
         case mtree'' of
-            Nothing -> bug "getForceReplace: unable to apply detected force replaces"
+            Nothing -> error "getForceReplace: unable to apply detected force replaces"
             Just tree'' -> do
                 put tree''
                 return patches
@@ -637,17 +695,15 @@ getReplaces YesLookForReplaces diffalg _repo pending working = do
 -- TODO: add witnesses for pending so we can make the types precise: currently
 -- the passed patch can be applied in any context, not just after pending.
 addPendingDiffToPending :: (RepoPatch p, ApplyState p ~ Tree)
-                          => Repository rt p wR wU wT -> UpdateWorking
-                          -> FreeLeft (FL (PrimOf p)) -> IO ()
-addPendingDiffToPending _ NoUpdateWorking  _ = return ()
-addPendingDiffToPending repo uw@YesUpdateWorking newP = do
-    (toPend :> _) <-
-        readPendingAndWorking (UseIndex, ScanKnown, MyersDiff)
-          NoLookForMoves NoLookForReplaces repo Nothing
+                        => Repository rt p wR wU wR
+                        -> FreeLeft (FL (PrimOf p)) -> IO ()
+addPendingDiffToPending repo newP = do
+    (_, Sealed toPend) <- readPending repo
     invalidateIndex repo
     case unFreeLeft newP of
-        (Sealed p) -> do recordedState <- readRecorded repo
-                         Pending.makeNewPending repo uw (toPend +>+ p) recordedState
+        (Sealed p) -> do
+            recordedState <- readRecorded repo
+            Pending.makeNewPending repo YesUpdatePending (toPend +>+ p) recordedState
 
 -- | Add an 'FL' of patches starting from the working state to the pending patch,
 -- including as much extra context as is necessary (context meaning
@@ -655,15 +711,17 @@ addPendingDiffToPending repo uw@YesUpdateWorking newP = do
 -- changes between pending and working as is possible, and including anything
 -- that doesn't commute, and the patch itself in the new pending patch.
 addToPending :: (RepoPatch p, ApplyState p ~ Tree)
-             => Repository rt p wR wU wT -> UpdateWorking
-             -> FL (PrimOf p) wU wY -> IO ()
-addToPending _ NoUpdateWorking  _ = return ()
-addToPending repo uw@YesUpdateWorking p = do
-   (toPend :> toUnrec) <- readPendingAndWorking (UseIndex, ScanKnown, MyersDiff)
+             => Repository rt p wR wU wR
+             -> UseIndex -> FL (PrimOf p) wU wY -> IO ()
+addToPending repo useidx p = do
+   (toPend :> toUnrec) <- readPendingAndWorking (useidx, ScanKnown, MyersDiff)
       NoLookForMoves NoLookForReplaces repo Nothing
    invalidateIndex repo
    case genCommuteWhatWeCanRL commuteFL (reverseFL toUnrec :> p) of
        (toP' :> p'  :> _excessUnrec) -> do
            recordedState <- readRecorded repo
-           Pending.makeNewPending repo uw
+           Pending.makeNewPending repo YesUpdatePending
             (toPend +>+ reverseRL toP' +>+ p') recordedState
+
+readPlainTree :: Repository rt p wR wU wT -> IO (Tree IO)
+readPlainTree repo  = PlainTree.readPlainTree (repoLocation repo)
