@@ -15,13 +15,13 @@
 --  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 --  Boston, MA 02110-1301, USA.
 
-{-# LANGUAGE CPP, TypeOperators, OverloadedStrings #-}
+{-# LANGUAGE CPP, OverloadedStrings #-}
 
 module Darcs.UI.Commands.Send ( send ) where
 
-import Prelude ()
 import Darcs.Prelude
 
+import System.Directory ( renameFile )
 import System.Exit
     ( exitSuccess
 #ifndef HAVE_MAPI
@@ -29,9 +29,8 @@ import System.Exit
     , exitWith
 #endif
     )
-import System.IO.Error ( ioeGetErrorString )
 import System.IO ( hClose )
-import Control.Exception ( catch, IOException )
+import Control.Exception ( catch, IOException, onException )
 import Control.Monad ( when, unless, forM_ )
 import Darcs.Util.Tree ( Tree )
 import Data.List ( intercalate, isPrefixOf )
@@ -46,6 +45,7 @@ import Darcs.UI.Commands
     , defaultRepo
     , amInHashedRepository
     )
+import Darcs.UI.Commands.Clone ( otherHelpInheritDefault )
 import Darcs.UI.Commands.Util ( printDryRunMessageAndExit, checkUnrelatedRepos )
 import Darcs.UI.Flags
     ( DarcsFlag
@@ -67,7 +67,7 @@ import Darcs.UI.Flags
     , editDescription
     )
 import Darcs.UI.Options
-    ( (^), odesc, ocheck, onormalise
+    ( (^), odesc, ocheck
     , defaultFlags, parseFlags, (?)
     )
 import qualified Darcs.UI.Options.All as O
@@ -78,6 +78,7 @@ import Darcs.Repository
     , repoLocation
     , PatchSet
     , identifyRepositoryFor
+    , ReadingOrWriting(..)
     , withRepository
     , RepoJob(..)
     , readRepo
@@ -85,13 +86,17 @@ import Darcs.Repository
     , prefsUrl )
 import Darcs.Patch.Set ( Origin )
 import Darcs.Patch.Apply( ApplyState )
-import Darcs.Patch ( IsRepoType, RepoPatch, description, applyToTree, invert )
+import Darcs.Patch ( IsRepoType, RepoPatch, description, applyToTree, effect, invert )
 import Darcs.Patch.Witnesses.Sealed ( Sealed(..) )
 import Darcs.Patch.Witnesses.Unsafe ( unsafeCoerceP )
 import Darcs.Patch.Witnesses.Ordered
     ( FL(..), (:>)(..), (:\/:)(..),
     mapFL, mapFL_FL, lengthFL, nullFL )
-import Darcs.Patch.Bundle ( minContext, makeBundleN, scanContextFile, patchFilename )
+import Darcs.Patch.Bundle
+    ( makeBundle
+    , minContext
+    , readContextFile
+    )
 import Darcs.Repository.Prefs ( addRepoSource, getPreflist )
 import Darcs.Repository.Flags ( DryRun(..) )
 import Darcs.Util.External ( fetchFilePS, Cachable(..) )
@@ -100,7 +105,6 @@ import Darcs.UI.External
     , sendEmailDoc
     , generateEmail
     , editFile
-    , catchall
     , getSystemEncoding
     , isUTF8Locale
 #ifndef HAVE_MAPI
@@ -117,7 +121,7 @@ import Darcs.Util.Lock
     )
 import Darcs.UI.SelectChanges
     ( WhichChanges(..)
-    , selectionContext
+    , selectionConfig
     , runSelection
     )
 import qualified Darcs.UI.SelectChanges as S ( PatchSelectionOptions(..) )
@@ -127,16 +131,16 @@ import Data.Text.Encoding       ( decodeUtf8' )
 import Darcs.Util.Progress ( debugMessage )
 import Darcs.UI.Email ( makeEmail )
 import Darcs.UI.Completion ( prefArgs )
+import Darcs.UI.Commands.Util ( getUniqueDPatchName )
 import Darcs.Util.Printer
-    ( Doc, vsep, text, ($$), (<+>), putDoc, putDocLn
-    , renderPS, vcat
+    ( Doc, formatWords, vsep, text, ($$), (<+>), putDoc, putDocLn
+    , quoted, renderPS, sentence, vcat
     )
 import Darcs.Util.English ( englishNum, Noun(..) )
-import Darcs.Util.Text ( sentence, quote )
+import Darcs.Util.Exception ( catchall )
 import Darcs.Util.Path ( FilePathLike, toFilePath, AbsolutePath, AbsolutePathOrStd,
                         getCurrentDirectory, useAbsoluteOrStd, makeAbsoluteOrStd )
-import Darcs.Util.Download.HTTP ( postUrl )
-import Darcs.Util.Workaround ( renameFile )
+import Darcs.Util.HTTP ( postUrl )
 import Darcs.Util.Global ( darcsSendMessage, darcsSendMessageFinal )
 import Darcs.Util.SignalHandler ( catchInterrupt )
 
@@ -146,11 +150,11 @@ patchSelOpts flags = S.PatchSelectionOptions
     , S.matchFlags = parseFlags O.matchSeveral flags
     , S.interactive = isInteractive True flags
     , S.selectDeps = selectDeps ? flags
-    , S.summary = O.summary ? flags
+    , S.withSummary = O.withSummary ? flags
     , S.withContext = O.NoContext
     }
 
-send :: DarcsCommand [DarcsFlag]
+send :: DarcsCommand
 send = DarcsCommand
     { commandProgramName = "darcs"
     , commandName = "send"
@@ -166,7 +170,6 @@ send = DarcsCommand
     , commandBasicOptions = odesc sendBasicOpts
     , commandDefaults = defaultFlags sendOpts
     , commandCheckOptions = ocheck sendOpts
-    , commandParseOptions = onormalise sendOpts
     }
   where
     sendBasicOpts
@@ -180,9 +183,10 @@ send = DarcsCommand
       ^ O.output
       ^ O.sign
       ^ O.dryRunXml
-      ^ O.summary
+      ^ O.withSummary
       ^ O.editDescription
       ^ O.setDefault
+      ^ O.inheritDefault
       ^ O.repoDir
       ^ O.minimize
       ^ O.allowUnrelatedRepos
@@ -199,10 +203,11 @@ sendCmd fps opts [""] = sendCmd fps opts []
 sendCmd (_,o) opts [unfixedrepodir] =
  withRepository (useCache ? opts) $ RepoJob $
   \(repository :: Repository rt p wR wU wR) -> do
-  context_ps <- the_context (O.sendToContext ? opts)
-  case context_ps of
-    Just them -> do
+  case O.sendToContext ? opts of
+    Just contextfile -> do
         wtds <- decideOnBehavior opts (Nothing :: Maybe (Repository rt p wR wU wR))
+        ref <- readRepo repository
+        Sealed them <- readContextFile ref (toFilePath contextfile)
         sendToThem repository opts wtds "CONTEXT" them
     Nothing -> do
         repodir <- fixUrl o unfixedrepodir
@@ -213,15 +218,13 @@ sendCmd (_,o) opts [unfixedrepodir] =
         old_default <- getPreflist "defaultrepo"
         when (old_default == [repodir]) $
             putInfo opts (creatingPatch repodir)
-        repo <- identifyRepositoryFor repository (useCache ? opts) repodir
+        repo <- identifyRepositoryFor Reading repository (useCache ? opts) repodir
         them <- readRepo repo
-        addRepoSource repodir (dryRun ? opts) (remoteRepos ? opts) (setDefault False opts)
+        addRepoSource repodir (dryRun ? opts) (remoteRepos ? opts)
+            (setDefault False opts) (O.inheritDefault ? opts) (isInteractive True opts)
         wtds <- decideOnBehavior opts (Just repo)
         sendToThem repository opts wtds repodir them
-    where
-        the_context Nothing = return Nothing
-        the_context (Just foo) = Just `fmap` scanContextFile (toFilePath foo)
-sendCmd _ _ _ = impossible
+sendCmd _ _ _ = error "impossible case"
 
 sendToThem :: (IsRepoType rt, RepoPatch p, ApplyState p ~ Tree)
            => Repository rt p wR wU wT -> [DarcsFlag] -> [WhatToDo] -> String
@@ -246,11 +249,11 @@ sendToThem repo opts wtds their_name them = do
       _     -> putVerbose opts $ selectionIs (mapFL description us')
   pristine <- readRecorded repo
   let direction = if changesReverse ? opts then FirstReversed else First
-      context = selectionContext direction "send" (patchSelOpts opts) Nothing Nothing
-  (to_be_sent :> _) <- runSelection us' context
+      selection_config = selectionConfig direction "send" (patchSelOpts opts) Nothing Nothing
+  (to_be_sent :> _) <- runSelection us' selection_config
   printDryRunMessageAndExit "send"
       (verbosity ? opts)
-      (O.summary ? opts)
+      (O.withSummary ? opts)
       (dryRun ? opts)
       O.NoXml
       (isInteractive True opts)
@@ -268,10 +271,10 @@ sendToThem repo opts wtds their_name them = do
                          Sealed (common' :> to_be_sent') -> prepareBundle opts common' (Left to_be_sent') )
                      `catchInterrupt` genFullBundle
   here   <- getCurrentDirectory
-  let make_fname (tb:>:_) = patchFilename $ patchDesc tb
-      make_fname _ = impossible
-      fname = make_fname to_be_sent
-      outname = case getOutput opts fname of
+  let make_fname (tb:>:_) = getUniqueDPatchName $ patchDesc tb
+      make_fname _ = error "impossible case"
+  fname <- make_fname to_be_sent
+  let outname = case getOutput opts fname of
                     Just f  -> Just f
                     Nothing | fst (O.sendmail ? opts) -> Nothing
                             | not $ null [ p | Post p <- wtds] -> Nothing
@@ -290,11 +293,11 @@ prepareBundle opts common e = do
   unsig_bundle <-
      case e of
        (Right (pristine, us' :\/: to_be_sent)) -> do
-         pristine' <- applyToTree (invert $ mapFL_FL hopefully us') pristine
-         makeBundleN (Just pristine')
+         pristine' <- applyToTree (invert $ effect us') pristine
+         makeBundle (Just pristine')
                      (unsafeCoerceP common)
                      (mapFL_FL hopefully to_be_sent)
-       Left to_be_sent -> makeBundleN Nothing
+       Left to_be_sent -> makeBundle Nothing
                                       (unsafeCoerceP common)
                                       (mapFL_FL hopefully to_be_sent)
   signString (parseFlags O.sign opts) unsig_bundle
@@ -354,14 +357,14 @@ sendBundle opts to_be_sent bundle fname wtds their_name=
                         (Just fname)
                contentAndBundle = Just (mailcontents, bundle)
 
-               sendmail = do
+               sendmail =
+                (do
                  sm_cmd <- getSendmailCmd opts
                  let to = generateEmailToString thetargets
                  sendEmailDoc from to thesubject (getCc opts)
-                               sm_cmd contentAndBundle body >>
-                  putInfo opts (success to (getCc opts))
-                 `catch` \e -> do warnMailBody
-                                  fail $ ioeGetErrorString e
+                               sm_cmd contentAndBundle body
+                 putInfo opts (success to (getCc opts)))
+                 `onException` warnMailBody
 
            when (null [ p | Post p <- thetargets]) sendmail
            nbody <- withOpenTemp $ \ (fh,fn) -> do
@@ -372,7 +375,7 @@ sendBundle opts to_be_sent bundle fname wtds their_name=
            forM_ [ p | Post p <- thetargets]
              (\url -> do
                 putInfo opts $ postingPatch url
-                postUrl url (BC.unpack nbody) "message/rfc822")
+                postUrl url nbody "message/rfc822")
              `catch` (\(_ :: IOException) -> sendmail)
            cleanup opts mailfile
 
@@ -448,7 +451,7 @@ collectTargets flags = [ f t | t <- O._to (O.headerFields ? flags) ] where
     f url | "http:" `isPrefixOf` url = Post url
     f em = SendMail em
 
-getDescription :: (RepoPatch p, ApplyState p ~ Tree)
+getDescription :: RepoPatch p
                => [DarcsFlag] -> String -> FL (PatchInfoAnd rt p) wX wY -> IO (Doc, Maybe String, Maybe String)
 getDescription opts their_name patches =
     case get_filename of
@@ -495,50 +498,51 @@ cmdDescription :: String
 cmdDescription =
     "Prepare a bundle of patches to be applied to some target repository."
 
-cmdHelp :: String
-cmdHelp = unlines
-    [ "Send is used to prepare a bundle of patches that can be applied to a target"
+cmdHelp :: Doc
+cmdHelp = vsep $
+  map formatWords
+  [ [ "Send is used to prepare a bundle of patches that can be applied to a target"
     , "repository.  Send accepts the URL of the repository as an argument.  When"
     , "called without an argument, send will use the most recent repository that"
     , "was either pushed to, pulled from or sent to.  By default, the patch bundle"
     , "is saved to a file, although you may directly send it by mail."
-    , ""
-    , "The `--output`, `--output-auto-name`, and `--to` flags determine"
+    ]
+  , [ "The `--output`, `--output-auto-name`, and `--to` flags determine"
     , "what darcs does with the patch bundle after creating it.  If you provide an"
     , "`--output` argument, the patch bundle is saved to that file.  If you"
     , "specify `--output-auto-name`, the patch bundle is saved to a file with an"
     , "automatically generated name.  If you give one or more `--to` arguments,"
     , "the bundle of patches is sent to those locations. The locations may either"
     , "be email addresses or urls that the patch should be submitted to via HTTP."
-    , ""
-    , "If you provide the `--mail` flag, darcs will look at the contents"
+    ]
+  , [ "If you provide the `--mail` flag, darcs will look at the contents"
     , "of the `_darcs/prefs/email` file in the target repository (if it exists),"
     , "and send the patch by email to that address.  In this case, you may use"
     , "the `--cc` option to specify additional recipients without overriding the"
     , "default repository email address."
-    , ""
-    , "If `_darcs/prefs/post` exists in the target repository, darcs will"
+    ]
+  , [ "If `_darcs/prefs/post` exists in the target repository, darcs will"
     , "upload to the URL contained in that file, which may either be a"
     , "`mailto:` URL, or an `http://` URL.  In the latter case, the"
     , "patch is posted to that URL."
-    , ""
-    , "If there is no email address associated with the repository, darcs will"
+    ]
+  , [ "If there is no email address associated with the repository, darcs will"
     , "prompt you for an email address."
-    , ""
-    , "Use the `--subject` flag to set the subject of the e-mail to be sent."
+    ]
+  , [ "Use the `--subject` flag to set the subject of the e-mail to be sent."
     , "If you don't provide a subject on the command line, darcs will make one up"
     , "based on names of the patches in the patch bundle."
-    , ""
-    , "Use the `--in-reply-to` flag to set the In-Reply-To and References headers"
+    ]
+  , [ "Use the `--in-reply-to` flag to set the In-Reply-To and References headers"
     , "of the e-mail to be sent. By default no additional headers are included so"
     , "e-mail will not be treated as reply by mail readers."
-    , ""
-    , "If you want to include a description or explanation along with the bundle"
+    ]
+  , [ "If you want to include a description or explanation along with the bundle"
     , "of patches, you need to specify the `--edit-description` flag, which"
     , "will cause darcs to open up an editor with which you can compose a message"
     , "to go along with your patches."
-    , ""
-    , "If you want to use a command different from the default one for sending"
+    ]
+  , [ "If you want to use a command different from the default one for sending"
     , "email, you need to specify a command line with the `--sendmail-command`"
     , "option. The command line can contain some format specifiers which are"
     , "replaced by the actual values. Accepted format specifiers are `%s` for"
@@ -548,39 +552,58 @@ cmdHelp = unlines
     , "Additionally you can add `%<` to the end of the command line if the command"
     , "expects the complete email message on standard input. E.g. the command lines"
     , "for evolution and msmtp look like this:"
-    , ""
-    , "    evolution \"mailto:%T?subject=%S&attach=%A&cc=%C&body=%B\""
+    ]
+  ]
+  ++
+  -- TODO autoformatting for indented paragraphs
+  [ vcat
+    [ "    evolution \"mailto:%T?subject=%S&attach=%A&cc=%C&body=%B\""
     , "    msmtp -t %<"
-    , ""
-    , "Do not confuse the `--author` options with the return address"
+    ]
+  ]
+  ++ map formatWords
+  [ [ "Do not confuse the `--author` options with the return address"
     , "that `darcs send` will set for your patch bundle."
-    , ""
-    , "For example, if you have two email addresses A and B:"
-    , ""
-    , "* If you use `--author A` but your machine is configured to send mail from"
-    , "  address B by default, then the return address on your message will be B."
-    , "* If you use `--from A` and your mail client supports setting the"
-    , "  From: address arbitrarily (some non-Unix-like mail clients, especially,"
-    , "  may not support this), then the return address will be A; if it does"
-    , "  not support this, then the return address will be B."
-    , "* If you supply neither `--from` nor `--author` then the return"
-    , "  address will be B."
-    , ""
-    , "In addition, unless you specify the sendmail command with"
+    ]
+  , [ "For example, if you have two email addresses A and B:"
+    ]
+  ]
+  ++
+  -- TODO autoformatting for bullet lists
+  [ vcat
+    [ "  * If you use `--author A` but your machine is configured to send"
+    , "    mail from address B by default, then the return address on your"
+    , "    message will be B."
+    , "  * If you use `--from A` and your mail client supports setting the"
+    , "    From: address arbitrarily (some non-Unix-like mail clients,"
+    , "    especially, may not support this), then the return address will"
+    , "    be A; if it does not support this, then the return address will"
+    , "    be B."
+    , "  * If you supply neither `--from` nor `--author` then the return"
+    , "    address will be B."
+    ]
+  ]
+  ++
+  [ formatWords
+    [ "In addition, unless you specify the sendmail command with"
     , "`--sendmail-command`, darcs sends email using the default email"
     , "command on your computer. This default command is determined by the"
     , "`configure` script. Thus, on some non-Unix-like OSes,"
     , "`--from` is likely to not work at all."
     ]
+  , otherHelpInheritDefault
+  ]
 
 cannotSendToSelf :: String
 cannotSendToSelf = "Can't send to current repository! Did you mean send --context?"
 
 creatingPatch :: String -> Doc
-creatingPatch repodir = "Creating patch to" <+> text (quote repodir) <> "..."
+creatingPatch repodir = "Creating patch to" <+> quoted repodir <> "..."
 
+#ifndef HAVE_MAPI
 noWorkingSendmail :: Doc
 noWorkingSendmail = "No working sendmail instance on your machine!"
+#endif
 
 nothingSendable :: Doc
 nothingSendable = "No recorded local changes to send!"
@@ -592,7 +615,7 @@ selectionIsNull :: Doc
 selectionIsNull = text "You don't want to send any patches, and that's fine with me!"
 
 emailBackedUp :: String -> Doc
-emailBackedUp mf = sentence $ "Email body left in" <+> text mf
+emailBackedUp mf = sentence $ "Email body left in" <+> text mf <> "."
 
 promptCharSetWarning :: String -> String
 promptCharSetWarning msg = "Warning: " ++ msg ++ "  Send anyway?"
