@@ -19,11 +19,6 @@ module Darcs.UI.Commands.ShowAuthors
     ( showAuthors, Spelling, compiledAuthorSpellings, canonizeAuthor, rankAuthors
     ) where
 
-import Prelude ()
-import Darcs.Prelude
-
-import Prelude hiding ( (^) )
-
 import Control.Arrow ( (&&&), (***) )
 import Data.Char ( toLower, isSpace )
 import Data.Function ( on )
@@ -35,8 +30,10 @@ import Text.ParserCombinators.Parsec hiding ( lower, count, Line )
 import Text.ParserCombinators.Parsec.Error
 import Text.Regex ( Regex, mkRegexWithOpts, matchRegex )
 
+import Darcs.Prelude
+
 import Darcs.UI.Flags ( DarcsFlag, useCache, verbose )
-import Darcs.UI.Options ( oid, odesc, ocheck, onormalise, defaultFlags, (?) )
+import Darcs.UI.Options ( oid, odesc, ocheck, defaultFlags, (?) )
 import qualified Darcs.UI.Options.All as O
 import Darcs.UI.Commands ( DarcsCommand(..), withStdOpts, nodefaults, putWarning, amInRepository )
 import Darcs.UI.Completion ( noArgs )
@@ -47,7 +44,7 @@ import Darcs.Patch.Set ( patchSet2RL )
 import Darcs.Repository ( readRepo, withRepository, RepoJob(..) )
 import Darcs.Patch.Witnesses.Ordered ( mapRL )
 import Darcs.Util.Lock ( readTextFile )
-import Darcs.Util.Printer ( text )
+import Darcs.Util.Printer ( Doc, text )
 import Darcs.Util.Path ( AbsolutePath )
 
 data Spelling = Spelling String String [Regex] -- name, email, regexps
@@ -56,8 +53,8 @@ type ParsedLine = Maybe Spelling -- Nothing for blank lines
 showAuthorsDescription :: String
 showAuthorsDescription = "List authors by patch count."
 
-showAuthorsHelp :: String
-showAuthorsHelp =
+showAuthorsHelp :: Doc
+showAuthorsHelp = text $
  "The `darcs show authors` command lists the authors of the current\n" ++
  "repository, sorted by the number of patches contributed.  With the\n" ++
  "`--verbose` option, this command simply lists the author of each patch\n" ++
@@ -96,7 +93,7 @@ showAuthorsHelp =
  "    John Snagge <snagge@bbc.co.uk>, John, snagge@, js@(si|mit).edu\n" ++
  "    Chuck Jones\\, Jr. <chuck@pobox.com>, cj\\+user@example.com\n"
 
-showAuthors :: DarcsCommand [DarcsFlag]
+showAuthors :: DarcsCommand
 showAuthors = DarcsCommand
     { commandProgramName = "darcs"
     , commandName = "authors"
@@ -112,7 +109,6 @@ showAuthors = DarcsCommand
     , commandBasicOptions = odesc showAuthorsBasicOpts
     , commandDefaults = defaultFlags showAuthorsOpts
     , commandCheckOptions = ocheck showAuthorsOpts
-    , commandParseOptions = onormalise showAuthorsOpts
     }
   where
     showAuthorsBasicOpts = O.repoDir
