@@ -8,7 +8,6 @@ module Darcs.UI.External
     , signString
     , verifyPS
     , execDocPipe
-    , execPipeIgnoreError
     , pipeDoc
     , pipeDocSSH
     , viewDoc
@@ -19,20 +18,17 @@ module Darcs.UI.External
     , darcsProgram
     , editText
     , editFile
-    , catchall
     --  * Locales
     , setDarcsEncodings
     , getSystemEncoding
     , isUTF8Locale
     ) where
 
-import Prelude ()
 import Darcs.Prelude
-import Darcs.Util.Text ( showCommandLine )
 
 import Data.Maybe ( isJust, isNothing, maybeToList )
 import Control.Monad ( unless, when, filterM, liftM2, void )
-import GHC.MVar ( MVar )
+import Control.Concurrent.MVar ( MVar )
 import System.Exit ( ExitCode(..) )
 import System.Environment
     ( getEnv
@@ -46,10 +42,12 @@ import System.FilePath.Posix ( (</>) )
 import System.Process ( createProcess, proc, CreateProcess(..), runInteractiveProcess, waitForProcess, StdStream(..) )
 import System.Process.Internals ( ProcessHandle )
 
+#ifndef WIN32
 import GHC.IO.Encoding
     ( getFileSystemEncoding
     , setForeignEncoding
     , setLocaleEncoding )
+#endif
 
 import Foreign.C.String ( CString, peekCString )
 
@@ -68,7 +66,6 @@ import Foreign.Ptr ( nullPtr )
 import Darcs.Util.Lock ( canonFilename, writeDocBinFile )
 #endif
 
-import Darcs.Util.SignalHandler ( catchNonSignal )
 import Darcs.UI.Options.All ( Sign(..), Verify(..), Compression(..) )
 import Darcs.Util.Path
     ( AbsolutePath
@@ -90,6 +87,7 @@ import Darcs.Util.Lock
     )
 import Darcs.Util.Ssh ( getSSH, SSHCmd(..) )
 import Darcs.Util.CommandLine ( parseCmd, addUrlencoded )
+import Darcs.Util.Exception ( catchall )
 import Darcs.Util.Exec ( execInteractive, exec, Redirect(..), withoutNonBlock )
 import Darcs.Util.URL ( SshFilePath, sshUhost )
 import Darcs.Util.Printer ( Doc, Printers, hPutDocLnWith, hPutDoc, hPutDocLn, hPutDocWith, ($$), renderPS,
@@ -126,7 +124,7 @@ data WhereToPipe = PipeToSsh Compression -- ^ if pipe to ssh, can choose to comp
 
 pipeDocInternal :: WhereToPipe -> String -> [String] -> Doc -> IO ExitCode
 pipeDocInternal whereToPipe c args inp = withoutNonBlock $ withoutProgress $
-    do debugMessage $ "Exec: " ++ showCommandLine (c:args)
+    do debugMessage $ "Exec: " ++ unwords (map show (c:args))
        (Just i,_,_,pid) <- createProcess (proc c args){ std_in = CreatePipe
                                                       , delegate_ctlc = True}
        debugMessage "Start transferring data"
@@ -321,15 +319,6 @@ execDocPipe c args instr = withoutProgress $ do
                           "' failed with exit code "++ show ec
          ExitSuccess -> return $ packedString out
 
--- The following is needed for diff, which returns non-zero whenever
--- the files differ.
-execPipeIgnoreError :: String -> [String] -> Doc -> IO Doc
-execPipeIgnoreError c args instr = withoutProgress $ do
-       (pid, mvare, out) <- execAndGetOutput c args instr
-       _ <- waitForProcess pid
-       takeMVar mvare
-       return $ if B.null out then empty else packedString out
-
 signString :: Sign -> Doc -> IO Doc
 signString NoSign d = return d
 signString Sign d = signPGP [] d
@@ -507,12 +496,13 @@ runEditor :: FilePath
           -> IO ExitCode
 runEditor f = do
     ed <- getEditor
-    execInteractive ed f
-         `ortryrunning` execInteractive "vi" f
-         `ortryrunning` execInteractive "emacs" f
-         `ortryrunning` execInteractive "emacs -nw" f
+    let mf = Just f
+    execInteractive ed mf
+         `ortryrunning` execInteractive "vi" mf
+         `ortryrunning` execInteractive "emacs" mf
+         `ortryrunning` execInteractive "emacs -nw" mf
 #ifdef WIN32
-         `ortryrunning` execInteractive "edit" f
+         `ortryrunning` execInteractive "edit" mf
 #endif
 
 
@@ -521,12 +511,6 @@ getEditor = getEnv "DARCS_EDITOR" `catchall`
              getEnv "VISUAL" `catchall`
              getEnv "EDITOR" `catchall` return "nano"
 
-catchall :: IO a
-         -> IO a
-         -> IO a
-a `catchall` b = a `catchNonSignal` (\_ -> b)
-
-
 -- | On Posix systems, GHC by default uses the user's locale encoding to
 -- determine how to decode/encode the raw byte sequences in the Posix API
 -- to/from 'String'. It also uses certain special variants of this
