@@ -16,6 +16,10 @@
 -- Boston, MA 02110-1301, USA.
 
 {-# LANGUAGE OverloadedStrings #-}
+-- | Helper functions to access option contents. Some of them are here only to
+-- ease the transition from the legacy system where we manually parsed the flag
+-- list to the new(er) option system. At some point this module should be
+-- renamed and the re-exports from "Darcs.UI.Options.All" removed.
 module Darcs.UI.Flags
     ( F.DarcsFlag
     , remoteDarcs
@@ -177,7 +181,10 @@ diffingOpts flags = diffOpts (O.useIndex ? flags) (lookForAdds flags)
 
 -- | This will become dis-entangled as soon as we inline these functions.
 wantGuiPause :: Config -> O.WantGuiPause
-wantGuiPause fs = if (hasDiffCmd fs || hasExternalMerge fs) && hasPause fs then O.YesWantGuiPause else O.NoWantGuiPause
+wantGuiPause fs =
+  if (hasDiffCmd fs || hasExternalMerge fs) && hasPause fs
+    then O.YesWantGuiPause
+    else O.NoWantGuiPause
   where
     hasDiffCmd = isJust . O.diffCmd . parseFlags O.extDiff
     hasExternalMerge = (/= O.NoExternalMerge) . parseFlags O.externalMerge
@@ -279,7 +286,9 @@ pathSetFromArgs fps args = do
 -- position of the path that cannot be converted.
 --
 -- It is intended for validating file arguments to darcs commands.
-maybeFixSubPaths :: (AbsolutePath, AbsolutePath) -> [String] -> IO [Maybe AnchoredPath]
+maybeFixSubPaths :: (AbsolutePath, AbsolutePath)
+                 -> [String]
+                 -> IO [Maybe AnchoredPath]
 maybeFixSubPaths (r, o) fs = do
   fixedFs <- mapM (fmap dropInDarcsdir . fixit) fs
   let bads = snd . unzip . filter (isNothing . fst) $ zip fixedFs fs
@@ -332,7 +341,10 @@ environmentHelpEmail = (["DARCS_EMAIL","EMAIL"], fileHelpAuthor)
 -- and if it's not possible, ask the user.
 getAuthor :: Maybe String -> Bool -> IO String
 getAuthor (Just author) _ = return author
-getAuthor Nothing pipe = if pipe then askUser "Who is the author? " else promptAuthor True False
+getAuthor Nothing pipe =
+  if pipe
+    then askUser "Who is the author? "
+    else promptAuthor True False
 
 -- | 'promptAuthor' try to guess the author, from repository or
 -- global preference files or environment variables, and
@@ -397,10 +409,10 @@ promptAuthor store alwaysAsk = do
                      unless exists $ createDirectory dir
                      return dir
 
--- | 'getEasyAuthor' tries to get the author name first from the repository preferences,
--- then from global preferences, then from environment variables.  Returns @[]@
--- if it could not get it.  Note that it may only return multiple possibilities when
--- reading from global preferences
+-- | 'getEasyAuthor' tries to get the author name first from the repository
+-- preferences, then from global preferences, then from environment variables.
+-- Returns @[]@ if it could not get it. Note that it may only return multiple
+-- possibilities when reading from global preferences.
 getEasyAuthor :: IO [String]
 getEasyAuthor =
   firstNotNullIO [ (take 1 . nonblank) `fmap` getPreflist "author"
@@ -440,7 +452,8 @@ environmentHelpSendmail = (["SENDMAIL"], [
 -- to be used by @darcs send@. Looks for a command specified by
 -- @SendmailCmd \"command\"@ in that list of flags, if any.
 -- This flag is present if darcs was invoked with @--sendmail-command=COMMAND@
--- Alternatively the user can set @$S@@ENDMAIL@ which will be used as a fallback if present.
+-- Alternatively the user can set @$S@@ENDMAIL@ which will be used as a
+-- fallback if present.
 getSendmailCmd :: Config -> IO String
 getSendmailCmd fs = case parseFlags O.sendmailCmd fs of
   Just cmd -> return cmd
