@@ -1,4 +1,3 @@
-{-# LANGUAGE RecordWildCards #-}
 {- | All the concrete options.
 
 Notes:
@@ -42,7 +41,7 @@ module Darcs.UI.Options.All
     , Verbosity (..) -- re-export
     , verbosity
     , timings
-    , anyVerbosity
+    , debugging
     , HooksConfig (..) -- re-export
     , HookConfig (..) -- re-export
     , preHook
@@ -71,13 +70,11 @@ module Darcs.UI.Options.All
     , maxCount
 
     -- local or remote repo(s)
-    , WorkRepo (..) -- re-export
-    , workRepo
     , repoDir
     , RemoteRepos (..) -- re-export
     , remoteRepos
     , possiblyRemoteRepo
-    , reponame
+    , newRepo
     , NotInRemote (..)
     , notInRemote
     , notInRemoteFlagName
@@ -89,6 +86,8 @@ module Darcs.UI.Options.All
     , withWorkingDir
     , SetDefault (..) -- re-export
     , setDefault
+    , InheritDefault (..) -- re-export
+    , inheritDefault
 
     -- patch meta-data
     , patchname
@@ -145,9 +144,6 @@ module Darcs.UI.Options.All
     , sendmailCmd
     , charset
     , editDescription
-    , ccApply
-    , reply
-    , happyForwarding
 
     -- patch bundles
     , applyAs
@@ -162,6 +158,7 @@ module Darcs.UI.Options.All
     , conflictsYes
     , ExternalMerge (..) -- re-export
     , externalMerge
+    , reorder
 
     -- optimizations
     , Compression (..) -- re-export
@@ -171,15 +168,14 @@ module Darcs.UI.Options.All
     , patchIndexNo
     , patchIndexYes
     , Reorder (..) -- re-export
-    , reorder
     , minimize
     , storeInMemory
 
     -- miscellaneous
     , Output (..)
     , output
-    , Summary (..)
-    , summary
+    , WithSummary (..)
+    , withSummary
     , maybeSummary
     , RemoteDarcs (..) -- re-export
     , NetworkOptions (..)
@@ -188,7 +184,6 @@ module Darcs.UI.Options.All
     , umask
     , SetScriptsExecutable (..) -- re-export
     , setScriptsExecutable
-    , restrictPaths
 
     -- command specific
 
@@ -243,16 +238,10 @@ module Darcs.UI.Options.All
 
     -- optimize
     , siblings
-    , optimizePatchIndex
     ) where
 
-import Prelude ()
 import Darcs.Prelude
 
-import Prelude hiding ( (^) )
-import Data.Char ( isDigit )
-import Data.List ( intercalate )
-
 import Darcs.Repository.Flags
     ( Compression (..)
     , RemoteDarcs (..)
@@ -270,11 +259,11 @@ import Darcs.Repository.Flags
     , LeaveTestDir (..)
     , RemoteRepos (..)
     , SetDefault (..)
+    , InheritDefault (..)
     , UseIndex (..)
     , ScanKnown (..)
     , CloneKind (..)
     , ExternalMerge (..)
-    , WorkRepo (..)
     , AllowConflicts (..)
     , WantGuiPause (..)
     , WithPatchIndex (..)
@@ -373,6 +362,10 @@ instance YesNo EnumPatches where
   yes NoEnumPatches = False
   yes YesEnumPatches = True
 
+instance YesNo InheritDefault where
+  yes NoInheritDefault = False
+  yes YesInheritDefault = True
+
 -- * Root command
 
 -- | Options for darcs iself that act like sub-commands.
@@ -399,7 +392,7 @@ data StdCmdAction = Help | ListOptions | Disable deriving (Eq, Show)
 stdCmdActions :: PrimDarcsOption (Maybe StdCmdAction)
 stdCmdActions = withDefault Nothing
   [ RawNoArg [] ["help"] F.Help (Just Help)
-    "show a brief description of the command and its options"
+    "show a description of the command and its options"
   , RawNoArg [] ["list-options"] F.ListOptions (Just ListOptions)
     "show plain list of available options and commands, for auto-completion"
   , RawNoArg [] ["disable"] F.Disable (Just Disable) "disable this command" ]
@@ -407,7 +400,7 @@ stdCmdActions = withDefault Nothing
 -- ** Verbosity related
 
 debug :: PrimDarcsOption Bool
-debug = singleNoArg [] ["debug"] F.Debug "give only debug output"
+debug = singleNoArg [] ["debug"] F.Debug "enable general debug output"
 
 debugHttp :: PrimDarcsOption Bool
 debugHttp = singleNoArg [] ["debug-http"] F.DebugHTTP "debug output from libcurl"
@@ -417,13 +410,13 @@ verbosity = withDefault NormalVerbosity
   [ RawNoArg ['q'] ["quiet"] F.Quiet Quiet "suppress informational output"
   , RawNoArg [] ["standard-verbosity"] F.NormalVerbosity NormalVerbosity
       "neither verbose nor quiet output"
-  , RawNoArg ['v'] ["verbose"] F.Verbose Verbose "give verbose output" ]
+  , RawNoArg ['v'] ["verbose"] F.Verbose Verbose "enable verbose output" ]
 
 timings :: PrimDarcsOption Bool
 timings = singleNoArg [] ["timings"] F.Timings "provide debugging timings information"
 
-anyVerbosity :: DarcsOption a (Bool -> Bool -> Verbosity -> Bool -> a)
-anyVerbosity = debug ^ debugHttp ^ verbosity ^ timings where
+debugging :: DarcsOption a (Bool -> Bool -> Bool -> a)
+debugging = debug ^ debugHttp ^ timings
 
 -- ** Hooks
 
@@ -480,12 +473,10 @@ useCache = (imap . cps) (Iso fw bw) $ singleNoArg [] ["no-cache"] F.NoCache "don
 
 -- * Interactivity related
 
-{- TODO: these options interact (no pun intended) in complex ways that are
+{- TODO: These options interact (no pun intended) in complex ways that are
 very hard to figure out for users as well as maintainers. I think the only
 solution here is a more radical (and probably incompatible) re-design
-involving all interactivity related options. That is beyond the goals of
-this sub-project (which is already large enough).
--}
+involving all interactivity related options. -}
 
 data XmlOutput = NoXml | YesXml deriving (Eq, Show)
 
@@ -549,52 +540,31 @@ changesReverse = withDefault False
   [ RawNoArg [] ["reverse"] F.Reverse True "show/consider changes in reverse order"
   , RawNoArg [] ["no-reverse"] F.Forward False "show/consider changes in the usual order" ]
 
--- | TODO: Returning @-1@ if the argument cannot be parsed as an integer is
--- not something I expected to find in a Haskell program. Instead, the flag
--- should take either a plain 'String' argument (leaving it to a later stage
--- to parse the 'String' to an 'Int'), or else a @'Maybe' 'Int'@, taking
--- the possibility of a failed parse into account.
 maxCount :: PrimDarcsOption (Maybe Int)
-maxCount = (withDefault Nothing
-    [ RawStrArg [] ["max-count"] F.MaxCount unF toV unV "NUMBER"
-      "return only NUMBER results" ])
-    {ocheck=check}
+maxCount = withDefault Nothing
+  [ RawStrArg [] ["max-count"] F.MaxCount unF toV unV "NUMBER" "return only NUMBER results" ]
   where
     unF f = [ s | F.MaxCount s <- [f] ]
-    unV x = [ show s | Just s <- [x] ]
-    toV s = if good s then Just (read s) else Nothing
-    check fs =
-      [ "invalid argument to --max-count: '"++s++"'" | s <- args, not (good s) ] ++
-      if length args > 1
-        then ["conflicting flags: " ++ intercalate ", " (map ("--max-count="++) args)]
-        else []
-      where
-        args = [ s | F.MaxCount s <- fs ]
-    good s = not (null s) && all isDigit s
+    unV x = [ showIntArg n | Just n <- [x] ]
+    toV = Just . parseIntArg "count" (>=0)
 
 -- * Local or remote repo
 
-workRepo :: PrimDarcsOption WorkRepo
-workRepo = imap (Iso fw bw) $ repoDir ^ possiblyRemoteRepo where
-  fw k (WorkRepoDir s)         = k (Just s) Nothing
-  fw k (WorkRepoPossibleURL s) = k Nothing  (Just s)
-  fw k WorkRepoCurrentDir      = k Nothing  Nothing
-  bw k (Just s) _              = k (WorkRepoDir s)
-  bw k Nothing  (Just s)       = k (WorkRepoPossibleURL s)
-  bw k Nothing  Nothing        = k WorkRepoCurrentDir
-
 repoDir :: PrimDarcsOption (Maybe String)
 repoDir = singleStrArg [] ["repodir"] F.WorkRepoDir arg "DIRECTORY"
     "specify the repository directory in which to run"
   where arg (F.WorkRepoDir s) = Just s
         arg _ = Nothing
 
--- | @--repodir@ is there for compatibility, should be removed eventually
+-- | This option is for when a new repo gets created. Used for clone, convert
+-- import, convert darcs-2, and initialize. For clone and initialize it has the
+-- same effect as giving the name as a normal argument.
 --
--- IMHO the whole option can disappear; it overlaps with using an extra (non-option)
--- argument, which is how e.g. @darcs get@ is usually invoked.
-reponame :: PrimDarcsOption (Maybe String)
-reponame = singleStrArg [] ["repo-name","repodir"] F.NewRepo arg "DIRECTORY" "path of output directory"
+-- The @--repodir@ alias is there for compatibility, should be removed eventually.
+--
+-- TODO We need a way to deprecate options / option names.
+newRepo :: PrimDarcsOption (Maybe String)
+newRepo = singleStrArg [] ["repo-name","repodir"] F.NewRepo arg "DIRECTORY" "path of output directory"
   where arg (F.NewRepo s) = Just s; arg _ = Nothing
 
 possiblyRemoteRepo :: PrimDarcsOption (Maybe String)
@@ -652,15 +622,20 @@ justThisRepo = singleNoArg [] ["just-this-repo"] F.JustThisRepo
 withWorkingDir :: PrimDarcsOption WithWorkingDir
 withWorkingDir = withDefault WithWorkingDir
   [ RawNoArg [] ["with-working-dir"] F.UseWorkingDir WithWorkingDir
-    "Create a working directory (normal repository)"
+    "Create a working tree (normal repository)"
   , RawNoArg [] ["no-working-dir"] F.UseNoWorkingDir NoWorkingDir
-    "Do not create a working directory (bare repository)" ]
+    "Do not create a working tree (bare repository)" ]
 
 setDefault :: PrimDarcsOption (Maybe Bool)
 setDefault = withDefault Nothing
   [ RawNoArg [] ["set-default"] F.SetDefault (Just True) "set default repository"
   , RawNoArg [] ["no-set-default"] F.NoSetDefault (Just False) "don't set default repository" ]
 
+inheritDefault :: PrimDarcsOption InheritDefault
+inheritDefault = withDefault NoInheritDefault
+  [ RawNoArg [] ["inherit-default"] F.InheritDefault YesInheritDefault "inherit default repository"
+  , RawNoArg [] ["no-inherit-default"] F.NoInheritDefault NoInheritDefault "don't inherit default repository" ]
+
 -- * Specifying patch meta-data
 
 patchname :: PrimDarcsOption (Maybe String)
@@ -678,7 +653,6 @@ author = singleStrArg ['A'] ["author"] F.Author arg
 data AskLongComment = NoEditLongComment | YesEditLongComment | PromptLongComment
   deriving (Eq, Show)
 
--- TODO: fix non-default behavior
 askLongComment :: PrimDarcsOption (Maybe AskLongComment)
 askLongComment = withDefault Nothing
   [ RawNoArg [] ["edit-long-comment"] F.EditLongComment (Just YesEditLongComment)
@@ -967,27 +941,6 @@ editDescription = withDefault True
   , RawNoArg [] ["dont-edit-description","no-edit-description"] F.NoEditDescription False
     "don't edit the patch bundle description" ]
 
--- TODO: turn these two into a combined option
-
-ccApply :: PrimDarcsOption (Maybe String)
-ccApply = singleStrArg [] ["cc"] F.Cc arg
-    "EMAIL" "mail results to additional EMAIL(s). Requires --reply"
-  where arg (F.Cc s) = Just s
-        arg _ = Nothing
-
-reply :: PrimDarcsOption (Maybe String)
-reply = singleStrArg [] ["reply"] F.Reply arg "FROM"
-    "reply to email-based patch using FROM address"
-  where arg (F.Reply s) = Just s
-        arg _ = Nothing
-
-happyForwarding :: PrimDarcsOption Bool
-happyForwarding = withDefault False
-  [ RawNoArg [] ["happy-forwarding"] F.HappyForwarding True
-    "forward unsigned messages without extra header"
-  , RawNoArg [] ["no-happy-forwarding"] F.NoHappyForwarding False
-    "don't forward unsigned messages without extra header" ]
-
 -- * Patch bundle related
 
 applyAs :: PrimDarcsOption (Maybe String)
@@ -1064,6 +1017,14 @@ externalMerge = imap (Iso fw bw) $ singleStrArg [] ["external-merge"] F.External
     fw k (YesExternalMerge s) = k (Just s)
     fw k NoExternalMerge = k Nothing
 
+-- | pull, apply, rebase pull, rebase apply
+reorder :: PrimDarcsOption Reorder
+reorder = withDefault NoReorder
+  [ RawNoArg [] ["reorder-patches"] F.Reorder Reorder
+    "put local-only patches on top of remote ones"
+  , RawNoArg [] ["no-reorder-patches"] F.NoReorder NoReorder
+    "put remote-only patches on top of local ones" ]
+
 -- * Optimizations
 
 compress :: PrimDarcsOption Compression
@@ -1122,15 +1083,15 @@ output = withDefault Nothing
 
 -- * Miscellaneous
 
-data Summary = NoSummary | YesSummary deriving (Eq, Show)
+data WithSummary = NoSummary | YesSummary deriving (Eq, Show)
 
-instance YesNo Summary where
+instance YesNo WithSummary where
   yes NoSummary = False
   yes YesSummary = True
 
 -- all commands except whatsnew
-summary :: PrimDarcsOption Summary
-summary = (imap . cps) (Iso fw bw) $ maybeSummary Nothing
+withSummary :: PrimDarcsOption WithSummary
+withSummary = (imap . cps) (Iso fw bw) $ maybeSummary Nothing
   where
     fw Nothing = NoSummary
     fw (Just NoSummary) = NoSummary
@@ -1139,7 +1100,7 @@ summary = (imap . cps) (Iso fw bw) $ maybeSummary Nothing
     bw YesSummary = Just YesSummary
 
 -- needed for whatsnew
-maybeSummary :: Maybe Summary -> PrimDarcsOption (Maybe Summary)
+maybeSummary :: Maybe WithSummary -> PrimDarcsOption (Maybe WithSummary)
 maybeSummary def = withDefault def
   [ RawNoArg ['s'] ["summary"] F.Summary (Just YesSummary) "summarize changes"
   , RawNoArg [] ["no-summary"] F.NoSummary (Just NoSummary) "don't summarize changes" ]
@@ -1182,14 +1143,6 @@ setScriptsExecutable = withDefault NoSetScriptsExecutable
   , RawNoArg [] ["dont-set-scripts-executable","no-set-scripts-executable"]
     F.DontSetScriptsExecutable NoSetScriptsExecutable "don't make scripts executable" ]
 
-restrictPaths :: PrimDarcsOption Bool
-restrictPaths = withDefault True
-  [ RawNoArg [] ["restrict-paths"] F.RestrictPaths True
-    "don't allow darcs to touch external files or repo metadata"
-  , RawNoArg [] ["dont-restrict-paths","no-restrict-paths"]
-    F.DontRestrictPaths False
-    "allow darcs to modify any file or directory (unsafe)" ]
-
 -- * Specific to a single command
 
 -- ** amend
@@ -1211,10 +1164,10 @@ machineReadable = withDefault False
   , __machineReadable True ]
 
 __humanReadable :: RawDarcsOption
-__humanReadable val = RawNoArg [] ["human-readable"] F.HumanReadable val "give human-readable output"
+__humanReadable val = RawNoArg [] ["human-readable"] F.HumanReadable val "normal human-readable output"
 
 __machineReadable :: RawDarcsOption
-__machineReadable val = RawNoArg [] ["machine-readable"] F.MachineReadable val "give machine-readable output"
+__machineReadable val = RawNoArg [] ["machine-readable"] F.MachineReadable val "machine-readable output"
 
 -- ** clone
 
@@ -1249,7 +1202,9 @@ hashed = deprecated
 
 patchFormat :: PrimDarcsOption PatchFormat
 patchFormat = withDefault PatchFormat2
-  [ RawNoArg [] ["darcs-2"] F.UseFormat2 PatchFormat2
+  [ RawNoArg [] ["darcs-3"] F.UseFormat3 PatchFormat3
+    "New darcs patch format"
+  , RawNoArg [] ["darcs-2"] F.UseFormat2 PatchFormat2
     "Standard darcs patch format"
   , RawNoArg [] ["darcs-1"] F.UseFormat1 PatchFormat1
     "Older patch format (for compatibility)" ]
@@ -1277,7 +1232,7 @@ data ChangesFormat
 
 changesFormat :: PrimDarcsOption (Maybe ChangesFormat)
 changesFormat = withDefault Nothing
-  [ RawNoArg [] ["context"] F.GenContext (Just GenContext) "give output suitable for clone --context"
+  [ RawNoArg [] ["context"] F.GenContext (Just GenContext) "produce output suitable for clone --context"
   , __xmloutput (Just GenXml)
   , __humanReadable (Just HumanReadable)
   , __machineReadable (Just MachineReadable)
@@ -1354,15 +1309,3 @@ siblings :: PrimDarcsOption [AbsolutePath]
 siblings = multiAbsPathArg [] ["sibling"] F.Sibling mkV "DIRECTORY"
     "specify a sibling directory"
   where mkV fs = [ s | F.Sibling s <- fs ]
-
-reorder :: PrimDarcsOption Reorder
-reorder = withDefault NoReorder
-  [ RawNoArg [] ["reorder-patches"] F.Reorder Reorder
-    "reorder the patches in the repository"
-  , RawNoArg [] ["no-reorder-patches"] F.NoReorder NoReorder
-    "don't reorder the patches in the repository" ]
-
-optimizePatchIndex :: PrimDarcsOption (Maybe WithPatchIndex)
-optimizePatchIndex = withDefault Nothing
-  [ __patchIndex (Just YesPatchIndex)
-  , __noPatchIndex (Just NoPatchIndex) ]
