@@ -15,28 +15,29 @@
 -- the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 -- Boston, MA 02110-1301, USA.
 
+{-# LANGUAGE RecordWildCards #-}
 module Darcs.UI.SelectChanges
     ( -- * Working with changes
       WhichChanges(..)
     , viewChanges
-    , withSelectedPatchFromRepo
+    , withSelectedPatchFromList
     , runSelection
-    , selectionContextPrim
-    , selectionContextGeneric
-    , selectionContext
-    , PatchSelectionContext(allowSkipAll)
-    , printSummary
+    , runInvertibleSelection
+    , selectionConfigPrim
+    , selectionConfigGeneric
+    , selectionConfig
+    , SelectionConfig(allowSkipAll)
     -- * Interactive selection utils
     , PatchSelectionOptions(..)
     , InteractiveSelectionM
-    , InteractiveSelectionContext(..)
+    , InteractiveSelectionState(..)
+    , initialSelectionState
     -- ** Navigating the patchset
     , currentPatch
     , skipMundane
     , skipOne
     , backOne
     , backAll
-    , showCur
     -- ** Decisions
     , decide
     , decideWholeFile
@@ -51,7 +52,6 @@ module Darcs.UI.SelectChanges
     , askAboutDepends
     ) where
 
-import Prelude ()
 import Darcs.Prelude
 
 import Control.Monad ( liftM, unless, when, (>=>) )
@@ -67,19 +67,18 @@ import Control.Monad.State
     )
 import Control.Monad.Trans ( liftIO )
 import Data.List ( intercalate, union )
-import Data.Maybe ( isJust, catMaybes )
+import Data.Maybe ( isJust )
 import System.Exit ( exitSuccess )
 
 import Darcs.Patch
     ( IsRepoType, RepoPatch, PrimOf
-    , commuteFLorComplain, invert
-    , listTouchedFiles, fromPrims
+    , commuteFL, invert
+    , listTouchedFiles
     )
-import qualified Darcs.Patch ( thing, things, summary )
-import Darcs.Patch.Apply ( Apply, ApplyState )
+import qualified Darcs.Patch ( thing, things )
+import Darcs.Patch.Apply ( ApplyState )
 import Darcs.Patch.Choices
     ( PatchChoices, Slot (..), LabelledPatch
-    , mkPatchChoices, forceFirsts
     , forceFirst, forceLast, forceMatchingFirst
     , forceMatchingLast, getChoices
     , makeEverythingLater, makeEverythingSooner
@@ -91,42 +90,54 @@ import Darcs.Patch.Choices
     , labelPatches
     )
 import Darcs.Patch.Commute ( Commute )
+import Darcs.Patch.Depends ( contextPatches )
+import Darcs.Patch.Ident ( Ident(..), PatchId )
 import Darcs.Patch.Info ( PatchInfo )
 import Darcs.Patch.Inspect ( PatchInspect )
 import Darcs.Patch.Invert ( Invert )
-import Darcs.Patch.Match ( haveNonrangeMatch, matchAPatch )
-import Darcs.Patch.Named.Wrapped ( anonymous )
-import Darcs.Patch.PatchInfoAnd ( PatchInfoAnd, info, n2pia )
-import Darcs.Patch.Set ( PatchSet(..), patchSet2RL )
+import Darcs.Patch.Invertible
+import Darcs.Patch.Match
+    ( Matchable
+    , MatchableRP
+    , haveNonrangeMatch
+    , matchAPatch
+    )
+import Darcs.Patch.Named ( adddeps, anonymous )
+import Darcs.Patch.PatchInfoAnd ( n2pia )
+import Darcs.Patch.Permutations ( commuteWhatWeCanRL )
 import Darcs.Patch.Show ( ShowPatch, ShowContextPatch )
-import Darcs.Patch.Split ( Splitter(applySplitter,canonizeSplit) )
+import Darcs.Patch.Split ( Splitter(..) )
 import Darcs.Patch.TouchesFiles ( selectNotTouching, deselectNotTouching )
-import Darcs.Patch.Type ( PatchType (..) )
-import Darcs.Patch.Witnesses.Eq ( unsafeCompare )
 import Darcs.Patch.Witnesses.Ordered
     ( (:>) (..), (:||:) (..), FL (..)
     , RL (..), filterFL, lengthFL, mapFL
     , mapFL_FL, spanFL, spanFL_M
-    , (+>+), (+<<+), (+>>+)
+    , (+>+), (+<<+)
+    , reverseFL, reverseRL
     )
 import Darcs.Patch.Witnesses.Sealed
     ( FlippedSeal (..), Sealed2 (..)
     , flipSeal, seal2, unseal2
     )
 import Darcs.Patch.Witnesses.WZipper
-    ( FZipper (..), left, right
+    ( FZipper (..), focus, jokers, left, right
     , rightmost, toEnd, toStart
     )
-import Darcs.Repository ( Repository, repoLocation, readRepo, readTentativeRepo )
+import Darcs.Repository ( Repository, repoLocation, readTentativeRepo )
 import Darcs.UI.External ( editText )
 import Darcs.UI.Options.All
-    ( Verbosity(..), Summary(..)
+    ( Verbosity(..), WithSummary(..)
     , WithContext(..), SelectDeps(..), MatchFlag )
 import Darcs.UI.PrintPatch
-    ( printFriendly, printPatch
-    , printPatchPager, showFriendly )
+    ( printContent
+    , printContentWithPager
+    , printFriendly
+    , printSummary
+    , showFriendly
+    )
 import Darcs.Util.English ( Noun (..), englishNum, capitalize )
-import Darcs.Util.Printer ( prefix, putDocLn, putDocLnWith, greenText )
+import Darcs.Util.Path ( AnchoredPath )
+import Darcs.Util.Printer ( putDocLnWith, greenText, vcat )
 import Darcs.Util.Printer.Color ( fancyPrinters )
 import Darcs.Util.Prompt ( PromptConfig (..), askUser, promptChar )
 import Darcs.Util.Tree ( Tree )
@@ -172,7 +183,7 @@ reversed w = w == LastReversed || w == FirstReversed
 -- given.
 data MatchCriterion p = MatchCriterion
    { mcHasNonrange :: Bool
-   , mcFunction :: forall wA wB. WhichChanges -> LabelledPatch p wA wB -> Bool
+   , mcFunction :: forall wA wB. p wA wB -> Bool
    }
 
 data PatchSelectionOptions = PatchSelectionOptions
@@ -180,31 +191,31 @@ data PatchSelectionOptions = PatchSelectionOptions
   , matchFlags :: [MatchFlag]
   , interactive :: Bool
   , selectDeps :: SelectDeps
-  , summary :: Summary
+  , withSummary :: WithSummary
   , withContext :: WithContext
   }
 
--- | A @PatchSelectionContext@ contains all the static settings for selecting
--- patches. See "PatchSelectionM"
-data PatchSelectionContext p = PSC { opts :: PatchSelectionOptions
-                                   , splitter :: Maybe (Splitter p)
-                                   , files :: Maybe [FilePath]
-                                   , matchCriterion :: MatchCriterion p
-                                   , jobname :: String
-                                   , allowSkipAll :: Bool
-                                   , pristine :: Maybe (Tree IO)
-                                   , whichChanges :: WhichChanges
-                                   }
+-- | All the static settings for selecting patches.
+data SelectionConfig p =
+  PSC { opts :: PatchSelectionOptions
+      , splitter :: Maybe (Splitter p)
+      , files :: Maybe [AnchoredPath]
+      , matchCriterion :: MatchCriterion p
+      , jobname :: String
+      , allowSkipAll :: Bool
+      , pristine :: Maybe (Tree IO)
+      , whichChanges :: WhichChanges
+      }
 
--- | A 'PatchSelectionContext' for selecting 'Prim' patches.
-selectionContextPrim :: WhichChanges
-                     -> String
-                     -> PatchSelectionOptions
-                     -> Maybe (Splitter prim)
-                     -> Maybe [FilePath]
-                     -> Maybe (Tree IO)
-                     -> PatchSelectionContext prim
-selectionContextPrim whch jn o spl fs p =
+-- | A 'SelectionConfig' for selecting 'Prim' patches.
+selectionConfigPrim :: WhichChanges
+                    -> String
+                    -> PatchSelectionOptions
+                    -> Maybe (Splitter prim)
+                    -> Maybe [AnchoredPath]
+                    -> Maybe (Tree IO)
+                    -> SelectionConfig prim
+selectionConfigPrim whch jn o spl fs p =
  PSC { opts = o
      , splitter = spl
      , files = fs
@@ -215,13 +226,15 @@ selectionContextPrim whch jn o spl fs p =
      , whichChanges = whch
      }
 
--- | A 'PatchSelectionContext' for selecting full patches ('PatchInfoAnd' patches)
-selectionContext :: (IsRepoType rt, RepoPatch p)
-                 => WhichChanges -> String -> PatchSelectionOptions
-                 -> Maybe (Splitter (PatchInfoAnd rt p))
-                 -> Maybe [FilePath]
-                 -> PatchSelectionContext (PatchInfoAnd rt p)
-selectionContext whch jn o spl fs =
+-- | A 'SelectionConfig' for selecting full ('Matchable') patches
+selectionConfig :: Matchable p
+                 => WhichChanges
+                 -> String
+                 -> PatchSelectionOptions
+                 -> Maybe (Splitter p)
+                 -> Maybe [AnchoredPath]
+                 -> SelectionConfig p
+selectionConfig whch jn o spl fs =
  PSC { opts = o
      , splitter = spl
      , files = fs
@@ -232,15 +245,15 @@ selectionContext whch jn o spl fs =
      , whichChanges = whch
      }
 
--- | A generic 'PatchSelectionContext'.
-selectionContextGeneric :: (IsRepoType rt, RepoPatch p, Invert q)
-                        => (forall wX wY . q wX wY -> Sealed2 (PatchInfoAnd rt p))
-                        -> WhichChanges
-                        -> String
-                        -> PatchSelectionOptions
-                        -> Maybe [FilePath]
-                        -> PatchSelectionContext q
-selectionContextGeneric extract whch jn o fs =
+-- | A generic 'SelectionConfig'.
+selectionConfigGeneric :: Matchable p
+                       => (forall wX wY . q wX wY -> Sealed2 p)
+                       -> WhichChanges
+                       -> String
+                       -> PatchSelectionOptions
+                       -> Maybe [AnchoredPath]
+                       -> SelectionConfig q
+selectionConfigGeneric extract whch jn o fs =
  PSC { opts = o
      , splitter = Nothing
      , files = fs
@@ -252,49 +265,67 @@ selectionContextGeneric extract whch jn o fs =
      }
 
 -- | The dynamic parameters for interactive selection of patches.
-data InteractiveSelectionContext p wX wY =
+data InteractiveSelectionState p wX wY =
  ISC { total :: Int                           -- ^ total number of patches
      , current :: Int                         -- ^ number of already-seen patches
      , lps :: FZipper (LabelledPatch p) wX wY -- ^ the patches we offer
      , choices :: PatchChoices p wX wY        -- ^ the user's choices
      }
 
-type PatchSelectionM p a = ReaderT (PatchSelectionContext p) a
+type PatchSelectionM p a = ReaderT (SelectionConfig p) a
 
 type InteractiveSelectionM p wX wY a =
-    StateT (InteractiveSelectionContext p wX wY)
+    StateT (InteractiveSelectionState p wX wY)
            (PatchSelectionM p IO) a
 
 -- Common match criteria
 
 -- | For commands without @--match@, 'triv' matches all patches
 triv :: MatchCriterion p
-triv = MatchCriterion { mcHasNonrange = False, mcFunction = \ _ _ -> True }
+triv = MatchCriterion { mcHasNonrange = False, mcFunction = \ _ -> True }
 
 -- | 'iswanted' selects patches according to the given match flags
-iswanted :: forall rt p q
-          . (IsRepoType rt, RepoPatch p, Invert q)
-         => (forall wX wY . q wX wY -> Sealed2 (PatchInfoAnd rt p))
+iswanted :: Matchable p
+         => (forall wX wY . q wX wY -> Sealed2 p)
          -> [MatchFlag]
          -> MatchCriterion q
 iswanted extract mflags = MatchCriterion
-    { mcHasNonrange = haveNonrangeMatch (PatchType :: PatchType rt p) mflags
-    , mcFunction = isWantedMcFunction
+    { mcHasNonrange = haveNonrangeMatch mflags
+    , mcFunction = unseal2 (matchAPatch mflags) . extract
     }
-  where
-    isWantedMcFunction w = unseal2 (matchAPatch mflags) . extract_reverse w . unLabel
-    -- TODO inverting should not be necessary here, at least I would expect
-    --
-    -- prop> matchAPatch (invert x) == matchAPatch x
-    extract_reverse w = if reversed w then extract . invert else extract
 
--- | Run a 'PatchSelection' action in the given 'PatchSelectionContext'.
-runSelection :: forall p wX wY . ( Invert p, Commute p, Apply p, PatchInspect p, ShowPatch p
-                                 , ShowContextPatch p, ApplyState p ~ Tree )
+-- | Run a 'PatchSelection' action in the given 'SelectionConfig',
+-- without assuming that patches are invertible.
+runSelection :: ( MatchableRP p, ShowPatch p, ShowContextPatch p
+                , ApplyState p ~ Tree, ApplyState p ~ ApplyState (PrimOf p)
+                )
              => FL p wX wY
-             -> PatchSelectionContext p
+             -> SelectionConfig p
              -> IO ((FL p :> FL p) wX wY)
-runSelection ps psc = runReaderT (selection ps) psc where
+runSelection _ PSC { splitter = Just _ } =
+  -- a Splitter makes sense for prim patches only and these are invertible anyway
+  error "cannot use runSelection with Splitter"
+runSelection ps PSC { matchCriterion = mc, .. } = do
+    unwrapOutput <$> runInvertibleSelection (wrapInput ps) ictx
+  where
+    convertMC :: MatchCriterion p -> MatchCriterion (Invertible p)
+    convertMC MatchCriterion { mcFunction = mcf, .. } =
+      MatchCriterion { mcFunction = withInvertible mcf, .. }
+    ictx = PSC { matchCriterion = convertMC mc, splitter = Nothing, .. }
+    wrapInput = mapFL_FL mkInvertible
+    unwrapOutput (xs :> ys) =
+      mapFL_FL fromPositiveInvertible xs :> mapFL_FL fromPositiveInvertible ys
+
+-- | Run a 'PatchSelection' action in the given 'SelectionConfig',
+-- assuming patches are invertible.
+runInvertibleSelection :: forall p wX wY .
+                          ( Invert p, MatchableRP p, ShowPatch p
+                          , ShowContextPatch p, ApplyState p ~ Tree
+                          )
+                       => FL p wX wY
+                       -> SelectionConfig p
+                       -> IO ((FL p :> FL p) wX wY)
+runInvertibleSelection ps psc = runReaderT (selection ps) psc where
   selection
     | reversed whch = fmap invert . doit . invert
     | otherwise = doit
@@ -361,9 +392,9 @@ runSelection ps psc = runReaderT (selection ps) psc where
   deselectUnwanted
     | backward whch = forceMatchingFirst (not . iswanted_)
     | otherwise     = forceMatchingLast (not . iswanted_)
-  iswanted_ = mcFunction crit whch
+  iswanted_ = mcFunction crit . unLabel
 
-  {- end of runSelection -}
+  {- end of runInvertibleSelection -}
 
 -- | The equivalent of 'runSelection' for the @darcs log@ command
 viewChanges :: (ShowPatch p, ShowContextPatch p, ApplyState p ~ Tree)
@@ -394,16 +425,15 @@ keysFor :: [[KeyPress]] -> [Char]
 keysFor = concatMap (map kp)
 
 -- | The function for selecting a patch to amend record. Read at your own risks.
-withSelectedPatchFromRepo
-    :: (IsRepoType rt, RepoPatch p, ApplyState p ~ Tree)
+withSelectedPatchFromList
+    :: (Commute p, Matchable p, ShowPatch p, ShowContextPatch p, ApplyState p ~ Tree)
     => String   -- name of calling command (always "amend" as of now)
-    -> Repository rt p wR wU wT
+    -> RL p wO wR
     -> PatchSelectionOptions
-    -> (forall wA . (FL (PatchInfoAnd rt p) :> PatchInfoAnd rt p) wA wR -> IO ())
+    -> (forall wA . (FL p :> p) wA wR -> IO ())
     -> IO ()
-withSelectedPatchFromRepo jn repository o job = do
-    patchSet <- readRepo repository
-    sp <- wspfr jn (matchAPatch $ matchFlags o) (patchSet2RL patchSet) NilFL
+withSelectedPatchFromList jn patches o job = do
+    sp <- wspfr jn (matchAPatch $ matchFlags o) patches NilFL
     case sp of
         Just (FlippedSeal (skipped :> selected')) -> job (skipped :> selected')
         Nothing ->
@@ -419,50 +449,53 @@ data WithSkipped p wX wY = WithSkipped
 -- | This ensures that the selected patch commutes freely with the skipped
 -- patches, including pending and also that the skipped sequences has an
 -- ending context that matches the recorded state, z, of the repository.
-wspfr :: forall rt p wX wY wU. (RepoPatch p, ApplyState p ~ Tree)
+wspfr :: forall p wX wY wU.
+         (Commute p, Matchable p, ShowPatch p, ShowContextPatch p, ApplyState p ~ Tree)
       => String
-      -> (forall wA wB . (PatchInfoAnd rt p) wA wB -> Bool)
-      -> RL (PatchInfoAnd rt p) wX wY
-      -> FL (WithSkipped (PatchInfoAnd rt p)) wY wU
-      -> IO (Maybe (FlippedSeal (FL (PatchInfoAnd rt p) :> PatchInfoAnd rt p) wU))
+      -> (forall wA wB . p wA wB -> Bool)
+      -> RL p wX wY
+      -> FL (WithSkipped p) wY wU
+      -> IO (Maybe (FlippedSeal (FL p :> p) wU))
 wspfr _ _ NilRL _ = return Nothing
 wspfr jn matches remaining@(pps:<:p) skipped
     | not $ matches p = wspfr jn matches pps
                             (WithSkipped SkippedAutomatically p :>: skipped)
     | otherwise =
-    case commuteFLorComplain (p :> mapFL_FL skippedPatch skipped) of
-    Left _  -> do putStrLn "\nSkipping depended-upon patch:"
+    case commuteFL (p :> mapFL_FL skippedPatch skipped) of
+    Nothing -> do putStrLn "\nSkipping depended-upon patch:"
                   defaultPrintFriendly p
                   wspfr jn matches pps (WithSkipped SkippedAutomatically p :>: skipped)
 
-    Right (skipped' :> p') -> do
+    Just (skipped' :> p') -> do
         defaultPrintFriendly p
-        yorn <- promptChar
+        let repeatThis = do
+              yorn <- promptChar
                     PromptConfig { pPrompt = prompt'
                                  , pBasicCharacters = keysFor basicOptions
                                  , pAdvancedCharacters = keysFor advancedOptions
                                  , pDefault = Just 'n'
                                  , pHelp = "?h" }
-        case yorn of
-            'y' -> return $ Just $ flipSeal $ skipped' :> p'
-            'n' -> nextPatch
-            'j' -> nextPatch
-            'k' -> previousPatch remaining skipped
-            'v' -> printPatch p >> repeatThis
-            'p' -> printPatchPager p >> repeatThis
-            'x' -> do putDocLn $ prefix "    " $ Darcs.Patch.summary p
-                      repeatThis
-            'q' -> do putStrLn $ (capitalize jn) ++ " cancelled."
-                      exitSuccess
-            _   -> do putStrLn $ helpFor jn basicOptions advancedOptions
-                      repeatThis
-  where repeatThis = wspfr jn matches (pps:<:p) skipped
-        prompt' = "Shall I " ++ jn ++ " this patch?"
+              case yorn of
+                'y' -> return $ Just $ flipSeal $ skipped' :> p'
+                'n' -> nextPatch
+                'j' -> nextPatch
+                'k' -> previousPatch remaining skipped
+                'v' -> printContent p >> repeatThis
+                'p' -> printContentWithPager p >> repeatThis
+                'x' -> do printSummary p
+                          repeatThis
+                'r' -> defaultPrintFriendly p >> repeatThis
+                'q' -> do putStrLn $ (capitalize jn) ++ " cancelled."
+                          exitSuccess
+                _   -> do putStrLn $ helpFor jn basicOptions advancedOptions
+                          repeatThis
+        repeatThis
+  where prompt' = "Shall I " ++ jn ++ " this patch?"
         nextPatch = wspfr jn matches pps (WithSkipped SkippedManually p:>:skipped)
-        previousPatch :: RL (PatchInfoAnd rt p) wX wQ
-                      -> FL (WithSkipped (PatchInfoAnd rt p)) wQ wU
+        previousPatch :: RL p wX wQ
+                      -> FL (WithSkipped p) wQ wU
                       -> IO (Maybe (FlippedSeal
-                              (FL (PatchInfoAnd rt p) :> PatchInfoAnd rt p) wU))
+                              (FL p :> p) wU))
         previousPatch remaining' NilFL = wspfr jn matches remaining' NilFL
         previousPatch remaining' (WithSkipped sk prev :>: skipped'') =
             case sk of
@@ -478,6 +511,7 @@ wspfr jn matches remaining@(pps:<:p) skipped
                     [[ KeyPress 'v' "view this patch in full"
                      , KeyPress 'p' "view this patch in full with pager"
                      , KeyPress 'x' "view a summary of this patch"
+                     , KeyPress 'r' "view this patch"
                      , KeyPress 'q' ("cancel " ++ jn)
                     ]]
         defaultPrintFriendly =
@@ -496,25 +530,30 @@ liftChoices act = do
 justDone :: Int -> InteractiveSelectionM p wX wY ()
 justDone n = modify $ \isc -> isc{ current = current isc + n}
 
+initialSelectionState :: FL (LabelledPatch p) wX wY
+                      -> PatchChoices p wX wY
+                      -> InteractiveSelectionState p wX wY
+initialSelectionState lps pcs =
+  ISC { total = lengthFL lps
+      , current = 0
+      , lps = FZipper NilRL lps
+      , choices = pcs
+      }
+
 -- | The actual interactive selection process.
 textSelect :: ( Commute p, Invert p, ShowPatch p, ShowContextPatch p
               , PatchInspect p, ApplyState p ~ Tree )
            => FL (LabelledPatch p) wX wY
            -> PatchChoices p wX wY
            -> PatchSelectionM p IO (PatchChoices p wX wY)
-textSelect lps' pcs = do
-    userSelection <- execStateT (skipMundane >>
-                                 showCur >>
-                                 textSelectIfAny)
-                     ISC { total = lengthFL lps'
-                         , current = 0
-                         , lps = FZipper NilRL lps'
-                         , choices = pcs }
-    return $ choices userSelection
-    where textSelectIfAny = do
-            z <- gets lps
-            unless (rightmost z) $
-              textSelect'
+textSelect lps' pcs =
+  choices <$>
+    execStateT (skipMundane >> printCurrent >> textSelectIfAny)
+      (initialSelectionState lps' pcs)
+  where
+    textSelectIfAny = do
+      z <- gets lps
+      unless (rightmost z) $ textSelect'
 
 textSelect' :: ( Commute p, Invert p, ShowPatch p, ShowContextPatch p
                , PatchInspect p, ApplyState p ~ Tree )
@@ -541,6 +580,7 @@ optionsView :: String -> String -> [KeyPress]
 optionsView aThing someThings =
     [ KeyPress 'v' ("view this "++aThing++" in full")
     , KeyPress 'p' ("view this "++aThing++" in full with pager")
+    , KeyPress 'r' ("view this "++aThing)
     , KeyPress 'l' ("list all selected "++someThings) ]
 
 optionsSummary :: String -> [KeyPress]
@@ -594,27 +634,21 @@ options single = do
          ,[optionsSplit split aThing]
          ++ [optionsFile jn | single]
          ++ [optionsView aThing someThings ++
-                if summary o == YesSummary
+                if withSummary o == YesSummary
                     then []
                     else optionsSummary aThing]
          ++ [optionsQuit jn allowsa someThings]
          ++ [optionsNav aThing False]
          )
 
--- | Returns a @Sealed2@ version of the patch we are asking the user
+-- | Returns a 'Sealed2' version of the patch we are asking the user
 -- about.
 currentPatch :: InteractiveSelectionM p wX wY (Maybe (Sealed2 (LabelledPatch p)))
-currentPatch = do
-  FZipper _ lps_todo <- gets lps
-  case lps_todo of
-    NilFL -> return Nothing
-    (lp:>:_) -> return $ Just (Sealed2 lp)
+currentPatch = focus <$> gets lps
 
 -- | Returns the patches we have yet to ask the user about.
 todo :: InteractiveSelectionM p wX wY (FlippedSeal (FL (LabelledPatch p)) wY)
-todo = do
-    (FZipper _ lps_todo) <- gets lps
-    return (FlippedSeal lps_todo)
+todo = jokers <$> gets lps
 
 -- | Modify the underlying @PatchChoices@ by some function
 modifyChoices :: (PatchChoices p wX wY -> PatchChoices p wX wY)
@@ -624,7 +658,7 @@ modifyChoices f = modify $ \isc -> isc{choices = f $ choices isc}
 -- | returns @Just f@ if the 'currentPatch' only modifies @f@,
 -- @Nothing@ otherwise.
 currentFile :: (PatchInspect p)
-            => InteractiveSelectionM p wX wY (Maybe FilePath)
+            => InteractiveSelectionM p wX wY (Maybe AnchoredPath)
 currentFile = do
   c <- currentPatch
   return $ case c of
@@ -648,12 +682,12 @@ decide takeOrDrop lp = do
 
 -- | like 'decide', but for all patches touching @file@
 decideWholeFile :: (Commute p, PatchInspect p)
-                => FilePath -> Bool -> InteractiveSelectionM p wX wY ()
-decideWholeFile file takeOrDrop =
+                => AnchoredPath -> Bool -> InteractiveSelectionM p wX wY ()
+decideWholeFile path takeOrDrop =
     do
       FlippedSeal lps_todo <- todo
       let patches_to_skip =
-              filterFL (\lp' -> listTouchedFiles lp' == [file]) lps_todo
+              filterFL (\lp' -> listTouchedFiles lp' == [path]) lps_todo
       mapM_ (unseal2 $ decide takeOrDrop) patches_to_skip
 
 -- | Undecide the current patch.
@@ -696,37 +730,21 @@ splitCurrent s = do
                                                    (choices isc)
                                       }
 
--- | Shows the patch that is actually being selected the way the user
--- should see it.
-repr :: Invert p => WhichChanges -> LabelledPatch p wX wY -> Sealed2 p
-repr w p
-  | reversed w = Sealed2 (invert (unLabel p))
-  | otherwise = Sealed2 (unLabel p)
-
--- | Returns a list of the currently selected patches, in
--- their original context, i.e., not commuted past unselected
--- patches.
-selected :: (Commute p, Invert p) => InteractiveSelectionM p wX wY [Sealed2 p]
-selected = do
-  w <- asks whichChanges
-  chs <- gets choices
-  (first_chs :> _ :> last_chs) <- return $ getChoices chs
-  return $ if backward w then mapFL (repr w) last_chs else mapFL (repr w) first_chs
-
--- | Prints the list of the selected patches. See 'selected'.
-printSelected :: (Invert p, Commute p, ShowPatch p) =>
-                InteractiveSelectionM p wX wY ()
+-- | Print the list of the selected patches. We currently choose to display
+-- them in "commuted" form, that is, in the order in which they have been
+-- selected and with deselected patches moved out of the way.
+printSelected :: (Commute p, ShowPatch p) => InteractiveSelectionM p wX wY ()
 printSelected = do
   someThings <- things
   o <- asks opts
-  s <- selected
-  liftIO $ do
-    putDocLnWith fancyPrinters $ greenText $ "---- selected "++someThings++" ----"
-    mapM_ (putDocLnWith fancyPrinters . unseal2 (showFriendly (verbosity o) (summary o))) s
-    putDocLnWith fancyPrinters $ greenText $ "---- end of selected "++someThings++" ----"
-
-printSummary :: ShowPatch p => p wX wY -> IO ()
-printSummary = putDocLn . prefix "    " . Darcs.Patch.summary
+  w <- asks whichChanges
+  let showFL = vcat . mapFL (showFriendly (verbosity o) (withSummary o) . unLabel)
+  (first_chs :> _ :> last_chs) <- getChoices <$> gets choices
+  liftIO $ putDocLnWith fancyPrinters $ vcat
+    [ greenText $ "---- selected "++someThings++" ----"
+    , if backward w then showFL last_chs else showFL first_chs
+    , greenText $ "---- end of selected "++someThings++" ----"
+    ]
 
 -- | Skips all remaining patches.
 skipAll ::  InteractiveSelectionM p wX wY ()
@@ -786,7 +804,7 @@ promptUser single def = do
                                    }
 
 -- | Ask the user what to do with the next patch.
-textSelectOne :: ( Invert p, Commute p, ShowPatch p, ShowContextPatch p, PatchInspect p
+textSelectOne :: ( Commute p, ShowPatch p, ShowContextPatch p, PatchInspect p
                  , ApplyState p ~ Tree )
               => InteractiveSelectionM p wX wY Bool
 textSelectOne = do
@@ -799,12 +817,12 @@ textSelectOne = do
          spl <- asks splitter
          whichch <- asks whichChanges
          let singleFile = isSingleFile (unLabel lp)
-             reprCur = repr whichch lp
+             p = unLabel lp
          (basicOptions,advancedOptions) <- options singleFile
          theSlot <- liftChoices $ state $ patchSlot lp
          let the_default = getDefault (backward whichch) theSlot
          yorn <- promptUser singleFile the_default
-         let nextPatch = skipMundane >> showCur
+         let nextPatch = skipMundane >> printCurrent
          case yorn of
                'y' -> decide True lp >> skipOne >> nextPatch
                       >> return False
@@ -812,7 +830,7 @@ textSelectOne = do
                       >> return False
                'w' -> postponeNext >> skipOne >> nextPatch
                       >> return False
-               'e' | (Just s) <- spl -> splitCurrent s >> showCur
+               'e' | (Just s) <- spl -> splitCurrent s >> printCurrent
                                         >> return False
                's' -> currentFile >>= maybe
                        (return ())
@@ -822,12 +840,13 @@ textSelectOne = do
                        (return ())
                        (\f -> decideWholeFile f True) >> nextPatch
                        >> return False
-               'v' -> liftIO $ unseal2 printPatch reprCur >> return False
-               'p' -> liftIO $ unseal2 printPatchPager reprCur >> return False
-               'l' -> printSelected >> showCur >> return False
-               'x' -> liftIO $ unseal2 printSummary reprCur >> return False
+               'v' -> liftIO $ printContent p >> return False
+               'p' -> liftIO $ printContentWithPager p >> return False
+               'r' -> printCurrent >> return False
+               'l' -> printSelected >> printCurrent >> return False
+               'x' -> liftIO $ printSummary p >> return False
                'd' -> skipAll >> return True
-               'g' -> backAll >> showCur >> return False
+               'g' -> backAll >> printCurrent >> return False
                'a' ->
                    do
                      askConfirmation
@@ -837,13 +856,13 @@ textSelectOne = do
                'q' -> liftIO $
                       do putStrLn $ capitalize jn ++ " cancelled."
                          exitSuccess
-               'j' -> skipOne >> showCur >> return False
-               'k' -> backOne >> showCur >> return False
+               'j' -> skipOne >> printCurrent >> return False
+               'k' -> backOne >> printCurrent >> return False
                _   -> do
                  liftIO . putStrLn $ helpFor jn basicOptions advancedOptions
                  return False
 
-lastQuestion :: (Commute p, Invert p, ShowPatch p, ShowContextPatch p, ApplyState p ~ Tree)
+lastQuestion :: (Commute p, ShowPatch p, ShowContextPatch p, ApplyState p ~ Tree)
              => InteractiveSelectionM p wX wY Bool
 lastQuestion = do
   jn <- asks jobname
@@ -861,27 +880,25 @@ lastQuestion = do
                  | c `elem` "qn" -> liftIO $
                                     do putStrLn $ jn ++" cancelled."
                                        exitSuccess
-               'g' -> backAll >> showCur >> return False
+               'g' -> backAll >> printCurrent >> return False
                'l' -> printSelected >> return False
-               'k' -> backOne >> showCur >> return False
+               'k' -> backOne >> printCurrent >> return False
                _ -> do
                  liftIO . putStrLn $ helpFor "this confirmation prompt"
                     basicOptions advancedOptions
                  return False
 
 -- | Shows the current patch as it should be seen by the user.
-showCur :: (Invert p, ShowPatch p, ShowContextPatch p, ApplyState p ~ Tree)
-        => InteractiveSelectionM p wX wY ()
-showCur = do
+printCurrent :: (ShowPatch p, ShowContextPatch p, ApplyState p ~ Tree)
+             => InteractiveSelectionM p wX wY ()
+printCurrent = do
   o <- asks opts
-  p <- asks pristine
+  pr <- asks pristine
   c <- currentPatch
-  whichch <- asks whichChanges
   case c of
-      Nothing -> return ()
-      Just (Sealed2 lp) -> do
-             let reprCur = repr whichch lp
-             liftIO . unseal2 (printFriendly p (verbosity o) (summary o) (withContext o)) $ reprCur
+    Nothing -> return ()
+    Just (Sealed2 lp) ->
+      liftIO $ printFriendly pr (verbosity o) (withSummary o) (withContext o) $ unLabel lp
 
 -- | The interactive part of @darcs changes@
 textView :: (ShowPatch p, ShowContextPatch p, ApplyState p ~ Tree)
@@ -891,9 +908,11 @@ textView :: (ShowPatch p, ShowContextPatch p, ApplyState p ~ Tree)
 textView _ _ _ _ [] = return ()
 textView o n_max n
             ps_done ps_todo@(p:ps_todo') = do
-      unseal2 (printFriendly Nothing (verbosity o) (summary o) (withContext o)) p
+      defaultPrintFriendly p
       repeatThis -- prompt the user
     where
+        defaultPrintFriendly =
+          unseal2 (printFriendly Nothing (verbosity o) (withSummary o) (withContext o))
         prev_patch :: IO ()
         prev_patch = case ps_done of
                        [] -> repeatThis
@@ -914,7 +933,8 @@ textView o n_max n
           , KeyPress 'n' "skip to the next patch" ]
         optionsView' =
           [ KeyPress 'v' "view this patch in full"
-          , KeyPress 'p' "view this patch in full with pager" ]
+          , KeyPress 'p' "view this patch in full with pager"
+          , KeyPress 'r' "view this patch" ]
         optionsSummary' =
           [ KeyPress 'x' "view a summary of this patch" ]
         optionsNav' =
@@ -926,7 +946,7 @@ textView o n_max n
         basicOptions = [ options_yn ]
         advancedOptions =
                      (optionsView' ++
-                        if summary o == YesSummary then [] else optionsSummary')
+                        if withSummary o == YesSummary then [] else optionsSummary')
                   : [ optionsNav' ]
         prompt' = "Shall I view this patch? "
                ++ "(" ++ show (n+1) ++ "/" ++ maybe "?" show n_max ++ ")"
@@ -934,11 +954,13 @@ textView o n_max n
         repeatThis = do
           yorn <- promptChar (PromptConfig prompt' (keysFor basicOptions) (keysFor advancedOptions) (Just 'n') "?h")
           case yorn of
-            'y' -> unseal2 printPatch p >> next_patch
+            'y' -> unseal2 printContent p >> next_patch
             'n' -> next_patch
-            'v' -> unseal2 printPatch p >> repeatThis
-            'p' -> unseal2 printPatchPager p >> repeatThis
-            'x' -> do putDocLn $ prefix "    " $ unseal2 Darcs.Patch.summary p
+            'v' -> unseal2 printContent p >> repeatThis
+            'p' -> unseal2 printContentWithPager p >> repeatThis
+            'r' -> do defaultPrintFriendly p
+                      repeatThis
+            'x' -> do unseal2 printSummary p
                       repeatThis
             'q' -> exitSuccess
             'k' -> prev_patch
@@ -959,7 +981,6 @@ skipMundane = do
   o <- asks opts
   crit <- asks matchCriterion
   jn <- asks jobname
-  whichch <- asks whichChanges
   (skipped :> unskipped) <- liftChoices $ spanFL_M
                                  (state . patchSlot >=> return . decided)
                                  lps_todo
@@ -967,8 +988,7 @@ skipMundane = do
   when (numSkipped > 0) . liftIO $ show_skipped o jn numSkipped skipped
   let boringThenInteresting =
           if selectDeps o == AutoDeps
-          then spanFL (not. mcFunction crit whichch)
-                                 unskipped
+          then spanFL (not . mcFunction crit . unLabel) unskipped
           else NilFL :> unskipped
   case boringThenInteresting of
     boring :> interesting ->
@@ -984,7 +1004,8 @@ skipMundane = do
       _these_ n  = show n ++ " already decided " ++ _elem_ n ""
       _elem_ n = englishNum n (Noun "patch")
       showskippedpatch :: ShowPatch p => FL (LabelledPatch p) wY wT -> IO ()
-      showskippedpatch = sequence_ . mapFL (printSummary . unLabel)
+      showskippedpatch =
+        putDocLnWith fancyPrinters . vcat . mapFL (showFriendly NormalVerbosity NoSummary . unLabel)
 
 decided :: Slot -> Bool
 decided InMiddle = False
@@ -1004,26 +1025,32 @@ askAboutDepends :: (IsRepoType rt, RepoPatch p, ApplyState p ~ Tree)
                 -> PatchSelectionOptions
                 -> [PatchInfo] -> IO [PatchInfo]
 askAboutDepends repository pa' ps_opts olddeps = do
-  -- ideally we'd just default the olddeps to yes but still ask about them.
+  -- Ideally we'd just default the olddeps to yes but still ask about them.
   -- SelectChanges doesn't currently (17/12/09) offer a way to do this so would
   -- have to have this support added first.
-  pps <- readTentativeRepo repository (repoLocation repository)
-  pa <- n2pia `fmap` anonymous (fromPrims pa')
-  -- FIXME: this code is completely unreadable
-  FlippedSeal ps <-
-    return $ case pps of PatchSet _ x -> FlippedSeal (x+>>+(pa:>:NilFL))
-  let my_lps = labelPatches Nothing ps
-      pc = mkPatchChoices my_lps
-      tas =
-        case catMaybes (mapFL (\lp -> if pa `unsafeCompare` unLabel lp || info (unLabel lp) `elem` olddeps
-                                          then Just (label lp) else Nothing) my_lps) of
+  pset <- readTentativeRepo repository (repoLocation repository)
+  -- Let the user select only from patches after the last clean tag.
+  -- We do this for efficiency, otherwise independentPatchIds can
+  -- take a /very/ long time to finish. The limitation this imposes
+  -- is a bit arbitrary from a user perspective. Note however that
+  -- contextPatches at least gives us this latest clean tag to select.
+  _ :> untagged <- return $ contextPatches pset
+  -- Note: using anonymous here seems to be safe since we don't store any patches
+  -- and only return a list of PatchInfo
+  pa <- n2pia . flip adddeps olddeps <$> anonymous pa'
+  -- get rid of all (implicit and explicit) dependencies of pa
+  _ :> _ :> non_deps <- return $ commuteWhatWeCanRL (untagged :> pa)
+  candidates :> _ <-
+    runSelection (reverseRL non_deps) $
+      selectionConfig FirstReversed "depend on" ps_opts
+        { matchFlags = [], interactive = True } Nothing Nothing
+  return $ olddeps `union` independentPatchIds (reverseFL candidates)
 
-                [] -> error "askAboutDepends: []"
-                tgs -> tgs
-  Sealed2 ps' <- return $
-    case getChoices (forceFirsts tas pc) of
-      _ :> mc :> _ -> Sealed2 $ mapFL_FL unLabel mc
-  (deps:>_) <- runSelection ps' $
-                 selectionContext FirstReversed "depend on" ps_opts
-                    { matchFlags = [], interactive = True } Nothing Nothing
-  return $ olddeps `union` mapFL info deps
+-- | From an 'RL' of patches select the identities of those that are
+-- not depended upon by later patches.
+independentPatchIds :: (Commute p, Ident p) => RL p wX wY -> [PatchId p]
+independentPatchIds NilRL = []
+independentPatchIds (ps :<: p) =
+  case commuteWhatWeCanRL (ps :> p) of
+    _ :> _ :> non_deps ->
+      ident p : independentPatchIds non_deps
