@@ -1,7 +1,6 @@
 --  Copyright (C) 2009-2011 Petr Rockai
 --
 --  BSD3
-{-# LANGUAGE ScopedTypeVariables, BangPatterns #-}
 
 -- | A few darcs-specific utility functions. These are used for reading and
 -- writing darcs and darcs-compatible hashed trees.
@@ -29,12 +28,10 @@ module Darcs.Util.Tree.Hashed
     , darcsUpdateHashes
     ) where
 
-import Prelude hiding ( lookup, (<$>) )
 import System.FilePath ( (</>) )
 
 import System.Directory( doesFileExist )
 import Codec.Compression.GZip( decompress, compress )
-import Control.Applicative( (<$>) )
 
 import qualified Data.ByteString.Char8 as BC
 import qualified Data.ByteString.Lazy.Char8 as BLC
@@ -43,14 +40,30 @@ import qualified Data.ByteString as B
 
 import Data.List( sortBy )
 import Data.Maybe( fromJust, isJust )
-import Control.Monad.State.Strict
+import Control.Monad.State.Strict (liftIO,when,unless)
 
-import Darcs.Util.Path
-import Darcs.Util.ByteString ( FileSegment, readSegment )
-import Darcs.Util.Hash
-import Darcs.Util.Progress ( debugMessage )
+import Darcs.Prelude
+
+import Darcs.Util.ByteString (FileSegment, readSegment)
+import Darcs.Util.Hash (Hash(..), decodeBase16, encodeBase16, sha256)
+import Darcs.Util.Path (Name, decodeWhiteName, encodeWhiteName)
+import Darcs.Util.Progress (debugMessage)
 import Darcs.Util.Tree
-import Darcs.Util.Tree.Monad
+    ( Blob(..)
+    , ItemType(..)
+    , Tree(..)
+    , TreeItem(..)
+    , addMissingHashes
+    , expand
+    , itemHash
+    , list
+    , listImmediate
+    , makeTreeWithHash
+    , readBlob
+    , updateSubtrees
+    , updateTree
+    )
+import Darcs.Util.Tree.Monad (TreeIO, initialState, runTreeMonad)
 
 ---------------------------------------------------------------------
 -- Utilities for coping with the darcs directory format.
@@ -76,7 +89,7 @@ darcsLocation dir (s,h) = case hash of
     where prefix Nothing = ""
           prefix (Just s') = formatSize s' ++ "-"
           formatSize s' = let n = show s' in replicate (10 - length n) '0' ++ n
-          hash = BC.unpack (encodeBase16 h)
+          hash = showHash h
 
 ----------------------------------------------
 -- Darcs directory format.
@@ -152,7 +165,7 @@ darcsAddMissingHashes = addMissingHashes darcsHash
 -- and with a given @hash@.
 readDarcsHashedDir :: FilePath -> (Maybe Int, Hash) -> IO [(ItemType, Name, Maybe Int, Hash)]
 readDarcsHashedDir dir h = do
-  debugMessage $ "readDarcsHashedDir: " ++ dir ++ " " ++ BC.unpack (encodeBase16 (snd h))
+  debugMessage $ "readDarcsHashedDir: " ++ dir ++ " " ++ showHash (snd h)
   exist <- doesFileExist $ fst (darcsLocation dir h)
   unless exist $ fail $ "error opening " ++ fst (darcsLocation dir h)
   compressed <- readSegment $ darcsLocation dir h
@@ -194,7 +207,8 @@ readDarcsHashedNosize dir hash = readDarcsHashed' True dir (Nothing, hash)
 -- | Write a Tree into a darcs-style hashed directory.
 writeDarcsHashed :: Tree IO -> FilePath -> IO Hash
 writeDarcsHashed tree' dir =
-    do t <- darcsUpdateDirHashes <$> expand tree'
+    do debugMessage $ "writeDarcsHashed " ++ dir
+       t <- darcsUpdateDirHashes <$> expand tree'
        sequence_ [ dump =<< readBlob b | (_, File b) <- list t ]
        let dirs = darcsFormatDir t : [ darcsFormatDir d | (_, SubTree d) <- list t ]
        _ <- mapM (dump . fromJust) dirs
@@ -211,6 +225,7 @@ writeDarcsHashed tree' dir =
 fsCreateHashedFile :: FilePath -> BLC.ByteString -> TreeIO ()
 fsCreateHashedFile fn content =
     liftIO $ do
+      debugMessage $ "fsCreateHashedFile " ++ fn
       exist <- doesFileExist fn
       unless exist $ BL.writeFile fn content
 
@@ -228,8 +243,9 @@ hashedTreeIO action t dir =
           updateItem _ x = return x
 
           updateFile b@(Blob _ !h) = do
+            liftIO $ debugMessage $ "hashedTreeIO.updateFile: " ++ showHash h
             content <- liftIO $ readBlob b
-            let fn = dir </> BC.unpack (encodeBase16 h)
+            let fn = dir </> showHash h
                 nblob = Blob (decompress <$> rblob) h
                 rblob = BL.fromChunks . return <$> B.readFile fn
                 newcontent = compress content
@@ -238,7 +254,10 @@ hashedTreeIO action t dir =
           updateSub s = do
             let !hash = treeHash s
                 Just dirdata = darcsFormatDir s
-                fn = dir </> BC.unpack (encodeBase16 hash)
+                fn = dir </> showHash hash
+            liftIO $ debugMessage $ "hashedTreeIO.updateSub: " ++ showHash hash
             fsCreateHashedFile fn (compress dirdata)
             return s
 
+showHash :: Hash -> String
+showHash = BC.unpack . encodeBase16
